class LockSingler
{
	HANDLE mutex;
	BOOL Locked;
public:
	LockSingler(const LPTSTR lpName = NULL)
	{
		mutex = CreateMutex(NULL, FALSE, lpName);
		Locked = FALSE;
	}
	~LockSingler()
	{
		CloseHandle(mutex);
	}
	BOOL IsLocked()
	{
		return Locked;
	}
	DWORD Lock( const DWORD dwTimeOut = INFINITE )
	{
		Locked = TRUE;
		return WaitForSingleObject(mutex, dwTimeOut);
	}
	BOOL Unlock()
	{
		Locked = FALSE;
		return ReleaseMutex(mutex);
	}
};

class AutoLocker
{
	LockSingler * sema;
public:
	AutoLocker(LockSingler * const s)
	{
		s->Lock();
		sema = s;
	}
	~AutoLocker()
	{
		sema->Unlock();
	}
};

class AutoUnLocker
{
	LockSingler * sema;
public:
	AutoUnLocker(LockSingler * const s)
	{
		sema = s;
	}
	~AutoUnLocker()
	{
		sema->Unlock();
	}
};
