enum TokenKind
{
	NULL_TOKEN = 0,
///////////////////////////////////////////////////////////////////////////////
	POINT_TOKEN,
	COLOCOLO_TOKEN,
	DIA_TOKEN, COLOYAJIRUSHI_TOKEN,
	EXCLAMATION_TOKEN, PLSPLS_TOKEN, MNSMNS_TOKEN,
	SHOUKAKKOOPEN_TOKEN, SHOUKAKKOCLOSE_TOKEN,
	CHUUKAKKOOPEN_TOKEN, CHUUKAKKOCLOSE_TOKEN,
	DAIKAKKOOPEN_TOKEN, DAIKAKKOCLOSE_TOKEN,
	BRACOLOOPEN_TOKEN, BRACOLOCLOSE_TOKEN,
	KAKKOCOLOOPEN_TOKEN, KAKKOCOLOCLOSE_TOKEN,
	SQUOT_TOKEN,
	DQUOT_TOKEN,
	BACKSLA_TOKEN,
	COLON_TOKEN,
	SEMICOLON_TOKEN,
///////////////////////////////////////////////////////////////////////////////
	SHARP_TOKEN,
	KAKERU_TOKEN, WARU_TOKEN, AMARI_TOKEN,
	PULUS_TOKEN, MINUS_TOKEN,
	LSHIFT_TOKEN, RSHIFT_TOKEN,
	SHOU_TOKEN, SHOUOREQUAL_TOKEN, DAI_TOKEN, DAIOREQUAL_TOKEN,
	EQUAL_TOKEN, NOTEQUAL_TOKEN,
	AND_TOKEN,
	XOR_TOKEN,
	OR_TOKEN,
	ANDAND_TOKEN,
	OROR_TOKEN,
	QUESTION_TOKEN,
	DAINYUU_TOKEN, PULUSDAINYUU_TOKEN, MINUSDAINYUU_TOKEN, KAKERUDAINYUU_TOKEN, WARUDAINYUU_TOKEN, AMARIDAINYUU_TOKEN,
	ANDDAINYUU_TOKEN, XORDAINYUU_TOKEN, ORDAINYUU_TOKEN, LSHIFTDAINYUU_TOKEN, RSHIFTDAINYUU_TOKEN,
	TENDAINYUU_TOKEN,
	TENTEN_TOKEN, COMMA_TOKEN,
///////////////////////////////////////////////////////////////////////////////
	TILDE_TOKEN,
///////////////////////////////////////////////////////////////////////////////
	OTHERKIND_TOKEN,
	BOOL_TOKEN, BRANCH_TOKEN, BREAK_TOKEN, BYTE_TOKEN,
	CASE_TOKEN, CAST_TOKEN, CHAR_TOKEN, CLASS_TOKEN, COMM_TOKEN, COND_TOKEN, CONST_TOKEN, CONSTEVAL_TOKEN, CONTINUE_TOKEN, CORE_TOKEN,
	DEFAULT_TOKEN, DELEGATE_TOKEN, DELETE_TOKEN, DIFF_TOKEN, DO_TOKEN, DOUBLE_TOKEN,
	ELSE_TOKEN, ENUM_TOKEN,
	FALSE_TOKEN, FLOAT_TOKEN, FUNC_TOKEN,
	IDNAME_TOKEN, IF_TOKEN, INDEX_TOKEN, INT_TOKEN,
	LEN_TOKEN, LIB_TOKEN, LIST_TOKEN, LOCK_TOKEN, LONG_TOKEN, LOOP_TOKEN,
	MBR_TOKEN, METHOD_TOKEN,
	NEW_TOKEN,
	PLUGIN_TOKEN, PROPERTY_TOKEN, PUBLIC_TOKEN,
	REF_TOKEN, RETURN_TOKEN,
	SELF_TOKEN, SHORT_TOKEN, SIZEOF_TOKEN, SWITCH_TOKEN,
	THREAD_TOKEN, TRUE_TOKEN, TYPE_TOKEN, TYPEID_TOKEN, TYPEIDOF_TOKEN, TYPESIZE_TOKEN,
	UBYTE_TOKEN, UINT_TOKEN, ULONG_TOKEN, USHORT_TOKEN,
	VOID_TOKEN,
	WAIT_TOKEN, WHILE_TOKEN
};

inline bool IsSoushokuToken(const TokenKind t)
{
	return t == PUBLIC_TOKEN || t == CONST_TOKEN || t == PROPERTY_TOKEN || t == CORE_TOKEN;
}

inline bool IsPublicToken(const TokenKind t)
{
	return t == PUBLIC_TOKEN || t == CONST_TOKEN || t == PROPERTY_TOKEN;
}

inline bool IsReadOnlyToken(const TokenKind t)
{
	return t == CONST_TOKEN || t == PROPERTY_TOKEN;
}

inline bool IsKoukaiHensuuToken(const TokenKind t)
{
	return t == PUBLIC_TOKEN || t == PROPERTY_TOKEN;
}

inline void SoushokuTobasi(const intptr_t *const ptypt, intptr_t &i)
{
	if(IsSoushokuToken((TokenKind)ptypt[i])) i++;
}


bool RefToken(PWCArray<intptr_t> &type, const intptr_t *const iap, intptr_t &i)
{
	if(iap[i] == CONST_TOKEN){
		type += CONST_TOKEN;
		i++;
	}
	switch(iap[i]){
	case BOOL_TOKEN:
	case CHAR_TOKEN:
	case DOUBLE_TOKEN:
	case FLOAT_TOKEN:
	case BYTE_TOKEN:
	case INT_TOKEN:
	case LONG_TOKEN:
	case SHORT_TOKEN:
	case UBYTE_TOKEN:
	case UINT_TOKEN:
	case ULONG_TOKEN:
	case USHORT_TOKEN:
	case INDEX_TOKEN:
	case VOID_TOKEN:
	case NULL_TOKEN:
		type += iap[i++];
		break;
	case CLASS_TOKEN:
	case ENUM_TOKEN:
		type += iap[i++];
		type += iap[i++];
		break;
	case REF_TOKEN:
	case LIST_TOKEN:
		type += iap[i++];
		RefToken(type, iap, i);
		break;
	default:
		return false;
	}
	return true;
}

bool FuncToken(PWCArray<intptr_t> &type, const intptr_t *const iap, intptr_t &i)
{
	if(IsReadOnlyToken((TokenKind)iap[i])){
		type += iap[i];
		i++;
	}
	switch(iap[i]){
	case BOOL_TOKEN:
	case CHAR_TOKEN:
	case DOUBLE_TOKEN:
	case FLOAT_TOKEN:
	case BYTE_TOKEN:
	case INT_TOKEN:
	case LONG_TOKEN:
	case SHORT_TOKEN:
	case UBYTE_TOKEN:
	case UINT_TOKEN:
	case ULONG_TOKEN:
	case USHORT_TOKEN:
	case INDEX_TOKEN:
	case VOID_TOKEN:
		type += iap[i++];
		break;
	case ENUM_TOKEN:
		type += iap[i++];
		type += iap[i++];
		break;
	case FUNC_TOKEN:
		i++;
		type += FUNC_TOKEN;
		if(!FuncToken(type, iap, i)) return false;
		break;
	default:
		return false;
	}
	if(iap[i] != KAKKOCOLOOPEN_TOKEN) return false;
	type += iap[i++];
	while(iap[i] != KAKKOCOLOCLOSE_TOKEN){
		type += iap[i++];
		if(iap[i-1] == FUNC_TOKEN){
			FuncToken(type, iap, i);
		}else if(iap[i-1] == CLASS_TOKEN || iap[i-1] == ENUM_TOKEN || iap[i-1] == DELEGATE_TOKEN){
			type += iap[i++];
		}
	}
	type += iap[i++];
	return true;
}

int32_t TokenTtoTypeId(const intptr_t *const iap, intptr_t &i)
{
	if(IsSoushokuToken((TokenKind)iap[i])) i++;
	switch(iap[i++]){
	case BYTE_TOKEN:
		return BYTE_TYPEID;
	case SHORT_TOKEN:
		return SHORT_TYPEID;
	case INT_TOKEN:
		return INT_TYPEID;
	case LONG_TOKEN:
		return LONG_TYPEID;
	case UBYTE_TOKEN:
		return UBYTE_TYPEID;
	case USHORT_TOKEN:
		return USHORT_TYPEID;
	case UINT_TOKEN:
		return UINT_TYPEID;
	case ULONG_TOKEN:
		return ULONG_TYPEID;
	case INDEX_TOKEN:
		return ADDRESS_TYPEID;
	case FLOAT_TOKEN:
		return FLOAT_TYPEID;
	case DOUBLE_TOKEN:
		return DOUBLE_TYPEID;
	case CHAR_TOKEN:
		return CHAR_TYPEID;
	case BOOL_TOKEN:
		return BOOL_TYPEID;
	case VOID_TOKEN:
		return VOID_TYPEID;
	case REF_TOKEN:
		if(iap[i] == NULL_TOKEN){
			i++;
			return DYNREF_TYPEID;
		}
		return REF_TYPEID;
	case LIST_TOKEN:
		if(iap[i] == NULL_TOKEN){
			i++;
			return DYNLIST_TYPEID;
		}
		return LIST_TYPEID;
	case FUNC_TOKEN:
		return FUNC_TYPEID;
	case METHOD_TOKEN:
		return METHOD_TYPEID;
	case DELEGATE_TOKEN:
		return DELEGATE_TYPEID;
	case THREAD_TOKEN:
		return ~(int32_t)iap[i++];
	case CLASS_TOKEN:
	case ENUM_TOKEN:
		return (int32_t)iap[i++];
	}
	return UNKNOWN_TYPEID;
}
