
struct OmCompiler
{
	typedef SourceTree<JunkanGoroshi> SrcTree;
	SSZStatic<JunkanGoroshi> *ss;
	PWCArray<wchar_t> *filename;
	intptr_t gyousuu;
	intptr_t yobunipop;
	struct KuwashiiType
	{
		int32_t id;
		SrcTree *pstre;
		PWCArray<intptr_t> kwsk;
		KuwashiiType()
		{
			pstre = NULL;
		}
		KuwashiiType& operator=(const KuwashiiType &x)
		{
			id = x.id;
			pstre = x.pstre;
			kwsk = x.kwsk;
			return *this;
		}
	};
	class XCompi
	{
	protected:
		OmCompiler *pomc;
		void Backpatch(PWCArray<uint8_t> &bytecode, PWCArray<intptr_t> &plist)
		{
			intptr_t i;
			for(i = 0; i < plist.GetLength(); i++){
				*(intptr_t*)(bytecode.GetConst()+plist.GetConst()[i]) = bytecode.GetLength();
			}
			plist.Clear();
		}
		void KansuStackSet(const SrcTree &stree, PWCArray<uint8_t> &bytecode)
		{
			intptr_t i, siz, fsi;
			fsi = 0;
			for(i = stree.tpcount; i < stree.tpcount+stree.hikisuucount; i++){
				fsi += stree.funchensuu.GetConst()[i].size;
			}
			siz = 0;
			for(; i < stree.funchensuu.GetLength(); i++){
				siz += stree.funchensuu.GetConst()[i].size;
			}
			if(siz > 0){
				siz = Aligner(siz);
				bytecode += HENSUUSTACKKAKUHO_MEIREI;
				AtaiCode(bytecode, siz);
			}else if(fsi > 0){
				bytecode += LOCALPUSH_MEIREI;
			}
		}
		void KansuStackClear(SrcTree &stree, PWCArray<uint8_t> &bytecode)
		{
			intptr_t siz;
			siz = stree.GetMemberSize();
			if(siz + pomc->yobunipop > 0){
				if(stree.GetTypeSize(stree.rettype) > 0){
					bytecode += POPTOHENSUU_MEIREI;
					AtaiCode(bytecode, siz-Aligner(stree.GetTypeSize(stree.rettype)));
				}
				bytecode += HENSUUSTACKKAIHOU_MEIREI;
				AtaiCode(bytecode, siz+pomc->yobunipop-Aligner(stree.GetTypeSize(stree.rettype)));
			}
			pomc->yobunipop = 0;
		}
		intptr_t TypeAtai(PWCArray<uint8_t> &bytecode, const MeireiCode typ, const void *const xp)
		{
			switch(typ){
			case BYTETYPE_MEIREI:
			case UBYTETYPE_MEIREI:
				AtaiCode(bytecode, *(int8_t*)xp);
				return sizeof(int8_t);
			case SHORTTYPE_MEIREI:
			case USHORTTYPE_MEIREI:
				AtaiCode(bytecode, *(int16_t*)xp);
				return sizeof(int16_t);
			case INTTYPE_MEIREI:
			case UINTTYPE_MEIREI:
				AtaiCode(bytecode, *(int32_t*)xp);
				return sizeof(int32_t);
			case LONGTYPE_MEIREI:
			case ULONGTYPE_MEIREI:
				AtaiCode(bytecode, *(int64_t*)xp);
				return sizeof(int64_t);
			case FLOATTYPE_MEIREI:
				AtaiCode(bytecode, *(float*)xp);
				return sizeof(float);
			case DOUBLETYPE_MEIREI:
				AtaiCode(bytecode, *(double*)xp);
				return sizeof(double);
			case CHARTYPE_MEIREI:
				AtaiCode(bytecode, *(wchar_t*)xp);
				return sizeof(wchar_t);
			case BOOLTYPE_MEIREI:
				AtaiCode(bytecode, *(bool*)xp);
				return sizeof(bool);
			case INDEXTYPE_MEIREI:
				AtaiCode(bytecode, *(intptr_t*)xp);
				return sizeof(intptr_t);
			}
			return 0;
		}
		MeireiCode TypeIdtoMeireiT(const SrcTree &stree, const int32_t typ)
		{
			switch(typ){
			case BYTE_TYPEID:
				return BYTETYPE_MEIREI;
			case UBYTE_TYPEID:
				return UBYTETYPE_MEIREI;
			case SHORT_TYPEID:
				return SHORTTYPE_MEIREI;
			case USHORT_TYPEID:
				return USHORTTYPE_MEIREI;
			case INT_TYPEID:
				return INTTYPE_MEIREI;
			case UINT_TYPEID:
				return UINTTYPE_MEIREI;
			case LONG_TYPEID:
				return LONGTYPE_MEIREI;
			case ULONG_TYPEID:
				return ULONGTYPE_MEIREI;
			case FLOAT_TYPEID:
				return FLOATTYPE_MEIREI;
			case DOUBLE_TYPEID:
				return DOUBLETYPE_MEIREI;
			case CHAR_TYPEID:
				return CHARTYPE_MEIREI;
			case BOOL_TYPEID:
				return BOOLTYPE_MEIREI;
			case ADDRESS_TYPEID:
				return INDEXTYPE_MEIREI;
			case FUNC_TYPEID:
			case METHOD_TYPEID:
			case DELEGATE_TYPEID:
				return INDEXTYPE_MEIREI;
			case REF_TYPEID:
			case LIST_TYPEID:
				return REFTYPE_MEIREI;
			case DYNREF_TYPEID:
			case DYNLIST_TYPEID:
				return DYNREFTYPE_MEIREI;
			default:
				if(typ < 0) return NOP_MEIREI;
				switch(stree.stat->funclist.Get(typ)->selftype){
				case ENUM_BLOCK:
					return INTTYPE_MEIREI;
				}
				break;
			}
			return NOP_MEIREI;
		}
		intptr_t Kansuu(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype, SrcTree &st, const bool thrd)
		{
			intptr_t ii, jj, il, bp, sizee;
			int32_t typ, fi;
			SrcTree *retcall = NULL;
			bool refari;
			PWCArray<wchar_t> name;
			if(!st.kakutei) return -1;
			refari = false;
			ii = 0;
			while(iap[i] != SHOUKAKKOCLOSE_TOKEN){
				if(ii >= st.hikisuucount){
					if(ii == 0){
						jj = tystack.GetLength();
						do{
							if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return -1;
							i++;
						}while(iap[i] != SHOUKAKKOCLOSE_TOKEN && iap[i] != COMMA_TOKEN);
						if(jj == tystack.GetLength()) break;
					}
					return -1;
				}
				typ = TokenTtoTypeId(st.funchensuu.GetConst()[st.tpcount+ii].type.GetConst(), il=0);
				tystack += typ;
				do{
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return -1;
					i++;
				}while(iap[i] != SHOUKAKKOCLOSE_TOKEN && iap[i] != COMMA_TOKEN);
				if(typ == UNKNOWN_TYPEID) return -1;
				if(st.funchensuu.GetConst()[st.tpcount+ii].type.GetTail() == ~DAINYUU_TOKEN){
					if(thrd){
						pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"Xbhɂ͎QƓnł܂B");
						return -1;
					}
					if(tystack.GetTail() != UNKNOWN_TYPEID || ktype.id != typ) return -1;
					if(ktype.id == REF_TYPEID || ktype.id == LIST_TYPEID || ktype.id == FUNC_TYPEID || ktype.id == METHOD_TYPEID){
						il = 0;
						if(IsSoushokuToken((TokenKind)ktype.kwsk.GetConst()[il])) il++;
						if(st.funchensuu.GetConst()[st.tpcount+ii].type.GetLength()-1 > ktype.kwsk.GetLength()-il) return -1;
						for(jj = 0; jj < st.funchensuu.GetConst()[st.tpcount+ii].type.GetLength()-1; jj++){
							if(st.funchensuu.GetConst()[st.tpcount+ii].type.GetConst()[jj] != ktype.kwsk.GetConst()[jj+il]) return -1;
						}
					}
					ktype.id = VOID_TYPEID;
					i++;
				}else{
					if(tystack.GetTail() == UNKNOWN_TYPEID) return -1;
					if(tystack.GetTail() == BOOL_TYPEID && ktype.id != BOOL_TYPEID){
						bytecode += BOOLTYPE_MEIREI;
						ktype.id = BOOL_TYPEID;
					}
					if(ktype.id == REF_TYPEID || ktype.id == LIST_TYPEID){
						if(!stree.YuruiKataCheck(ktype.kwsk, st.funchensuu.GetConst()[st.tpcount+ii].type, *pomc->filename, pomc->gyousuu)) return -1;
						if(thrd){
							bytecode += REFCNTUP_MEIREI;
							refari = true;
						}
					}else if(ktype.id == DYNREF_TYPEID || ktype.id == DYNLIST_TYPEID){
						if(!stree.YuruiKataCheck(ktype.kwsk, st.funchensuu.GetConst()[st.tpcount+ii].type, *pomc->filename, pomc->gyousuu)) return -1;
						if(thrd){
							bytecode += REFCNTUP_MEIREI;
							refari = true;
						}
					}else if(ktype.id == FUNC_TYPEID || ktype.id == METHOD_TYPEID){
						if(!stree.YuruiKataCheck(ktype.kwsk, st.funchensuu.GetConst()[st.tpcount+ii].type, *pomc->filename, pomc->gyousuu)) return -1;
					}else if(ktype.id == DELEGATE_TYPEID){
						if(thrd){
							pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"delegate^̓Xbhɓn܂B");
							return -1;
						}
						if(!stree.YuruiKataCheck(ktype.kwsk, st.funchensuu.GetConst()[st.tpcount+ii].type, *pomc->filename, pomc->gyousuu)) return -1;
					}else if(0 <= ktype.id && ktype.id <= ENDCLASS_TYPEID){
						if(!JidouCast(stree, bytecode, tystack, typ, ktype.id)) return -1;
						if(stree.stat->funclist.Get(ktype.id)->selftype == CLASS_BLOCK){
							if(thrd){
								name = L"break";
								if((fi = stree.stat->funclist.Get(ktype.id)->GetFuncId(name, false)) >= 0 && stree.stat->funclist.Get(fi)->token.GetLength() > 0){
									pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"fXgN^̂NX̓Xbhɓn܂B");
									return -1;
								}
							}
							bytecode += LOADSANSHOUKARASANSHOUSAKI_MEIREI;
							AtaiCode(bytecode, stree.stat->funclist.Get(ktype.id)->GetMemberSize());
						}
					}else{
						if(!JidouCast(stree, bytecode, tystack, typ, ktype.id)) return -1;
						ktype.id = typ;
					}
				}
				if(iap[i] == COMMA_TOKEN){
					i++;
					if(iap[i] == SHOUKAKKOCLOSE_TOKEN) return -1;
				}
				ii++;
			}
			tystack.SetLength(tystack.GetLength()-ii*2);
			if(ii != st.hikisuucount) return -1;
			sizee = 0;
			for(jj = st.tpcount; jj < st.tpcount+st.hikisuucount; jj++){
				if(st.funchensuu.GetConst()[jj].type.GetTail() == ~DAINYUU_TOKEN){
					sizee += Aligner(sizeof(intptr_t));
				}else{
					sizee += Aligner(stree.GetTypeSize(st.funchensuu.GetConst()[jj].type));
				}
			}
			if(thrd){
				if(ktype.id != ADDRESS_TYPEID) bytecode += INDEXTYPE_MEIREI;
				bytecode += PUSHTEISUU_MEIREI;
				AtaiCode(bytecode, sizee);
				bytecode += PUSHTEISUU_MEIREI;
				if(st.byteadress == 0){
					st.backpatch += bytecode.GetLength();
				}
				AtaiCode(bytecode, st.byteadress);
				if(refari){
					if(st.sonota.thatoshori == 0){
						bytecode += JUMP_MEIREI;
						bp = bytecode.GetLength();
						AtaiCode(bytecode, (intptr_t)0);
						st.sonota.thatoshori = bytecode.GetLength();
						for(jj = st.tpcount+st.hikisuucount-1; jj >= st.tpcount; jj--){
							if(st.funchensuu.GetConst()[jj].type.GetConst()[0]==REF_TOKEN || st.funchensuu.GetConst()[jj].type.GetConst()[0]==LIST_TOKEN){
								bytecode += MEMBERKARASANSHOU_MEIREI;
								AtaiCode(bytecode, st.GetMemberOffset(jj) - (st.GetMemberSize() - sizee));
								if(!RefCountDown(stree, bytecode, ktype, SANSHOUREFCNTDOWN_MEIREI, st.funchensuu.GetConst()[jj].type, true)) return -1;
							}
						}
						bytecode += RETURN_MEIREI;
						*(intptr_t *)(bytecode.GetConst()+bp) = bytecode.GetLength();
					}
					bytecode += PUSHTEISUU_MEIREI;
					AtaiCode(bytecode, st.sonota.thatoshori);
				}else{
					bytecode += PUSHTEISUU_MEIREI;
					AtaiCode(bytecode, (intptr_t)0);
				}
				ktype.id = ADDRESS_TYPEID;
			}
			return sizee;
		}
		intptr_t KansuuPointer(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype, const SrcTree *const pmbrstree, PWCArrayQuad< PWCArray<intptr_t> > &waketa)
		{
			intptr_t ii, jj, il, sizee;
			int32_t typ;
			SrcTree *retcall = NULL;
			if((ktype.id != FUNC_TYPEID && ktype.id != METHOD_TYPEID && ktype.id != DELEGATE_TYPEID) || !stree.KansuuKata(pmbrstree, ktype.kwsk, waketa)) return -1;
			ii = 0;
			while(iap[i] != KAKKOCOLOCLOSE_TOKEN){
				if(ii+1 >= waketa.GetLength()) return -1;
				typ = TokenTtoTypeId(waketa.GetConst()[ii+1].GetConst(), il=0);
				tystack += typ;
				do{
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return -1;
					i++;
				}while(iap[i] != KAKKOCOLOCLOSE_TOKEN && iap[i] != COMMA_TOKEN);
				if(typ == UNKNOWN_TYPEID) return -1;
				if(waketa.GetConst()[ii+1].GetTail() == ~DAINYUU_TOKEN){
					if(tystack.GetTail() != UNKNOWN_TYPEID || ktype.id != typ) return -1;
					if(ktype.id == REF_TYPEID || ktype.id == LIST_TYPEID || ktype.id == FUNC_TYPEID || ktype.id == METHOD_TYPEID){
						il = 0;
						if(IsSoushokuToken((TokenKind)ktype.kwsk.GetConst()[il])) il++;
						if(waketa.GetConst()[ii+1].GetLength()-1 > ktype.kwsk.GetLength()-il) return -1;
						for(jj = 0; jj < waketa.GetConst()[ii+1].GetLength()-1; jj++){
							if(waketa.GetConst()[ii+1].GetConst()[jj] != ktype.kwsk.GetConst()[jj+il]) return -1;
						}
					}
					ktype.id = VOID_TYPEID;
					i++;
				}else{
					if(tystack.GetTail() == UNKNOWN_TYPEID) return -1;
					if(tystack.GetTail() == BOOL_TYPEID && ktype.id != BOOL_TYPEID){
						bytecode += BOOLTYPE_MEIREI;
						ktype.id = BOOL_TYPEID;
					}
					if(ktype.id == REF_TYPEID || ktype.id == LIST_TYPEID || ktype.id == DYNREF_TYPEID || ktype.id == DYNLIST_TYPEID 
						|| ktype.id == FUNC_TYPEID || ktype.id == METHOD_TYPEID || ktype.id == DELEGATE_TYPEID){
						if(!stree.YuruiKataCheck(ktype.kwsk, waketa.GetConst()[ii+1], *pomc->filename, pomc->gyousuu)) return -1;
					}else if(0 <= ktype.id && ktype.id <= ENDCLASS_TYPEID){
						if(!JidouCast(stree, bytecode, tystack, typ, ktype.id)) return -1;
						if(stree.stat->funclist.Get(ktype.id)->selftype == CLASS_BLOCK){
							bytecode += LOADSANSHOUKARASANSHOUSAKI_MEIREI;
							AtaiCode(bytecode, stree.stat->funclist.Get(ktype.id)->GetMemberSize());
						}
					}else{
						if(!JidouCast(stree, bytecode, tystack, typ, ktype.id)) return -1;
						ktype.id = typ;
					}
				}
				if(iap[i] == COMMA_TOKEN){
					i++;
					if(iap[i] == KAKKOCOLOCLOSE_TOKEN) return -1;
				}
				ii++;
			}
			tystack.SetLength(tystack.GetLength()-ii*2);
			if(ii+1 != waketa.GetLength()) return -1;
			sizee = 0;
			for(jj = 1; jj < waketa.GetLength(); jj++){
				if(waketa.GetConst()[jj].GetTail() == ~DAINYUU_TOKEN){
					sizee += Aligner(sizeof(intptr_t));
				}else{
					sizee += Aligner(stree.GetTypeSize(waketa.GetConst()[jj]));
				}
			}
			return sizee;
		}
		intptr_t PluginKansuu(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype, const SrcTree *const pmbrstree, PWCArrayQuad< PWCArray<intptr_t> > &waketa)
		{
			intptr_t ii, jj, il, sizee;
			int32_t typ;
			SrcTree *retcall = NULL;
			if(!stree.KansuuKata(pmbrstree, ktype.kwsk, waketa)) return -1;
			ii = 0;
			while(iap[i] != KAKKOCOLOCLOSE_TOKEN){
				if(ii+1 >= waketa.GetLength()) return -1;
				typ = TokenTtoTypeId(waketa.GetConst()[ii+1].GetConst(), il=0);
				tystack += typ;
				do{
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return -1;
					i++;
				}while(iap[i] != KAKKOCOLOCLOSE_TOKEN && iap[i] != COMMA_TOKEN);
				if(typ == UNKNOWN_TYPEID) return -1;
				if(waketa.GetConst()[ii+1].GetTail() == ~DAINYUU_TOKEN){
					if(tystack.GetTail() != UNKNOWN_TYPEID || ktype.id != typ) return -1;
					if(ktype.id == REF_TYPEID || ktype.id == LIST_TYPEID || ktype.id == FUNC_TYPEID || ktype.id == METHOD_TYPEID){
						il = 0;
						if(IsSoushokuToken((TokenKind)ktype.kwsk.GetConst()[il])) il++;
						if(waketa.GetConst()[ii+1].GetLength()-1 > ktype.kwsk.GetLength()-il) return -1;
						for(jj = 0; jj < waketa.GetConst()[ii+1].GetLength()-1; jj++){
							if(waketa.GetConst()[ii+1].GetConst()[jj] != ktype.kwsk.GetConst()[jj+il]) return -1;
						}
					}
					bytecode += INDEXTYPE_MEIREI;
					bytecode += POPTOARG_MEIREI;
					ktype.id = ADDRESS_TYPEID;
					i++;
				}else{
					if(tystack.GetTail() == UNKNOWN_TYPEID) return -1;
					if(tystack.GetTail() == BOOL_TYPEID && ktype.id != BOOL_TYPEID){
						bytecode += BOOLTYPE_MEIREI;
						ktype.id = BOOL_TYPEID;
					}
					if(ktype.id == REF_TYPEID || ktype.id == LIST_TYPEID || ktype.id == DYNREF_TYPEID || ktype.id == DYNLIST_TYPEID 
						|| ktype.id == FUNC_TYPEID || ktype.id == METHOD_TYPEID || ktype.id == DELEGATE_TYPEID){
						if(!stree.YuruiKataCheck(ktype.kwsk, waketa.GetConst()[ii+1], *pomc->filename, pomc->gyousuu)) return -1;
						bytecode += POPTOARG_MEIREI;
					}else if(0 <= ktype.id && ktype.id <= ENDCLASS_TYPEID){
						if(!JidouCast(stree, bytecode, tystack, typ, ktype.id)) return -1;
						if(stree.stat->funclist.Get(ktype.id)->selftype == CLASS_BLOCK){
							bytecode += LOADSANSHOUTOARG_MEIREI;
							AtaiCode(bytecode, stree.stat->funclist.Get(ktype.id)->GetMemberSize());
						}else{
							bytecode += POPTOARG_MEIREI;
						}
					}else{
						if(!JidouCast(stree, bytecode, tystack, typ, ktype.id)) return -1;
						ktype.id = typ;
						bytecode += POPTOARG_MEIREI;
					}
				}
				if(iap[i] == COMMA_TOKEN){
					i++;
					if(iap[i] == KAKKOCOLOCLOSE_TOKEN) return -1;
				}
				ii++;
			}
			tystack.SetLength(tystack.GetLength()-ii*2);
			if(ii+1 != waketa.GetLength()) return -1;
			sizee = 0;
			for(jj = 1; jj < waketa.GetLength(); jj++){
				if(waketa.GetConst()[jj].GetTail() == ~DAINYUU_TOKEN){
					bytecode += ARGSANSHOUKARAADDRESS_MEIREI;
					AtaiCode(bytecode, sizee/g_AlignmentSize);
					sizee += Aligner(sizeof(intptr_t));
				}else{
					sizee += Aligner(stree.GetTypeSize(waketa.GetConst()[jj]));
				}
			}
			return sizee;
		}
		bool RefCountDown(SrcTree &stree, PWCArray<uint8_t> &bytecode, KuwashiiType &ktype, const MeireiCode shurui, const PWCArray<intptr_t> &hes, const bool popsuru)
		{
			SrcTree *pst;
			int32_t fi;
			PWCArray<intptr_t> type;
			PWCArray<wchar_t> name;
			intptr_t bkp, bp1, bp2, bp3, il;
			il = 0;
			if(IsSoushokuToken((TokenKind)hes.GetConst()[il])) il++;
			if(hes.GetConst()[il] != REF_TOKEN && hes.GetConst()[il] != LIST_TOKEN) return false;
			if(hes.GetConst()[il+1] == NULL_TOKEN){
				if(ktype.id != DYNREF_TYPEID || ktype.id != DYNLIST_TYPEID){
					bytecode += DYNREFTYPE_MEIREI;
					ktype.id = VOID_TYPEID;
				}
				bytecode += shurui;
				if(popsuru) bytecode += POPADDRESS_MEIREI;
				return true;
			}
			il++;
			if(hes.GetConst()[il] == CONST_TOKEN) il++;
			if(hes.GetConst()[il] == CLASS_TOKEN){
				pst = stree.stat->funclist.Get((int32_t)hes.GetConst()[il+1]);
			}else{
				pst = NULL;
			}
			if(ktype.id != REF_TYPEID && ktype.id != LIST_TYPEID){
				bytecode += REFTYPE_MEIREI;
				ktype.id = VOID_TYPEID;
			}
			if(pst != NULL && pst->IsJunkanAble()){
				switch(shurui){
				case SANSHOUREFCNTDOWN_MEIREI:
					bytecode += SANSHOUCIRCREFCNTDOWN_MEIREI;
					break;
				case SANSHOUREFCOPY_MEIREI:
					bytecode += SANSHOUCIRCREFCOPY_MEIREI;
					break;
				default:
					return false;
				}
			}else{
				bytecode += shurui;
			}
			bkp = bytecode.GetLength();
			AtaiCode(bytecode, (intptr_t)0);
			bytecode += INDEXTYPE_MEIREI;
			if( hes.GetConst()[il] == REF_TOKEN || hes.GetConst()[il] == LIST_TOKEN || hes.GetConst()[il] == THREAD_TOKEN
				|| (pst != NULL && (name = L"break", fi = pst->GetFuncId(name, false)) >= 0 && stree.stat->funclist.Get(fi)->token.GetLength() > 0) )
			{
				bytecode += LOCALPUSH_MEIREI;
				bytecode += PUSHTEISUU_MEIREI;
				AtaiCode(bytecode, (intptr_t)0);
				switch(shurui){
				case SANSHOUREFCNTDOWN_MEIREI:
					bytecode += PUSHHENSUU_MEIREI;
					AtaiCode(bytecode, (intptr_t)0);
					bytecode += REFTYPE_MEIREI;
					bytecode += SANSHOUKARASANSHOUSAKI_MEIREI;
					break;
				case SANSHOUREFCOPY_MEIREI:
					bytecode += REFTYPE_MEIREI;
					bytecode += PUSHHENSUU_MEIREI;
					AtaiCode(bytecode, (intptr_t)0);
					break;
				default:
					return false;
				}
				bytecode += LEN_MEIREI;
				if(pst != NULL){
					AtaiCode(bytecode, pst->GetMemberSize());
				}else if(hes.GetConst()[il] == THREAD_TOKEN){
					AtaiCode(bytecode, (intptr_t)sizeof(Thread));
				}else{
					AtaiCode(bytecode, (intptr_t)sizeof(Reference));
				}
				//
				switch(shurui){
				case SANSHOUREFCNTDOWN_MEIREI:
					bytecode += INDEXTYPE_MEIREI;
					bytecode += PUSHHENSUU_MEIREI;
					AtaiCode(bytecode, (intptr_t)0);
					bytecode += REFTYPE_MEIREI;
					bytecode += SANSHOUKARASANSHOUSAKI_MEIREI;
					break;
				case SANSHOUREFCOPY_MEIREI:
					bytecode += PUSHHENSUU_MEIREI;
					AtaiCode(bytecode, (intptr_t)0);
					break;
				default:
					return false;
				}
				bytecode += INDEXTYPE_MEIREI;
				bytecode += PUSHTEISUU_MEIREI;
				AtaiCode(bytecode, (intptr_t)0);
				bytecode += LISTIDXSANSHOU_MEIREI;
				AtaiCode(bytecode, (intptr_t)1);
				//
				bp1 = bytecode.GetLength();
				//
				bytecode += PUSHHENSUU_MEIREI;
				AtaiCode(bytecode, -Aliszof<intptr_t>());
				bytecode += PUSHHENSUU_MEIREI;
				AtaiCode(bytecode, -Aliszof<intptr_t>()*2);
				bytecode += SHOU_MEIREI;
				//
				bytecode += POPNFALSEJUMP_MEIREI;
				bp2 = bytecode.GetLength();
				AtaiCode(bytecode, (intptr_t)0);
				//
				if(pst != NULL){
					bytecode += PUSHSTACK_MEIREI;
					bytecode += POPTOMEMBERSTACK_MEIREI;
					bytecode += CALL_MEIREI;
					if(stree.stat->funclist.Get(fi)->byteadress == 0){
						stree.stat->funclist.Get(fi)->backpatch += bytecode.GetLength();
					}
					AtaiCode(bytecode, stree.stat->funclist.Get(fi)->byteadress);
					bytecode += MEMBERPOP_MEIREI;
					//
					bytecode += INDEXTYPE_MEIREI;
					bytecode += PUSHTEISUU_MEIREI;
					AtaiCode(bytecode, pst->GetMemberSize());
					bytecode += TASU_MEIREI;
				}else if(hes.GetConst()[il] == THREAD_TOKEN){
					bytecode += PUSHSTACK_MEIREI;
					bytecode += SANSHOUTHREADWAIT_MEIREI;
					bytecode += PUSHTEISUU_MEIREI;
					AtaiCode(bytecode, (intptr_t)sizeof(Thread));
					bytecode += TASU_MEIREI;
				}else{
					type.Clear();
					type.AddTail(hes.GetConst()+il, hes.GetLength()-il);
					RefCountDown(stree, bytecode, ktype, SANSHOUREFCNTDOWN_MEIREI, type, false);
					bytecode += INDEXTYPE_MEIREI;
					bytecode += PUSHTEISUU_MEIREI;
					AtaiCode(bytecode, (intptr_t)sizeof(Reference));
					bytecode += TASU_MEIREI;
				}
				bytecode += HENSUUKARASANSHOU_MEIREI;
				AtaiCode(bytecode, -Aliszof<intptr_t>());
				bytecode += INCL_MEIREI;
				//
				bytecode += JUMP_MEIREI;
				AtaiCode(bytecode, bp1);
				//
				*(intptr_t *)(bytecode.GetConst()+bp2) = bytecode.GetLength();
				switch(shurui){
				case SANSHOUREFCNTDOWN_MEIREI:
					if(!popsuru){
						bytecode += PUSHHENSUU_MEIREI;
						AtaiCode(bytecode, (intptr_t)0);
						bytecode += POPTOHENSUU_MEIREI;
						AtaiCode(bytecode, -Aliszof<intptr_t>());
						bytecode += HENSUUSTACKKAIHOU_MEIREI;
						AtaiCode(bytecode, Aliszof<intptr_t>()*2);
						break;
					}
				case SANSHOUREFCOPY_MEIREI:
					bytecode += HENSUUSTACKKAIHOU_MEIREI;
					AtaiCode(bytecode, Aliszof<intptr_t>()*3);
					break;
				default:
					return false;
				}
			}else if(!popsuru && shurui == SANSHOUREFCNTDOWN_MEIREI){
				bytecode += PUSHSTACK_MEIREI;
			}
			switch(shurui){
			case SANSHOUREFCNTDOWN_MEIREI:
				if(hes.GetConst()[il] == CLASS_TOKEN && pst->IsJunkanAble()){
					bytecode += SANSHOUCIRCREFDELETE_MEIREI;
				}else{
					bytecode += SANSHOUREFDELETE_MEIREI;
				}
				if(popsuru){
					bytecode += JUMP_MEIREI;
					bp3 = bytecode.GetLength();
					AtaiCode(bytecode, (intptr_t)0);
				}
				break;
			case SANSHOUREFCOPY_MEIREI:
				bytecode += LOCALPUSH_MEIREI;
				bytecode += HENSUUKARASANSHOU_MEIREI;
				AtaiCode(bytecode, (intptr_t)0);
				if(hes.GetConst()[il] == CLASS_TOKEN && pst->IsJunkanAble()){
					bytecode += SANSHOUCIRCREFDELETE_MEIREI;
				}else{
					bytecode += SANSHOUREFDELETE_MEIREI;
				}
				if(popsuru){
					bytecode += HENSUUSTACKKAIHOU_MEIREI;
					AtaiCode(bytecode, Aliszof<intptr_t>()+Aliszof<Reference>());
				}else{
					bytecode += HENSUUSTACKKAIHOU_MEIREI;
					AtaiCode(bytecode, Aliszof<Reference>());
				}
				bytecode += JUMP_MEIREI;
				bp3 = bytecode.GetLength();
				AtaiCode(bytecode, (intptr_t)0);
				break;
			default:
				return false;
			}
			*(intptr_t *)(bytecode.GetConst()+bkp) = bytecode.GetLength();
			switch(shurui){
			case SANSHOUREFCNTDOWN_MEIREI:
				if(popsuru){
					bytecode += POPADDRESS_MEIREI;
					*(intptr_t *)(bytecode.GetConst()+bp3) = bytecode.GetLength();
				}
				break;
			case SANSHOUREFCOPY_MEIREI:
				bytecode += REFTYPE_MEIREI;
				bytecode += POP_MEIREI;
				if(popsuru) bytecode += POPADDRESS_MEIREI;
				*(intptr_t *)(bytecode.GetConst()+bp3) = bytecode.GetLength();
				break;
			default:
				return false;
			}
			ktype.id = VOID_TYPEID;
			return true;
		}
		bool RefNew(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype, const MeireiCode shurui, const PWCArray<intptr_t> &hes)
		{
			SrcTree *pst;
			SrcTree *retcall = NULL;
			int32_t fi;
			intptr_t bp1, bp2, bp3, bp4, il;
			PWCArray<intptr_t> type;
			PWCArray<wchar_t> name;
			il = 0;
			if(IsSoushokuToken((TokenKind)hes.GetConst()[il])) il++;
			if(hes.GetConst()[il] != REF_TOKEN && hes.GetConst()[il] != LIST_TOKEN) return false;
			il++;
			if(hes.GetConst()[il] == CONST_TOKEN) il++;
			if(hes.GetConst()[il] == CLASS_TOKEN){
				pst = stree.stat->funclist.Get((int32_t)hes.GetConst()[il+1]);
			}else{
				pst = NULL;
			}
			tystack += ADDRESS_TYPEID;
			do{
				if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
				i++;
			}while(iap[i] != SHOUKAKKOCLOSE_TOKEN);
			if(!JidouCast(stree, bytecode, tystack, ADDRESS_TYPEID, ktype.id)) return false;
			ktype.id = ADDRESS_TYPEID;
			tystack.Pop();
			if(pst != NULL && pst->IsJunkanAble()){
				switch(shurui){
				case SANSHOUREFNEW_MEIREI:
					bytecode += SANSHOUCIRCREFNEW_MEIREI;
					AtaiCode(bytecode, pst->sonota.junkanid);
					break;
				case SANSHOULISTNEW_MEIREI:
					bytecode += SANSHOUCIRCLISTNEW_MEIREI;
					AtaiCode(bytecode, pst->sonota.junkanid);
					break;
				default:
					return false;
				}
			}else{
				switch(shurui){
				case SANSHOUREFNEW_MEIREI:
				case SANSHOULISTNEW_MEIREI:
					bytecode += shurui;
					break;
				default:
					return false;
				}
			}
			if(pst != NULL){
				AtaiCode(bytecode, (intptr_t)pst->GetMemberSize());
			}else{
				type.Clear();
				type.AddTail(hes.GetConst()+il, hes.GetLength()-il);
				AtaiCode(bytecode, (intptr_t)stree.GetTypeSize(type));
			}
			if(
				hes.GetConst()[il] == REF_TOKEN || hes.GetConst()[il] == LIST_TOKEN || hes.GetConst()[il] == THREAD_TOKEN
				|| (pst != NULL && (name = L"default", fi = pst->GetFuncId(name, false)) >= 0 && stree.stat->funclist.Get(fi)->token.GetLength() > 0))
			{
				bytecode += PUSHTEISUU_MEIREI;
				AtaiCode(bytecode, (intptr_t)0);
				bytecode += LOCALPUSH_MEIREI;
				//
				bytecode += PUSHHENSUU_MEIREI;
				AtaiCode(bytecode, Aliszof<intptr_t>()*2);
				//
				bytecode += PUSHHENSUU_MEIREI;
				AtaiCode(bytecode, (intptr_t)0);
				bytecode += PUSHHENSUU_MEIREI;
				AtaiCode(bytecode, Aliszof<intptr_t>());
				bytecode += SHOU_MEIREI;
				//
				bytecode += POPNFALSEJUMP_MEIREI;
				bp4 = bytecode.GetLength();
				AtaiCode(bytecode, (intptr_t)0);
				//
				bytecode += REFTYPE_MEIREI;
				bytecode += SANSHOUKARASANSHOUSAKI_MEIREI;
				bytecode += REFPOSSANSHOU_MEIREI;
				bytecode += INDEXTYPE_MEIREI;
				//
				bytecode += JUMP_MEIREI;
				bp3 = bytecode.GetLength();
				AtaiCode(bytecode, (intptr_t)0);
				//
				bp1 = bytecode.GetLength();
				//
				bytecode += PUSHHENSUU_MEIREI;
				AtaiCode(bytecode, (intptr_t)0);
				bytecode += PUSHHENSUU_MEIREI;
				AtaiCode(bytecode, Aliszof<intptr_t>());
				bytecode += SHOU_MEIREI;
				//
				bytecode += POPNFALSEJUMP_MEIREI;
				bp2 = bytecode.GetLength();
				AtaiCode(bytecode, (intptr_t)0);
				//
				*(intptr_t *)(bytecode.GetConst()+bp3) = bytecode.GetLength();
				//
				if(pst != NULL){
					bytecode += PUSHSTACK_MEIREI;
					bytecode += POPTOMEMBERSTACK_MEIREI;
					bytecode += CALL_MEIREI;
					if(stree.stat->funclist.Get(fi)->byteadress == 0){
						stree.stat->funclist.Get(fi)->backpatch += bytecode.GetLength();
					}
					AtaiCode(bytecode, stree.stat->funclist.Get(fi)->byteadress);
					bytecode += MEMBERPOP_MEIREI;
					//
					bytecode += INDEXTYPE_MEIREI;
					bytecode += PUSHTEISUU_MEIREI;
					AtaiCode(bytecode, (intptr_t)pst->GetMemberSize());
					bytecode += TASU_MEIREI;
				}else{
					bytecode += PUSHSTACK_MEIREI;
					if(hes.GetConst()[il] == THREAD_TOKEN){
						bytecode += SANSHOUSETNULL_MEIREI;
					}else if(hes.GetConst()[il+1] == NULL_TOKEN){
						bytecode += DYNREFTYPE_MEIREI;
						bytecode += SANSHOUSETNULL_MEIREI;
						bytecode += INDEXTYPE_MEIREI;
					}else{
						bytecode += REFTYPE_MEIREI;
						bytecode += SANSHOUSETNULL_MEIREI;
						bytecode += INDEXTYPE_MEIREI;
					}
					bytecode += PUSHTEISUU_MEIREI;
					AtaiCode(bytecode, (intptr_t)stree.GetTypeSize(type));
					bytecode += TASU_MEIREI;
				}
				bytecode += HENSUUKARASANSHOU_MEIREI;
				AtaiCode(bytecode, (intptr_t)0);
				bytecode += INCL_MEIREI;
				//
				bytecode += JUMP_MEIREI;
				AtaiCode(bytecode, bp1);
				//
				*(intptr_t *)(bytecode.GetConst()+bp2) = bytecode.GetLength();
				*(intptr_t *)(bytecode.GetConst()+bp4) = bytecode.GetLength();
				//
				bytecode += HENSUUSTACKKAIHOU_MEIREI;
				AtaiCode(bytecode, Aliszof<intptr_t>()*4);
			}else{
				bytecode += POP_MEIREI;
				bytecode += POP_MEIREI;
			}
			ktype.id = VOID_TYPEID;
			tystack += ktype.id;
			return true;
		}
		bool RefIdxNew(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype, const PWCArray<intptr_t> &hes)
		{
			SrcTree *pst;
			SrcTree *retcall = NULL;
			int32_t fi;
			intptr_t bp1, bp2, il;
			PWCArray<intptr_t> type;
			PWCArray<wchar_t> name;
			il = 0;
			if(IsSoushokuToken((TokenKind)hes.GetConst()[il])) il++;
			if(hes.GetConst()[il] != LIST_TOKEN) return false;
			il++;
			if(hes.GetConst()[il] == CONST_TOKEN) il++;
			if(hes.GetConst()[il] == CLASS_TOKEN){
				pst = stree.stat->funclist.Get((int32_t)hes.GetConst()[il+1]);
			}else{
				pst = NULL;
			}
			do{
				if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
				i++;
			}while(iap[i] != DAIKAKKOCLOSE_TOKEN);
			if(!JidouCast(stree, bytecode, tystack, ADDRESS_TYPEID, ktype.id)) return false;
			ktype.id = ADDRESS_TYPEID;
			if(pst != NULL){
				if(pst->IsJunkanAble()){
					bytecode += SANSHOUCIRCLISTIDXNEW_MEIREI;
					AtaiCode(bytecode, pst->sonota.junkanid);
				}else{
					bytecode += SANSHOULISTIDXNEW_MEIREI;
				}
				AtaiCode(bytecode, (intptr_t)pst->GetMemberSize());
			}else{
				bytecode += SANSHOULISTIDXNEW_MEIREI;
				type.Clear();
				type.AddTail(hes.GetConst()+il, hes.GetLength()-il);
				AtaiCode(bytecode, (intptr_t)stree.GetTypeSize(type));
			}
			if(
				hes.GetConst()[il] == REF_TOKEN || hes.GetConst()[il] == LIST_TOKEN || hes.GetConst()[il] == THREAD_TOKEN
				|| (pst != NULL && (name = L"default", fi = pst->GetFuncId(name, false)) >= 0))
			{
				bytecode += PUSHTEISUU_MEIREI;
				AtaiCode(bytecode, (intptr_t)0);
				bytecode += LOCALPUSH_MEIREI;
				//
				bytecode += PUSHHENSUU_MEIREI;
				AtaiCode(bytecode, Aliszof<intptr_t>()*2);
				//
				bp1 = bytecode.GetLength();
				//
				bytecode += HENSUUKARASANSHOU_MEIREI;
				AtaiCode(bytecode, (intptr_t)0);
				bytecode += INCLKARAMAENOATAI_MEIREI;
				bytecode += PUSHHENSUU_MEIREI;
				AtaiCode(bytecode, Aliszof<intptr_t>());
				bytecode += SHOU_MEIREI;
				//
				bytecode += POPNFALSEJUMP_MEIREI;
				bp2 = bytecode.GetLength();
				AtaiCode(bytecode, (intptr_t)0);
				//
				if(pst != NULL){
					bytecode += PUSHSTACK_MEIREI;
					bytecode += POPTOMEMBERSTACK_MEIREI;
					bytecode += CALL_MEIREI;
					if(stree.stat->funclist.Get(fi)->byteadress == 0){
						stree.stat->funclist.Get(fi)->backpatch += bytecode.GetLength();
					}
					AtaiCode(bytecode, stree.stat->funclist.Get(fi)->byteadress);
					bytecode += MEMBERPOP_MEIREI;
					//
					bytecode += INDEXTYPE_MEIREI;
					bytecode += PUSHTEISUU_MEIREI;
					AtaiCode(bytecode, (intptr_t)pst->GetMemberSize());
					bytecode += TASU_MEIREI;
				}else{
					bytecode += PUSHSTACK_MEIREI;
					if(hes.GetConst()[il] == THREAD_TOKEN){
						bytecode += SANSHOUSETNULL_MEIREI;
					}else if(hes.GetConst()[il+1] == NULL_TOKEN){
						bytecode += DYNREFTYPE_MEIREI;
						bytecode += SANSHOUSETNULL_MEIREI;
						bytecode += INDEXTYPE_MEIREI;
					}else{
						bytecode += REFTYPE_MEIREI;
						bytecode += SANSHOUSETNULL_MEIREI;
						bytecode += INDEXTYPE_MEIREI;
					}
					bytecode += PUSHTEISUU_MEIREI;
					AtaiCode(bytecode, (intptr_t)stree.GetTypeSize(type));
					bytecode += TASU_MEIREI;
				}
				//
				bytecode += JUMP_MEIREI;
				AtaiCode(bytecode, bp1);
				//
				*(intptr_t *)(bytecode.GetConst()+bp2) = bytecode.GetLength();
				//
				bytecode += HENSUUSTACKKAIHOU_MEIREI;
				AtaiCode(bytecode, Aliszof<intptr_t>()*4);
			}else{
				bytecode += POP_MEIREI;
				bytecode += POP_MEIREI;
			}
			return true;
		}
		bool Sanshou(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype, const PWCArray<intptr_t> &otype, const bool dicnstf)
		{
			SrcTree *pst = NULL;
			SrcTree *retcall = NULL;
			MeireiCode deco;
			int32_t tyki;
			KuwashiiType typ;
			intptr_t ig, tmpip;
			PWCArray<intptr_t> typer, type;
			intptr_t indt, il;
			bool tmpb, cnstf, tendeta;
			PWCArrayQuad< PWCArray<intptr_t> > waketa;
			ig = 0;
			typer = otype;
			cnstf = dicnstf || IsReadOnlyToken((TokenKind)typer.GetConst()[0]);
			tendeta = false;
			while(iap[i+1] == POINT_TOKEN){
				tendeta = true;
				il = 0;
				if(IsSoushokuToken((TokenKind)typer.GetConst()[il])) il++;
				if(typer.GetConst()[il] != CLASS_TOKEN && typer.GetConst()[il] != ENUM_TOKEN) return false;
				il++;
				tyki = (int32_t)typer.GetConst()[il];
				i += 2;
				if(iap[i] == MBR_TOKEN) i++;
				pst = stree.stat->funclist.Get(tyki);
				if(iap[i] == SHOUKAKKOOPEN_TOKEN){
					i++;
					if(iap[i] < 0){
						pst = stree.stat->funclist.Get((int32_t)~iap[i]);
						i++;
						if(!pst->kakutei && (pst = pst->GetIjousaki(ig, stree, iap+i)) == NULL) return false;
						if(stree.frp->rettype.GetConst()[0] == CONST_TOKEN && pst->rettype.GetConst()[0] != CONST_TOKEN) return false;
						if(cnstf && !IsReadOnlyToken((TokenKind)pst->rettype.GetConst()[0])) return false;
						if((tmpip = Kansuu(stree, bytecode, iap, i, tystack, ktype, *pst, false)) < 0) return false;
						bytecode += LOCALPUSH_MEIREI;
						if(ktype.id != ADDRESS_TYPEID){
							bytecode += INDEXTYPE_MEIREI;
							ktype.id = ADDRESS_TYPEID;
						}
						bytecode += PUSHHENSUU_MEIREI;
						AtaiCode(bytecode, tmpip);
						if(ig != 0){
							bytecode += PUSHTEISUU_MEIREI;
							AtaiCode(bytecode, ig);
							bytecode += TASU_MEIREI;
						}
						bytecode += POPTOMEMBERSTACK_MEIREI;
						bytecode += CALL_MEIREI;
						if(pst->byteadress == 0){
							pst->backpatch += bytecode.GetLength();
						}
						AtaiCode(bytecode, pst->byteadress);
						if((ktype.id = TokenTtoTypeId(pst->rettype.GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
						ktype.kwsk = pst->rettype;
						tystack += ktype.id;
						if(ktype.id != VOID_TYPEID){
							bytecode += POPTOHENSUU_MEIREI;
							AtaiCode(bytecode, tmpip+Aliszof<intptr_t>()-Aligner(stree.GetTypeSize(pst->rettype)));
						}
						bytecode += HENSUUSTACKKAIHOU_MEIREI;
						AtaiCode(bytecode, Aliszof<intptr_t>()-Aligner(stree.GetTypeSize(pst->rettype)));
						bytecode += MEMBERPOP_MEIREI;
					}else{
						return false;
					}
					break;
				}
				if(iap[i] >= 0) return false;
				indt = ~iap[i];
				ig += pst->GetMemberOffset(indt);
				typer = pst->funchensuu.GetConst()[indt].type;
				cnstf = cnstf || IsReadOnlyToken((TokenKind)typer.GetConst()[0]);
			}
			if(iap[i] >= 0 && iap[i] != DIA_TOKEN && iap[i] != DAIKAKKOCLOSE_TOKEN) return true;
			if(tendeta && typer.GetConst()[0] == CONST_TOKEN){
				bytecode += POPADDRESS_MEIREI;
			}else if(ig != 0){
				if(ktype.id != ADDRESS_TYPEID){
					bytecode += INDEXTYPE_MEIREI;
					ktype.id = ADDRESS_TYPEID;
				}
				bytecode += PUSHTEISUU_MEIREI;
				AtaiCode(bytecode, ig);
				bytecode += TASU_MEIREI;
			}
			il = 0;
			if(IsSoushokuToken((TokenKind)typer.GetConst()[il])) il++;
			if(iap[i+1] == VOID_TOKEN){
				i++;
				if(typer.GetConst()[il] == THREAD_TOKEN){
					if(ktype.id != ADDRESS_TYPEID){
						bytecode += INDEXTYPE_MEIREI;
						ktype.id = ADDRESS_TYPEID;
					}
					bytecode += SANSHOUSETNULL_MEIREI;
				}else if(typer.GetConst()[il] == REF_TOKEN || typer.GetConst()[il] == LIST_TOKEN){
					if(typer.GetConst()[il+1] == NULL_TOKEN){
						if(ktype.id != DYNREF_TYPEID && ktype.id != DYNLIST_TYPEID){
							bytecode += DYNREFTYPE_MEIREI;
							ktype.id = VOID_TYPEID;
						}
						bytecode += SANSHOUSETNULL_MEIREI;
					}else{
						if(ktype.id != REF_TYPEID && ktype.id != LIST_TYPEID){
							bytecode += REFTYPE_MEIREI;
							ktype.id = VOID_TYPEID;
						}
						bytecode += SANSHOUSETNULL_MEIREI;
					}
				}else{
					return false;
				}
				i++;//Z~R΂
				if(iap[i] != SEMICOLON_TOKEN) return false;
				return true;
			}
			if(iap[i+1] == DELETE_TOKEN || iap[i+1] == NEW_TOKEN){
				if(cnstf) return false;
				i++;
				if(typer.GetConst()[il] == THREAD_TOKEN){
					bytecode += SANSHOUTHREADWAIT_MEIREI;
				}else if(typer.GetConst()[il] == REF_TOKEN || typer.GetConst()[il] == LIST_TOKEN){
					if(iap[i] != NEW_TOKEN || iap[i+1] != DAIKAKKOOPEN_TOKEN){
						if(!RefCountDown(stree, bytecode, ktype, SANSHOUREFCNTDOWN_MEIREI, typer, iap[i]==DELETE_TOKEN)) return false;
					}
				}else{
					return false;
				}
				if(iap[i] == NEW_TOKEN){
					i++;
					if(typer.GetConst()[il+1] == CONST_TOKEN) return false;
					if(iap[i] == SHOUKAKKOOPEN_TOKEN){
						i++;
						if(typer.GetConst()[il] == REF_TOKEN){
							if(!RefNew(stree, bytecode, iap, i, tystack, ktype, SANSHOUREFNEW_MEIREI, typer)) return false;
						}else{
							if(!RefNew(stree, bytecode, iap, i, tystack, ktype, SANSHOULISTNEW_MEIREI, typer)) return false;
						}
					}else if(iap[i] == DAIKAKKOOPEN_TOKEN){
						i++;
						if(typer.GetConst()[il] != LIST_TOKEN) return false;
						if(!RefIdxNew(stree, bytecode, iap, i, tystack, ktype, typer)) return false;
						il++;
						type.Clear();
						type.AddTail(typer.GetConst()+il, typer.GetLength()-il);
						if(!Sanshou(stree, bytecode, iap, i, tystack, ktype, type, false)) return false;
					}else{
						return false;
					}
				}else{
					i++;
					if(iap[i] != SEMICOLON_TOKEN) return false;
				}
				return true;
			}
			if(iap[i+1] == SHOUKAKKOOPEN_TOKEN && iap[i+2] == TENTEN_TOKEN){
				if(cnstf) return false;
				i += 3;
				if(typer.GetConst()[il] != THREAD_TOKEN) return false;
				il++;
				if(Kansuu(stree, bytecode, iap, i, tystack, ktype, *stree.stat->funclist.Get((int32_t)typer.GetConst()[il]), true) < 0) return false;
				bytecode += SANSHOUTHREADRUN_MEIREI;
				ktype.id = VOID_TYPEID;
				return true;
			}
			tmpb = typer.GetConst()[il] == CLASS_TOKEN;
			typ.kwsk = typer;
			if((typ.id = TokenTtoTypeId(typer.GetConst(), il=0)) == UNKNOWN_TYPEID){
				if(typer.GetConst()[0] == PLUGIN_TOKEN && iap[i+1] == KAKKOCOLOOPEN_TOKEN){
					i += 2;
					bytecode += POPADDRESS_MEIREI;
					ktype = typ;
					if((tmpip = PluginKansuu(stree, bytecode, iap, i, tystack, ktype, pst, waketa)) < 0) return false;
					if(stree.frp->rettype.GetConst()[0] == CONST_TOKEN && waketa.GetConst()[0].GetConst()[0] != CONST_TOKEN) return false;
					if(cnstf && !IsReadOnlyToken((TokenKind)waketa.GetConst()[0].GetConst()[0])) return false;
					if((ktype.id = TokenTtoTypeId(waketa.GetConst()[0].GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
					ktype.kwsk = waketa.GetConst()[0];
					bytecode += PLUGINCALL_MEIREI;
					bytecode += (uint8_t)(stree.TypeIdtoSize(ktype.id) + (ktype.id == FLOAT_TYPEID || ktype.id == DOUBLE_TYPEID ? 128 : 0));
					AtaiCode(bytecode, typer.GetTail());
					if(ktype.id != VOID_TYPEID){
						bytecode += TypeIdtoMeireiT(stree, ktype.id);
					}
					tystack += ktype.id;
					return true;
				}else{
					return false;
				}
			}
			if(iap[i+1] == NULL_TOKEN && iap[i+2] == TENDAINYUU_TOKEN){
				if(cnstf) return false;
				i += 2;
				if(!Append(stree, bytecode, iap, i, tystack, ktype, typ)) return false;
				return true;
			}
			if(iap[i+1] == NULL_TOKEN && (DAINYUU_TOKEN <= iap[i+2] && iap[i+2] <= TENDAINYUU_TOKEN)){
				if(cnstf) return false;
				i += 2;
				if(!Dainyuu(stree, bytecode, iap, i, tystack, ktype, typ, ig, PUSHSTACKSANSHOUSAKI_MEIREI, tmpb, false)) return false;
				return true;
			}
			if(
				(iap[i+1] == NULL_TOKEN && (iap[i+2] == PLSPLS_TOKEN || iap[i+2] == MNSMNS_TOKEN))
				|| iap[i+1] == PLSPLS_TOKEN || iap[i+1] == MNSMNS_TOKEN)
			{
				if(cnstf) return false;
				i++;
				if(!IncDec(stree, bytecode, iap, i, tystack, ktype, typ)) return false;
				return true;
			}
			ktype.pstre = typ.pstre;
			ktype.kwsk = typ.kwsk;
			if(typ.id != ktype.id){
				if(typ.id == VOID_TYPEID) return false;
				ktype.id = typ.id;
				if(!tmpb && !(ENDTHREAD_TYPEID <= ktype.id && ktype.id < 0)){
					if((deco = TypeIdtoMeireiT(stree, ktype.id)) == NOP_MEIREI) return false;
					bytecode += deco;
				}
			}
			if(cnstf && ENDTHREAD_TYPEID <= ktype.id && ktype.id < 0) return false;
			if(ktype.id != VOID_TYPEID){
				if(tendeta && typer.GetConst()[0] == CONST_TOKEN){
					if(ktype.id == FUNC_TYPEID){
						bytecode += PUSHTEISUU_MEIREI;
						if(!stree.stat->funclist.Get((int32_t)typer.GetTail())->kakutei) return false;
						if(stree.stat->funclist.Get((int32_t)typer.GetTail())->byteadress == 0){
							stree.stat->funclist.Get((int32_t)typer.GetTail())->backpatch += bytecode.GetLength();
						}
						AtaiCode(bytecode, stree.stat->funclist.Get((int32_t)typer.GetTail())->byteadress);
					}else if(ktype.id == REF_TYPEID){
						bytecode += PUSHGLOBALHENSUU_MEIREI;
						AtaiCode(bytecode, typer.GetTail());
					}else{
						if((deco = TypeIdtoMeireiT(stree, ktype.id)) == NOP_MEIREI) return false;
						bytecode += PUSHTEISUU_MEIREI;
						TypeAtai(bytecode, deco, typer.GetConst()+il);
					}
				}else{
					if(ktype.id == METHOD_TYPEID && iap[i+1] == KAKKOCOLOOPEN_TOKEN){
						i += 2;
						ig = -pst->GetMemberOffset(indt);
						if((tmpip = KansuuPointer(stree, bytecode, iap, i, tystack, ktype, pst, waketa)) < 0) return false;
						if(stree.frp->rettype.GetConst()[0] == CONST_TOKEN && waketa.GetConst()[0].GetConst()[0] != CONST_TOKEN) return false;
						if(cnstf && !IsReadOnlyToken((TokenKind)waketa.GetConst()[0].GetConst()[0])) return false;
						bytecode += INDEXTYPE_MEIREI;
						if(tmpip > 0){
							bytecode += LOCALPUSH_MEIREI;
							bytecode += PUSHHENSUU_MEIREI;
							AtaiCode(bytecode, tmpip);
						}else{
							bytecode += PUSHSTACK_MEIREI;
						}
						if(ig != 0){
							bytecode += PUSHTEISUU_MEIREI;
							AtaiCode(bytecode, ig);
							bytecode += TASU_MEIREI;
						}
						bytecode += POPTOMEMBERSTACK_MEIREI;
						if(tmpip > 0){
							bytecode += PUSHHENSUU_MEIREI;
							AtaiCode(bytecode, tmpip);
						}
						bytecode += SANSHOUKARASANSHOUSAKI_MEIREI;
						bytecode += STACKCALL_MEIREI;
						if((ktype.id = TokenTtoTypeId(waketa.GetConst()[0].GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
						ktype.kwsk = waketa.GetConst()[0];
						if(tmpip > 0){
							if(ktype.id != VOID_TYPEID){
								bytecode += POPTOHENSUU_MEIREI;
								AtaiCode(bytecode, tmpip+Aliszof<intptr_t>()-Aligner(stree.GetTypeSize(waketa.GetConst()[0])));
							}
							bytecode += HENSUUSTACKKAIHOU_MEIREI;
							AtaiCode(bytecode, Aliszof<intptr_t>()-Aligner(stree.GetTypeSize(waketa.GetConst()[0])));
						}
						bytecode += MEMBERPOP_MEIREI;
					}else{
						if(tmpb || (ENDTHREAD_TYPEID <= ktype.id && ktype.id < 0)){
						}else if(ktype.id == DELEGATE_TYPEID){
							bytecode += LOADSANSHOUKARASANSHOUSAKI_MEIREI;
							AtaiCode(bytecode, stree.TypeIdtoSize(ktype.id));
						}else{
							bytecode += SANSHOUKARASANSHOUSAKI_MEIREI;
						}
					}
				}
			}
			tystack += ktype.id;
			return true;
		}
		bool IsVoidHikinashiRet(SrcTree &stree, const intptr_t *const iap, intptr_t &i, const SrcTree &st)
		{
			intptr_t oi;
			const PWCArray<intptr_t> *pty;
			if(iap[i] != SHOUKAKKOCLOSE_TOKEN){
				if(st.hikisuucount != 0) return false;
				oi = i;
				if(stree.frp == stree.root){
					if(iap[i] != TILDE_TOKEN) return false;
					i++;
				}
				if(iap[i] >= 0){
					i = oi;
					return false;
				}
				pty = &stree.frp->funchensuu.GetConst()[~iap[i]].type;
				if(pty->GetConst()[0] == REF_TOKEN && pty->GetConst()[1] == OTHERKIND_TOKEN){
					i++;
					i += (iap[i] == VOID_TOKEN ? 2 + (stree.frp == stree.root ? 2 : 1) : 0);//VOID_TOKEN,SEMICOLON_TOKEN
					i += 3;//NULL_TOKEN,DAINYUU_TOKEN,COMMA_TOKEN
				}else{
					i = oi;
					return false;
				}
				if(iap[i] != SHOUKAKKOCLOSE_TOKEN){
					i = oi;
					return false;
				}
			}
			if(iap[i+1] != NULL_TOKEN) return false;
			if(iap[i+2] != RETURN_TOKEN){
				if(iap[i+2] != SHARP_TOKEN) return false;
				oi = i;
				pty = &stree.frp->funchensuu.GetConst()[~iap[i+3+(stree.frp == stree.root ? 1 : 0)]].type;
				if(pty->GetConst()[0] == REF_TOKEN && pty->GetConst()[1] == OTHERKIND_TOKEN){
					i += 2 + (stree.frp == stree.root ? 2 : 1);
				}else{
					return false;
				}
				if(iap[i+1] != NULL_TOKEN || iap[i+2] != RETURN_TOKEN){
					i = oi;
					return false;
				}
				i = oi;
			}
			return true;
		}
		bool MainSwitch(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype, SrcTree *&retcall)
		{
			static intptr_t glbbp;
			SrcTree *pst;
			MeireiCode deco;
			intptr_t ig, bp1, bp2, tmpip;
			int32_t tyki;
			KuwashiiType typ;
			TokenKind tmptk;
			intptr_t indt, il;
			PWCArray<intptr_t> type;
			const PWCArray<intptr_t> *pty;
			bool tmpb, cnstf;
			PWCArrayQuad< PWCArray<intptr_t> > waketa;
			int64_t castvar;
			ktype.pstre = NULL;
			switch(iap[i]){
			case CONST_TOKEN:
				i++;
				if((typ.id = TokenTtoTypeId(iap, i)) == UNKNOWN_TYPEID) return false;
				if(typ.id == REF_TYPEID){
					if(typ.id != ktype.id){
						ktype = typ;
						if((deco = TypeIdtoMeireiT(stree, ktype.id)) == NOP_MEIREI) return false;
						bytecode += deco;
					}
					ktype.kwsk = REF_TOKEN;
					if(!RefToken(ktype.kwsk, iap, i)) return false;
					bytecode += PUSHGLOBALHENSUU_MEIREI;
					AtaiCode(bytecode, iap[i]);
				}else if(typ.id == FUNC_TYPEID || typ.id == METHOD_TYPEID){
					if(typ.id != ktype.id){
						ktype = typ;
						if((deco = TypeIdtoMeireiT(stree, ktype.id)) == NOP_MEIREI) return false;
						bytecode += deco;
					}
					if(ktype.id == METHOD_TYPEID){
						ktype.kwsk = METHOD_TOKEN;
						ktype.kwsk += iap[i++];
					}else{
						ktype.kwsk = FUNC_TOKEN;
					}
					if(!FuncToken(ktype.kwsk, iap, i)) return false;
					bytecode += PUSHTEISUU_MEIREI;
					if(!stree.stat->funclist.Get((int32_t)iap[i])->kakutei) return false;
					if(stree.stat->funclist.Get((int32_t)iap[i])->byteadress == 0){
						stree.stat->funclist.Get((int32_t)iap[i])->backpatch += bytecode.GetLength();
					}
					AtaiCode(bytecode, stree.stat->funclist.Get((int32_t)iap[i])->byteadress);
				}else{
					il = i + (stree.TypeIdtoSize(typ.id)+sizeof(intptr_t)-1)/sizeof(intptr_t);
					if(((KAKERU_TOKEN <= iap[il] && iap[il] <= RSHIFTDAINYUU_TOKEN)
						|| iap[il] == COMMA_TOKEN || iap[il] == SHOUKAKKOCLOSE_TOKEN || iap[il] == COLON_TOKEN)
						&& tystack.GetTail() != WILDCARD_TYPEID){
						if(stree.CastConst(typ.id, iap+i, tystack.GetTail(), &castvar) < 0){
							pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"^G[łB");
							return false;
						}
						if(tystack.GetTail() != ktype.id){
							ktype.id = tystack.GetTail();
							if((deco = TypeIdtoMeireiT(stree, ktype.id)) == NOP_MEIREI) return false;
							bytecode += deco;
						}else{
							if((deco = TypeIdtoMeireiT(stree, ktype.id)) == NOP_MEIREI) return false;
						}
						bytecode += PUSHTEISUU_MEIREI;
						TypeAtai(bytecode, deco, &castvar);
					}else{
						if(typ.id != ktype.id){
							ktype = typ;
							if((deco = TypeIdtoMeireiT(stree, ktype.id)) == NOP_MEIREI) return false;
							bytecode += deco;
						}else{
							if((deco = TypeIdtoMeireiT(stree, ktype.id)) == NOP_MEIREI) return false;
						}
						bytecode += PUSHTEISUU_MEIREI;
						TypeAtai(bytecode, deco, iap+i);
					}
					i = il - 1;
				}
				ktype.pstre = &stree;
				tystack += ktype.id;
				break;
			case DELEGATE_TOKEN:
				i++;
				bytecode += INDEXTYPE_MEIREI;
				bytecode += PUSHMEMBERREGISTER_MEIREI;
				bytecode += PUSHLOCALREGISTER_MEIREI;
				ktype.id = DELEGATE_TYPEID;
				ktype.kwsk = DELEGATE_TOKEN;
				if(!FuncToken(ktype.kwsk, iap, i)) return false;
				bytecode += PUSHTEISUU_MEIREI;
				if(((SrcTree *)iap[i])->byteadress == 0) ((SrcTree *)iap[i])->backpatch += bytecode.GetLength();
				AtaiCode(bytecode, ((SrcTree *)iap[i])->byteadress);
				tystack += ktype.id;
				break;
			case PULUS_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id == BOOL_TYPEID || ktype.id <= ENDCLASS_TYPEID) return false;
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += TASU_MEIREI;
				break;
			case MINUS_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id == BOOL_TYPEID || ktype.id <= ENDCLASS_TYPEID) return false;
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += HIKU_MEIREI;
				break;
			case KAKERU_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id == BOOL_TYPEID || ktype.id <= ENDCLASS_TYPEID) return false;
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += KAKERU_MEIREI;
				break;
			case WARU_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id == BOOL_TYPEID || ktype.id <= ENDCLASS_TYPEID) return false;
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += WARU_MEIREI;
				break;
			case AMARI_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id == BOOL_TYPEID || ktype.id == FLOAT_TYPEID || ktype.id == DOUBLE_TYPEID || ktype.id <= ENDCLASS_TYPEID) return false;
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += AMARI_MEIREI;
				break;
			case AND_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id == FLOAT_TYPEID || ktype.id == DOUBLE_TYPEID || ktype.id <= ENDCLASS_TYPEID) return false;
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += AND_MEIREI;
				break;
			case XOR_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id == FLOAT_TYPEID || ktype.id == DOUBLE_TYPEID || ktype.id <= ENDCLASS_TYPEID) return false;
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += XOR_MEIREI;
				break;
			case OR_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id == FLOAT_TYPEID || ktype.id == DOUBLE_TYPEID || ktype.id <= ENDCLASS_TYPEID) return false;
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += OR_MEIREI;
				break;
			case LSHIFT_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id == BOOL_TYPEID || ktype.id == FLOAT_TYPEID || ktype.id == DOUBLE_TYPEID || ktype.id <= ENDCLASS_TYPEID) return false;
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += LSHIFT_MEIREI;
				break;
			case RSHIFT_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id == BOOL_TYPEID || ktype.id == FLOAT_TYPEID || ktype.id == DOUBLE_TYPEID || ktype.id <= ENDCLASS_TYPEID) return false;
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += RSHIFT_MEIREI;
				break;
			case EQUAL_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id <= ENDCLASS_TYPEID){
					if(ktype.id < 0){
						if((ktype.id != FUNC_TYPEID && ktype.id != METHOD_TYPEID)) return false;
					}else{
						if(stree.stat->funclist.Get(ktype.id)->selftype != ENUM_BLOCK) return false;
					}
				}
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += EQUAL_MEIREI;
				tystack.Tail() = BOOL_TYPEID;
				break;
			case NOTEQUAL_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id <= ENDCLASS_TYPEID){
					if(ktype.id < 0){
						if((ktype.id != FUNC_TYPEID && ktype.id != METHOD_TYPEID)) return false;
					}else{
						if(stree.stat->funclist.Get(ktype.id)->selftype != ENUM_BLOCK) return false;
					}
				}
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += NOTEQUAL_MEIREI;
				tystack.Tail() = BOOL_TYPEID;
				break;
			case SHOU_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id == BOOL_TYPEID) return false;
				if(ktype.id <= ENDCLASS_TYPEID){
					if(ktype.id < 0) return false;
					if(stree.stat->funclist.Get(ktype.id)->selftype != ENUM_BLOCK) return false;
				}
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += SHOU_MEIREI;
				tystack.Tail() = BOOL_TYPEID;
				break;
			case SHOUOREQUAL_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id == BOOL_TYPEID) return false;
				if(ktype.id <= ENDCLASS_TYPEID){
					if(ktype.id < 0) return false;
					if(stree.stat->funclist.Get(ktype.id)->selftype != ENUM_BLOCK) return false;
				}
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += SHOUOREQUAL_MEIREI;
				tystack.Tail() = BOOL_TYPEID;
				break;
			case DAI_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id == BOOL_TYPEID) return false;
				if(ktype.id <= ENDCLASS_TYPEID){
					if(ktype.id < 0) return false;
					if(stree.stat->funclist.Get(ktype.id)->selftype != ENUM_BLOCK) return false;
				}
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += DAI_MEIREI;
				tystack.Tail() = BOOL_TYPEID;
				break;
			case DAIOREQUAL_TOKEN:
				ktype.id = tystack.GetConst()[tystack.GetLength()-2];
				if(ktype.id == BOOL_TYPEID) return false;
				if(ktype.id <= ENDCLASS_TYPEID){
					if(ktype.id < 0) return false;
					if(stree.stat->funclist.Get(ktype.id)->selftype != ENUM_BLOCK) return false;
				}
				if(!JidouCast(stree, bytecode, tystack, ktype.id, tystack.Pop())) return false;
				bytecode += DAIOREQUAL_MEIREI;
				tystack.Tail() = BOOL_TYPEID;
				break;
			case EXCLAMATION_TOKEN:
				if(tystack.GetTail() == UNKNOWN_TYPEID) return false;
				if(tystack.GetTail() == BOOL_TYPEID && ktype.id != BOOL_TYPEID){
					bytecode += BOOLTYPE_MEIREI;
					ktype.id = BOOL_TYPEID;
				}
				if(ktype.id == FLOAT_TYPEID || ktype.id == DOUBLE_TYPEID || ktype.id <= ENDCLASS_TYPEID) return false;
				bytecode += NOT_MEIREI;
				break;
			case DIA_TOKEN:
				if(
					tystack.GetTail() == UNKNOWN_TYPEID || tystack.GetTail() == BOOL_TYPEID
					|| (ktype.id != REF_TYPEID && ktype.id != LIST_TYPEID))
				{
					return false;
				}
				il = 0;
				if(IsSoushokuToken((TokenKind)ktype.kwsk.GetConst()[il])) il++;
				if(ktype.kwsk.GetConst()[il] != REF_TOKEN && ktype.kwsk.GetConst()[il] != LIST_TOKEN) return false;
				il++;
				type.Clear();
				type.AddTail(ktype.kwsk.GetConst()+il, ktype.kwsk.GetLength()-il);
				bytecode += REFPOSSANSHOU_MEIREI;
				bytecode += INDEXTYPE_MEIREI;
				ktype.id = ADDRESS_TYPEID;
				tystack.Pop();
				if(!Sanshou(stree, bytecode, iap, i, tystack, ktype, type, false)) return false;
				break;
			case DAIKAKKOOPEN_TOKEN:
				if(ktype.id == REF_TYPEID){
					deco = REFIDXSANSHOU_MEIREI;
				}else if(ktype.id == LIST_TYPEID){
					deco = LISTIDXSANSHOU_MEIREI;
				}else{
					return false;
				}
				il = 0;
				if(IsSoushokuToken((TokenKind)ktype.kwsk.GetConst()[il])) il++;
				if(ktype.kwsk.GetConst()[il] != REF_TOKEN && ktype.kwsk.GetConst()[il] != LIST_TOKEN) return false;
				il++;
				type.Clear();
				type.AddTail(ktype.kwsk.GetConst()+il, ktype.kwsk.GetLength()-il);
				i++;
				do{
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
					i++;
				}while(iap[i] != DAIKAKKOCLOSE_TOKEN && iap[i] != TENTEN_TOKEN);
				if(!JidouCast(stree, bytecode, tystack, ADDRESS_TYPEID, ktype.id)) return false;
				if(iap[i] == TENTEN_TOKEN){
					i++;
					if(iap[i] != DAIKAKKOCLOSE_TOKEN) return false;
					bytecode += REFTYPE_MEIREI;
					bytecode += SLICE_MEIREI;
					AtaiCode(bytecode, (intptr_t)stree.GetTypeSize(type));
					ktype.id = REF_TYPEID;
					ktype.kwsk = REF_TOKEN;
					ktype.kwsk += type;
					tystack.SetLength(tystack.GetLength()-2);
					tystack.Tail() = ktype.id;
				}else{
					bytecode += deco;
					AtaiCode(bytecode, (intptr_t)stree.GetTypeSize(type));
					ktype.id = ADDRESS_TYPEID;
					tystack.SetLength(tystack.GetLength()-2);
					if(!Sanshou(stree, bytecode, iap, i, tystack, ktype, type, false)) return false;
				}
				break;
			case NULL_TOKEN:
				i++;
				switch(iap[i]){
				case PULUS_TOKEN:
					if(ktype.id == BOOL_TYPEID || ktype.id <= ENDCLASS_TYPEID) return false;
					if(tystack.GetTail() == UNKNOWN_TYPEID || tystack.GetTail() == BOOL_TYPEID) return false;
					break;
				case MINUS_TOKEN:
					if(ktype.id == BOOL_TYPEID || ktype.id <= ENDCLASS_TYPEID) return false;
					if(tystack.GetTail() == UNKNOWN_TYPEID) return false;
					if(tystack.GetTail() == BOOL_TYPEID) return false;
					bytecode += MINUS_MEIREI;
					break;
				case ANDAND_TOKEN:
				case OROR_TOKEN:
					if(!AAOO(stree, bytecode, iap, i, tystack, ktype)) return false;
					break;
				case QUESTION_TOKEN:
					if(!Sankou(stree, bytecode, iap, i, tystack, ktype)) return false;
					break;
				case MBR_TOKEN:
					if(ktype.id != ADDRESS_TYPEID){
						bytecode += INDEXTYPE_MEIREI;
						ktype.id = ADDRESS_TYPEID;
					}
					bytecode += MEMBERKARASANSHOU_MEIREI;
					AtaiCode(bytecode, (intptr_t)0);
					tystack += ADDRESS_TYPEID;
					if(iap[i+1] == NULL_TOKEN && iap[i+2] == DAINYUU_TOKEN){
						if(IsReadOnlyToken((TokenKind)stree.frp->rettype.GetConst()[0])) return false;
						i += 2;
						typ.id = stree.frp->parent->funcid;
						if(!Dainyuu(stree, bytecode, iap, i, tystack, ktype, typ, 0, PUSHSTACKSANSHOUSAKI_MEIREI, true, false)) return false;
					}else{
						ktype.id = stree.frp->parent->funcid;
						tystack.Tail() = ktype.id;
					}
					break;
				case TENTEN_TOKEN:
					if(!JidouCast(stree, bytecode, tystack, ADDRESS_TYPEID, ktype.id)) return false;
					ktype.id = ADDRESS_TYPEID;
					break;
				case SHARP_TOKEN:
					pty = &stree.frp->funchensuu.GetConst()[~iap[i+(stree.frp == stree.root ? 2 : 1)]].type;
					if(pty->GetConst()[0] == REF_TOKEN && pty->GetConst()[1] == OTHERKIND_TOKEN){
						i += (stree.frp == stree.root ? 2 : 1);
					}else{
						tystack.Pop();
					}
					break;
				case RETURN_TOKEN:
					if((typ.id = TokenTtoTypeId(stree.crp->rettype.GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
					if(!JidouCast(stree, bytecode, tystack, typ.id, ktype.id)) return false;
					ktype.id = typ.id;
					if(tystack.GetLength() > 0){
						if(ktype.id == VOID_TYPEID) return false;
						tystack.Pop();
					}
					if(stree.frp == stree.root){//TuubN̕ϐIɉ
						bytecode += GBLDELPUSH_MEIREI;
						AtaiCode(bytecode, (intptr_t)(bytecode.GetLength()+sizeof(intptr_t)+1+sizeof(intptr_t)));
						bytecode += JUMP_MEIREI;
						glbbp = bytecode.GetLength();
						AtaiCode(bytecode, (intptr_t)0);
					}
					break;
				default:
					return false;
				}
				break;
			case RETURN_TOKEN:
				if(stree.frp == stree.root){//TuubN̕ϐIɉ
					if(retcall == NULL) *(intptr_t *)(bytecode.GetData()+glbbp) = bytecode.GetLength();
				}
				typ.id = TokenTtoTypeId(stree.crp->rettype.GetConst(), il=0);
				if(stree.frp != stree.root && typ.id != VOID_TYPEID &&
					TypeIdtoMeireiT(stree, ktype.id) != TypeIdtoMeireiT(stree, typ.id)){
					bytecode += TypeIdtoMeireiT(stree, typ.id);
				}
				if(stree.frp != stree.root && stree.crp->selftype != DELEGATE_BLOCK){
					KansuStackClear(stree, bytecode);
				}
				ktype.id = typ.id;
				if(retcall == NULL){
					if(stree.crp->selftype == DELEGATE_BLOCK){
						bytecode += LOCALPOP_MEIREI;
						bytecode += MEMBERPOP_MEIREI;
					}
					bytecode += RETURN_MEIREI;
				}else{
					bytecode += JUMP_MEIREI;
					if(retcall->byteadress == 0){
						retcall->backpatch += bytecode.GetLength();
					}
					AtaiCode(bytecode, retcall->byteadress);
				}
				break;
			case SHARP_TOKEN:
			case SEMICOLON_TOKEN:
				if(ktype.id != VOID_TYPEID){
					if(tystack.GetTail() == UNKNOWN_TYPEID) return false;
					if(tystack.GetTail() == BOOL_TYPEID && ktype.id != BOOL_TYPEID) bytecode += BOOLTYPE_MEIREI;
					bytecode += POP_MEIREI;
					if(ktype.id == DELEGATE_TYPEID){
						bytecode += POP_MEIREI;
						bytecode += POP_MEIREI;
					}
					ktype.id = VOID_TYPEID;
				}
				tystack.Pop();
				break;
			case LEN_TOKEN:
				if(
					tystack.GetTail() == UNKNOWN_TYPEID || tystack.GetTail() == BOOL_TYPEID
					|| (ktype.id != REF_TYPEID && ktype.id != LIST_TYPEID && ktype.id != DYNREF_TYPEID && ktype.id != DYNLIST_TYPEID))
				{
					return false;
				}
				bytecode += LEN_MEIREI;
				if(ktype.id == REF_TYPEID || ktype.id == LIST_TYPEID){
					il = 0;
					if(IsSoushokuToken((TokenKind)ktype.kwsk.GetConst()[il])) il++;
					if(ktype.kwsk.GetConst()[il] != REF_TOKEN && ktype.kwsk.GetConst()[il] != LIST_TOKEN) return false;
					il++;
					if((typ.id = TokenTtoTypeId(ktype.kwsk.GetConst(), il)) == UNKNOWN_TYPEID) return false;
					AtaiCode(bytecode, stree.TypeIdtoSize(typ.id));
				}
				bytecode += INDEXTYPE_MEIREI;
				ktype.id = ADDRESS_TYPEID;
				tystack.Tail() = ktype.id;
				break;
			case IDNAME_TOKEN:
				if(!JidouCast(stree, bytecode, tystack, INT_TYPEID, ktype.id)) return false;
				i++;
				bytecode += (stree.frp == stree.root ? GLOBALKARASANSHOU_MEIREI : HENSUUKARASANSHOU_MEIREI);
				AtaiCode(bytecode, (stree.frp == stree.root ? stree.sonota.golbalgeta : 0) + stree.crp->GetMemberOffset(iap[i]));
				bytecode += IDNAME_MEIREI;
				ktype.id = REF_TYPEID;
				ktype.kwsk = REF_TOKEN;
				ktype.kwsk += CHAR_TOKEN;
				tystack.Tail() = ktype.id;
				bytecode += REFTYPE_MEIREI;
				break;
			case TYPESIZE_TOKEN:
				if(
					tystack.GetTail() == UNKNOWN_TYPEID || tystack.GetTail() == BOOL_TYPEID
					|| (ktype.id != DYNREF_TYPEID && ktype.id != DYNLIST_TYPEID))
				{
					return false;
				}
				bytecode += DYNREFTYPESIZE_MEIREI;
				bytecode += INDEXTYPE_MEIREI;
				ktype.id = ADDRESS_TYPEID;
				tystack.Tail() = ktype.id;
				break;
			case TYPEID_TOKEN:
				if(
					tystack.GetTail() == UNKNOWN_TYPEID || tystack.GetTail() == BOOL_TYPEID
					|| (ktype.id != DYNREF_TYPEID && ktype.id != DYNLIST_TYPEID))
				{
					return false;
				}
				bytecode += DYNREFTYPEID_MEIREI;
				bytecode += INTTYPE_MEIREI;
				ktype.id = INT_TYPEID;
				tystack.Tail() = ktype.id;
				break;
			case CAST_TOKEN:
				if(
					tystack.GetTail() == UNKNOWN_TYPEID || tystack.GetTail() == BOOL_TYPEID
					|| (ktype.id != DYNREF_TYPEID && ktype.id != DYNLIST_TYPEID))
				{
					return false;
				}
				typ.id = ktype.id;
				i++;
				if(iap[i] != SHOUKAKKOOPEN_TOKEN) return false;
				i++;
				tystack += UNKNOWN_TYPEID;
				do{
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
					i++;
					if(iap[i] == SHOUKAKKOCLOSE_TOKEN) return false;
				}while(iap[i] != COMMA_TOKEN);
				i++;
				if(iap[i] != SHOUKAKKOCLOSE_TOKEN) return false;
				if(
					tystack.GetTail() != UNKNOWN_TYPEID
					|| (ktype.id != REF_TYPEID && (typ.id == DYNREF_TYPEID || ktype.id != LIST_TYPEID)))
				{
					return false;
				}
				il = 0;
				if(IsSoushokuToken((TokenKind)ktype.kwsk.GetConst()[il])) il++;
				if(ktype.kwsk.GetConst()[il] != REF_TOKEN && ktype.kwsk.GetConst()[il] != LIST_TOKEN) return false;
				il++;
				if((typ.id = TokenTtoTypeId(ktype.kwsk.GetConst(), il)) == UNKNOWN_TYPEID) return false;
				bytecode += DYNREFGET_MEIREI;
				AtaiCode(bytecode, typ.id);
				tystack.SetLength(tystack.GetLength()-2);
				tystack.Tail() = BOOL_TYPEID;
				break;
			case WAIT_TOKEN:
				i++;
				if(iap[i] != SHOUKAKKOOPEN_TOKEN) return false;
				do{
					i++;
					tystack += UNKNOWN_TYPEID;
					do{
						if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
						i++;
					}while(iap[i] != SHOUKAKKOCLOSE_TOKEN && iap[i] != COMMA_TOKEN);
					if(tystack.GetTail() == UNKNOWN_TYPEID || tystack.GetTail() == BOOL_TYPEID) return false;
					if(ktype.id == REF_TYPEID || ktype.id == LIST_TYPEID || ktype.id == DYNREF_TYPEID || ktype.id == DYNLIST_TYPEID){
						bytecode += LOCK_MEIREI;
						bytecode += UNLOCK_MEIREI;
						bytecode += POP_MEIREI;
					}else if(ENDTHREAD_TYPEID <= ktype.id && ktype.id < 0){
						bytecode += SANSHOUTHREADWAIT_MEIREI;
					}else{
						return false;
					}
					tystack.SetLength(tystack.GetLength()-2);
				}while(iap[i] != SHOUKAKKOCLOSE_TOKEN);
				ktype.id = VOID_TYPEID;
				break;
			case CONSTEVAL_TOKEN:
				i++;
				if(iap[i] != SHOUKAKKOOPEN_TOKEN) return false;
				i++;
				tystack += UNKNOWN_TYPEID;
				do{
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
					i++;
					if(iap[i] == SHOUKAKKOCLOSE_TOKEN) return false;
				}while(iap[i] != COMMA_TOKEN);
				i++;
				if(iap[i] != COMMA_TOKEN) return false;
				i++;
				if(tystack.GetTail() != UNKNOWN_TYPEID) return false;
				typ = ktype;
				do{
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
					i++;
				}while(iap[i] != SHOUKAKKOCLOSE_TOKEN);
				if(tystack.GetTail() == UNKNOWN_TYPEID || tystack.GetTail() == BOOL_TYPEID) return false;
				type = REF_TOKEN;
				type += CONST_TOKEN;
				type += CHAR_TOKEN;
				if(!stree.YuruiKataCheck(ktype.kwsk, type, *pomc->filename, pomc->gyousuu)) return false;
				il = 0;
				if(IsSoushokuToken((TokenKind)typ.kwsk.GetConst()[il])) il++;
				indt = il;
				switch(TokenTtoTypeId(typ.kwsk.GetConst(), il)){
				case REF_TYPEID:
				case LIST_TYPEID:
				case DYNREF_TYPEID:
				case DYNLIST_TYPEID:
					return false;
				case METHOD_TYPEID:
					il++;
				case FUNC_TYPEID:
				case DELEGATE_TYPEID:
					type.Clear();
					if(!FuncToken(type, typ.kwsk.GetConst(), il)) return false;
					break;
				case UNKNOWN_TYPEID:
					return false;
				}
				bytecode += INDEXTYPE_MEIREI;
				ktype.id = ADDRESS_TYPEID;
				for(il--; il >= indt; il--){
					bytecode += PUSHTEISUU_MEIREI;
					AtaiCode(bytecode, typ.kwsk.GetConst()[il]);
				}
				bytecode += CONSTEVAL_MEIREI;
				AtaiCode(bytecode, &stree);
				tystack.SetLength(tystack.GetLength()-2);
				tystack.Tail() = BOOL_TYPEID;
				break;
			case CHUUKAKKOOPEN_TOKEN:
				i++;
				pst = stree.children.GetConst()[iap[i]];
				switch(pst->selftype){
				case NORMAL_BLOCK:
					if(!pomc->ncom.FuncTreeCompile(*pst, bytecode)) return false;
					break;
				case BRANCH_BLOCK:
					if(!pomc->brcom.FuncTreeCompile(*pst, bytecode)) return false;
					break;
				case LOOP_BLOCK:
					if(!pomc->lpcom.FuncTreeCompile(*pst, bytecode)) return false;
					break;
				case LOCK_BLOCK:
					i++;
					if(iap[i] != SHOUKAKKOOPEN_TOKEN) return false;
					indt = 0;
					do{
						i++;
						tystack += UNKNOWN_TYPEID;
						do{
							if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
							i++;
						}while(iap[i] != SHOUKAKKOCLOSE_TOKEN && iap[i] != COMMA_TOKEN);
						if(
							tystack.GetTail() == UNKNOWN_TYPEID || tystack.GetTail() == BOOL_TYPEID
							|| (ktype.id != REF_TYPEID && ktype.id != LIST_TYPEID && ktype.id != DYNREF_TYPEID && ktype.id != DYNLIST_TYPEID))
						{
							return false;
						}
						tystack.Pop();
						tystack.Tail() = ktype.id;
						bytecode += LOCK_MEIREI;
						bytecode += REFCNTUP_MEIREI;
						bytecode += INDEXTYPE_MEIREI;
						bytecode += PUSHTEISUU_MEIREI;
						bp1 = bytecode.GetLength();
						AtaiCode(bytecode, (intptr_t)0);
						bytecode += JUMP_MEIREI;
						bp2 = bytecode.GetLength();
						AtaiCode(bytecode, (intptr_t)0);
						//
						*(intptr_t *)(bytecode.GetConst()+bp1) = bytecode.GetLength();
						//
						if(ktype.id == DYNREF_TYPEID || ktype.id == DYNLIST_TYPEID){
							bytecode += DYNREFTYPE_MEIREI;
							bytecode += UNLOCK_MEIREI;
							bytecode += LOCALPUSH_MEIREI;
							bytecode += HENSUUKARASANSHOU_MEIREI;
							AtaiCode(bytecode, (intptr_t)0);
							typ = ktype;
							if(!RefCountDown(stree, bytecode, ktype, SANSHOUREFCNTDOWN_MEIREI, typ.kwsk, false)) return false;
							bytecode += HENSUUSTACKKAIHOU_MEIREI;
							AtaiCode(bytecode, Aliszof<DynamicRef>()+Aliszof<intptr_t>());
						}else{
							bytecode += REFTYPE_MEIREI;
							bytecode += UNLOCK_MEIREI;
							bytecode += LOCALPUSH_MEIREI;
							bytecode += HENSUUKARASANSHOU_MEIREI;
							AtaiCode(bytecode, (intptr_t)0);
							typ = ktype;
							if(!RefCountDown(stree, bytecode, ktype, SANSHOUREFCNTDOWN_MEIREI, typ.kwsk, false)) return false;
							bytecode += HENSUUSTACKKAIHOU_MEIREI;
							AtaiCode(bytecode, Aliszof<Reference>()+Aliszof<intptr_t>());
						}
						bytecode += RETURN_MEIREI;
						//
						*(intptr_t *)(bytecode.GetConst()+bp2) = bytecode.GetLength();
						ktype.id = ADDRESS_TYPEID;
						//
						indt++;
					}while(iap[i] != SHOUKAKKOCLOSE_TOKEN);
					pst->hikisuucount = (int)indt;
					if(!pomc->lkcom.FuncTreeCompile(*pst, bytecode)) return false;
					tystack.SetLength(tystack.GetLength()-indt);
					break;
				case SWITCH_BLOCK:
					i++;
					if(iap[i] != SHOUKAKKOOPEN_TOKEN) return false;
					i++;
					tystack += WILDCARD_TYPEID;
					do{
						if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
						i++;
					}while(iap[i] != SHOUKAKKOCLOSE_TOKEN);
					if(ktype.id == VOID_TYPEID) return false;
					ktype.id = tystack.Pop();
					tystack.Tail() = ktype.id;
					if(ktype.id < 0 || (ktype.id <= ENDCLASS_TYPEID && stree.stat->funclist.Get(ktype.id)->selftype != ENUM_BLOCK)) return false;
					pst->extswitch->swtype = ktype.id;
					if(!pomc->swcom.FuncTreeCompile(*pst, bytecode)) return false;
					tystack.Pop();
					break;
				case DELEGATE_BLOCK:
					bytecode += JUMP_MEIREI;
					bp1 = bytecode.GetLength();
					AtaiCode(bytecode, (intptr_t)0);
					if(!pomc->dgcom.FuncTreeCompile(*pst, bytecode)) return false;
					*(intptr_t *)(bytecode.GetConst()+bp1) = bytecode.GetLength();
					ktype.pstre = &stree;
					break;
				default:
					return false;
				}
				i++;
				if(iap[i] != CHUUKAKKOCLOSE_TOKEN) return false;
				break;
			case SHOUKAKKOOPEN_TOKEN:
				i++;
				if(iap[i] < 0){
					pst = stree.stat->funclist.Get((int32_t)~iap[i]);
					i++;
					if(Kansuu(stree, bytecode, iap, i, tystack, ktype, *pst, false) < 0) return false;
					if(stree.frp != stree.root && !stree.frp->touka) bytecode += MEMBERPUSH_MEIREI;
					bytecode += CALL_MEIREI;
					if((ktype.id = TokenTtoTypeId(pst->rettype.GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
					ktype.kwsk = pst->rettype;
					tystack += ktype.id;
					if(pst->byteadress == 0){
						pst->backpatch += bytecode.GetLength();
					}
					AtaiCode(bytecode, pst->byteadress);
					if(stree.frp != stree.root && !stree.frp->touka) bytecode += MEMBERPOP_MEIREI;
				}else{
					if((typ.id = TokenTtoTypeId(iap, i)) == UNKNOWN_TYPEID) return false;
					if(!MeijiCast(stree, bytecode, tystack, typ.id, ktype.id)) return false;
					ktype = typ;
					tystack.Tail() = ktype.id;
				}
				break;
			case KAKKOCOLOOPEN_TOKEN:
				i++;
				tmpb = ktype.id == DELEGATE_TYPEID;
				if((tmpip = KansuuPointer(stree, bytecode, iap, i, tystack, ktype, stree.frp, waketa)) < 0) return false;
				if(stree.glbconstant && waketa.GetConst()[0].GetConst()[0] != CONST_TOKEN) return false;
				if(stree.frp->rettype.GetConst()[0] == PROPERTY_TOKEN && !IsReadOnlyToken((TokenKind)waketa.GetConst()[0].GetConst()[0])) return false;
				if(tmpip > 0){
					bytecode += INDEXTYPE_MEIREI;
					bytecode += LOCALPUSH_MEIREI;
					if(tmpb){
						bytecode += PUSHHENSUU_MEIREI;
						AtaiCode(bytecode, tmpip+Aliszof<intptr_t>());
					}
					bytecode += PUSHHENSUU_MEIREI;
					AtaiCode(bytecode, tmpip);
					if(tmpb){
						bytecode += LOCALPOP_MEIREI;
					}
				}
				bytecode += STACKCALL_MEIREI;
				if((ktype.id = TokenTtoTypeId(waketa.GetConst()[0].GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
				ktype.kwsk = waketa.GetConst()[0];
				tystack.Tail() = ktype.id;
				if(!tmpb && tmpip > 0){
					if(ktype.id != VOID_TYPEID){
						bytecode += TypeIdtoMeireiT(stree, ktype.id);
						bytecode += POPTOHENSUU_MEIREI;
						AtaiCode(bytecode, tmpip+Aliszof<intptr_t>()-Aligner(stree.GetTypeSize(waketa.GetConst()[0])));
					}
					bytecode += HENSUUSTACKKAIHOU_MEIREI;
					AtaiCode(bytecode, Aliszof<intptr_t>()-Aligner(stree.GetTypeSize(waketa.GetConst()[0])));
				}else{
					if(ktype.id != VOID_TYPEID){
						bytecode += TypeIdtoMeireiT(stree, ktype.id);
					}
				}
				break;
			case LIB_TOKEN:
			case TILDE_TOKEN:
				tmptk = (TokenKind)iap[i];
				i++;
				if(tmptk == LIB_TOKEN) i += 2;
				if(iap[i] == SHOUKAKKOOPEN_TOKEN){
					i++;
					if(iap[i] < 0){
						pst = stree.stat->funclist.Get((int32_t)~iap[i]);
						i++;
						if(stree.glbconstant && pst->rettype.GetConst()[0] != CONST_TOKEN) return false;
						if((typ.id = TokenTtoTypeId(pst->rettype.GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
						typ.kwsk = pst->rettype;
						tystack += typ.id;
						if(stree.frp != stree.root && typ.id == VOID_TYPEID && IsVoidHikinashiRet(stree, iap, i, *pst)){
							if(TokenTtoTypeId(stree.frp->rettype.GetConst(), il=0) != VOID_TYPEID) return false;
							tystack.Pop();
							retcall = pst;
						}else{
							if(Kansuu(stree, bytecode, iap, i, tystack, ktype, *pst, false) < 0) return false;
							bytecode += CALL_MEIREI;
							if(pst->byteadress == 0){
								pst->backpatch += bytecode.GetLength();
							}
							AtaiCode(bytecode, pst->byteadress);
						}
						ktype = typ;
					}else{
						return false;
					}
				}else{
					if(iap[i] >= 0) return MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall);
					indt = ~iap[i];
					pst = (tmptk == LIB_TOKEN ? (SrcTree *)iap[i-2] : stree.root);
					cnstf = stree.glbconstant || (tmptk == LIB_TOKEN ? IsReadOnlyToken((TokenKind)pst->funchensuu.GetConst()[indt].type.GetConst()[0]) : false);
					ig = pst->sonota.golbalgeta;
					while(true){
						ig += pst->GetMemberOffset(indt);
						if(iap[i+1] != POINT_TOKEN) break;
						i += 2;
						if(iap[i] == SHOUKAKKOOPEN_TOKEN){
							i++;
							if(iap[i] < 0){
								pst = stree.stat->funclist.Get((int32_t)~iap[i]);
								i++;
								if(!pst->kakutei && (pst = pst->GetIjousaki(ig, stree, iap+i)) == NULL) return false;
								if(stree.frp->rettype.GetConst()[0] == CONST_TOKEN && pst->rettype.GetConst()[0] != CONST_TOKEN) return false;
								if(cnstf && !IsReadOnlyToken((TokenKind)pst->rettype.GetConst()[0])) return false;
								if(Kansuu(stree, bytecode, iap, i, tystack, ktype, *pst, false) < 0) return false;
								bytecode += MEMBERPUSHGLOBAL_MEIREI;
								AtaiCode(bytecode, ig);
								bytecode += CALL_MEIREI;
								if(pst->byteadress == 0){
									pst->backpatch += bytecode.GetLength();
								}
								AtaiCode(bytecode, pst->byteadress);
								if((ktype.id = TokenTtoTypeId(pst->rettype.GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
								ktype.kwsk = pst->rettype;
								tystack += ktype.id;
								bytecode += MEMBERPOP_MEIREI;
							}else{
								return false;
							}
							break;
						}
						if(iap[i] >= 0) return false;
						il = 0;
						if(IsSoushokuToken((TokenKind)pst->funchensuu.GetConst()[indt].type.GetConst()[il])) il++;
						if(pst->funchensuu.GetConst()[indt].type.GetConst()[il] != CLASS_TOKEN
							&& pst->funchensuu.GetConst()[indt].type.GetConst()[il] != ENUM_TOKEN) return false;
						il++;
						tyki = (int32_t)pst->funchensuu.GetConst()[indt].type.GetConst()[il];
						pst = stree.stat->funclist.Get(tyki);
						indt = ~iap[i];
						cnstf = cnstf || IsReadOnlyToken((TokenKind)pst->funchensuu.GetConst()[indt].type.GetConst()[0]);
					}
					if(iap[i] >= 0) break;
					pty = &pst->funchensuu.GetConst()[indt].type;
					il = 0;
					if(IsSoushokuToken((TokenKind)pty->GetConst()[il])) il++;
					if(iap[i+1] == VOID_TOKEN){
						i++;
						if(pty->GetConst()[il] == THREAD_TOKEN){
							if(ktype.id != ADDRESS_TYPEID){
								bytecode += INDEXTYPE_MEIREI;
								ktype.id = ADDRESS_TYPEID;
							}
							bytecode += GLOBALSETNULL_MEIREI;
							AtaiCode(bytecode, ig);
						}else if(pty->GetConst()[il] == REF_TOKEN || pty->GetConst()[il] == LIST_TOKEN){
							if(pty->GetConst()[il+1] == NULL_TOKEN){
								if(ktype.id != DYNREF_TYPEID && ktype.id != DYNLIST_TYPEID){
									bytecode += DYNREFTYPE_MEIREI;
									ktype.id = VOID_TYPEID;
								}
								bytecode += GLOBALSETNULL_MEIREI;
								AtaiCode(bytecode, ig);
							}else if(pty->GetConst()[il+1] != OTHERKIND_TOKEN){
								if(ktype.id != REF_TYPEID && ktype.id != LIST_TYPEID){
									bytecode += REFTYPE_MEIREI;
									ktype.id = VOID_TYPEID;
								}
								bytecode += GLOBALSETNULL_MEIREI;
								AtaiCode(bytecode, ig);
							}
						}else{
							return false;
						}
						i++;//Z~R΂
						if(iap[i] != SEMICOLON_TOKEN) return false;
						break;
					}
					if(iap[i+1] == DELETE_TOKEN || iap[i+1] == NEW_TOKEN){
						if(cnstf) return false;
						i++;
						if(pty->GetConst()[il] == THREAD_TOKEN){
							bytecode += GLOBALKARASANSHOU_MEIREI;
							AtaiCode(bytecode, ig);
							bytecode += SANSHOUTHREADWAIT_MEIREI;
						}else if(pty->GetConst()[il] == REF_TOKEN || pty->GetConst()[il] == LIST_TOKEN){
							if(pty->GetConst()[il+1] != OTHERKIND_TOKEN){
								bytecode += GLOBALKARASANSHOU_MEIREI;
								AtaiCode(bytecode, ig);
								if(iap[i] != NEW_TOKEN || iap[i+1] != DAIKAKKOOPEN_TOKEN){
									if(!RefCountDown(stree, bytecode, ktype, SANSHOUREFCNTDOWN_MEIREI, *pty, iap[i]==DELETE_TOKEN)) return false;
								}
							}
						}else{
							return false;
						}
						if(iap[i] == NEW_TOKEN){
							i++;
							if(pty->GetConst()[il+1] == CONST_TOKEN || pty->GetConst()[il+1] == OTHERKIND_TOKEN) return false;
							if(iap[i] == SHOUKAKKOOPEN_TOKEN){
								i++;
								if(pty->GetConst()[il] == REF_TOKEN){
									if(!RefNew(stree, bytecode, iap, i, tystack, ktype, SANSHOUREFNEW_MEIREI, *pty)) return false;
								}else{
									if(!RefNew(stree, bytecode, iap, i, tystack, ktype, SANSHOULISTNEW_MEIREI, *pty)) return false;
								}
							}else if(iap[i] == DAIKAKKOOPEN_TOKEN){
								i++;
								if(pty->GetConst()[il] != LIST_TOKEN) return false;
								if(!RefIdxNew(stree, bytecode, iap, i, tystack, ktype, *pty)) return false;
								il++;
								type.Clear();
								type.AddTail(pty->GetConst()+il, pty->GetLength()-il);
								if(!Sanshou(stree, bytecode, iap, i, tystack, ktype, type, false)) return false;
							}else{
								return false;
							}
						}else{
							i++;
							if(iap[i] != SEMICOLON_TOKEN) return false;
						}
						break;
					}
					if(iap[i+1] == SHOUKAKKOOPEN_TOKEN && iap[i+2] == TENTEN_TOKEN){
						if(cnstf) return false;
						i += 3;
						bytecode += GLOBALKARASANSHOU_MEIREI;
						AtaiCode(bytecode, ig);
						if(pty->GetConst()[il] != THREAD_TOKEN) return false;
						il++;
						if(Kansuu(stree, bytecode, iap, i, tystack, ktype, *stree.stat->funclist.Get((int32_t)pty->GetConst()[il]), true) < 0) return false;
						bytecode += SANSHOUTHREADRUN_MEIREI;
						ktype.id = VOID_TYPEID;
						break;
					}
					tmpb = pty->GetConst()[il] == CLASS_TOKEN;
					typ.pstre = pst;
					typ.kwsk = *pty;
					if((typ.id = TokenTtoTypeId(pty->GetConst(), il=0)) == UNKNOWN_TYPEID){
						if(pty->GetConst()[0] == PLUGIN_TOKEN && iap[i+1] == KAKKOCOLOOPEN_TOKEN){
							i += 2;
							ktype = typ;
							if((tmpip = PluginKansuu(stree, bytecode, iap, i, tystack, ktype, pst, waketa)) < 0) return false;
							if(stree.frp->rettype.GetConst()[0] == CONST_TOKEN && waketa.GetConst()[0].GetConst()[0] != CONST_TOKEN) return false;
							if(cnstf && !IsReadOnlyToken((TokenKind)waketa.GetConst()[0].GetConst()[0])) return false;
							if((ktype.id = TokenTtoTypeId(waketa.GetConst()[0].GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
							ktype.kwsk = waketa.GetConst()[0];
							bytecode += PLUGINCALL_MEIREI;
							bytecode += (uint8_t)(stree.TypeIdtoSize(ktype.id) + (ktype.id == FLOAT_TYPEID || ktype.id == DOUBLE_TYPEID ? 128 : 0));
							AtaiCode(bytecode, pty->GetTail());
							if(ktype.id != VOID_TYPEID){
								bytecode += TypeIdtoMeireiT(stree, ktype.id);
							}
							tystack += ktype.id;
							break;
						}else{
							return false;
						}
					}
					if(iap[i+1] == NULL_TOKEN && iap[i+2] == TENDAINYUU_TOKEN){
						if(cnstf) return false;
						i += 2;
						bytecode += GLOBALKARASANSHOU_MEIREI;
						AtaiCode(bytecode, ig);
						if(!Append(stree, bytecode, iap, i, tystack, ktype, typ)) return false;
						break;
					}
					if(iap[i+1] == NULL_TOKEN && (DAINYUU_TOKEN <= iap[i+2] && iap[i+2] <= TENDAINYUU_TOKEN)){
						if(cnstf) return false;
						i += 2;
						if(!Dainyuu(stree, bytecode, iap, i, tystack, ktype, typ, ig, PUSHGLOBALHENSUU_MEIREI, tmpb , false)) return false;
						break;
					}
					if(
						(iap[i+1] == NULL_TOKEN && (iap[i+2] == PLSPLS_TOKEN || iap[i+2] == MNSMNS_TOKEN))
						|| iap[i+1] == PLSPLS_TOKEN || iap[i+1] == MNSMNS_TOKEN)
					{
						if(cnstf) return false;
						i++;
						bytecode += GLOBALKARASANSHOU_MEIREI;
						AtaiCode(bytecode, ig);
						if(!IncDec(stree, bytecode, iap, i, tystack, ktype, typ)) return false;
						break;
					}
					ktype.pstre = typ.pstre;
					ktype.kwsk = typ.kwsk;
					if(typ.id != ktype.id){
						if(typ.id == VOID_TYPEID) return false;
						ktype.id = typ.id;
						if(!tmpb && !(ENDTHREAD_TYPEID <= ktype.id && ktype.id < 0)){
							if((deco = TypeIdtoMeireiT(stree, ktype.id)) == NOP_MEIREI) return false;
							bytecode += deco;
						}
					}
					if(cnstf && ENDTHREAD_TYPEID <= ktype.id && ktype.id < 0) return false;
					if(ktype.id != VOID_TYPEID){
						if(pty->GetConst()[0] == CONST_TOKEN){
							if(ktype.id == FUNC_TYPEID){
								bytecode += PUSHTEISUU_MEIREI;
								if(!stree.stat->funclist.Get((int32_t)pty->GetTail())->kakutei) return false;
								if(stree.stat->funclist.Get((int32_t)pty->GetTail())->byteadress == 0){
									stree.stat->funclist.Get((int32_t)pty->GetTail())->backpatch += bytecode.GetLength();
								}
								AtaiCode(bytecode, stree.stat->funclist.Get((int32_t)pty->GetTail())->byteadress);
							}else if(ktype.id == REF_TYPEID){
								bytecode += PUSHGLOBALHENSUU_MEIREI;
								AtaiCode(bytecode, pty->GetTail());
							}else{
								if((deco = TypeIdtoMeireiT(stree, ktype.id)) == NOP_MEIREI) return false;
								bytecode += PUSHTEISUU_MEIREI;
								TypeAtai(bytecode, deco, pty->GetConst()+il);
							}
						}else{
							if(tmpb || (ENDTHREAD_TYPEID <= ktype.id && ktype.id < 0)){
								bytecode += GLOBALKARASANSHOU_MEIREI;
								AtaiCode(bytecode, ig);
							}else if(ktype.id == DELEGATE_TYPEID){
								bytecode += GLOBALKARASANSHOU_MEIREI;
								AtaiCode(bytecode, ig);
								bytecode += LOADSANSHOUKARASANSHOUSAKI_MEIREI;
								AtaiCode(bytecode, stree.TypeIdtoSize(ktype.id));
							}else{
								bytecode += PUSHGLOBALHENSUU_MEIREI;
								AtaiCode(bytecode, ig);
							}
						}
						if(ktype.id == METHOD_TYPEID && iap[i+1] == KAKKOCOLOOPEN_TOKEN){
							i += 2;
							ig -= pst->GetMemberOffset(indt);
							if((tmpip = KansuuPointer(stree, bytecode, iap, i, tystack, ktype, pst, waketa)) < 0) return false;
							if(stree.frp->rettype.GetConst()[0] == CONST_TOKEN && waketa.GetConst()[0].GetConst()[0] != CONST_TOKEN) return false;
							if(cnstf && !IsReadOnlyToken((TokenKind)waketa.GetConst()[0].GetConst()[0])) return false;
							bytecode += MEMBERPUSHGLOBAL_MEIREI;
							AtaiCode(bytecode, ig);
							if(tmpip > 0){
								bytecode += INDEXTYPE_MEIREI;
								bytecode += LOCALPUSH_MEIREI;
								bytecode += PUSHHENSUU_MEIREI;
								AtaiCode(bytecode, tmpip);
							}
							bytecode += STACKCALL_MEIREI;
							if((ktype.id = TokenTtoTypeId(waketa.GetConst()[0].GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
							ktype.kwsk = waketa.GetConst()[0];
							if(tmpip > 0){
								if(ktype.id != VOID_TYPEID){
									bytecode += POPTOHENSUU_MEIREI;
									AtaiCode(bytecode, tmpip+Aliszof<intptr_t>()-Aligner(stree.GetTypeSize(waketa.GetConst()[0])));
								}
								bytecode += HENSUUSTACKKAIHOU_MEIREI;
								AtaiCode(bytecode, Aliszof<intptr_t>()-Aligner(stree.GetTypeSize(waketa.GetConst()[0])));
							}
							bytecode += MEMBERPOP_MEIREI;
						}
					}
					tystack += ktype.id;
				}
				break;
			case MBR_TOKEN:
				i++;
				if(iap[i] == SHOUKAKKOOPEN_TOKEN){
					i++;
					if(iap[i] < 0){
						pst = stree.stat->funclist.Get(~iap[i]);
						i++;
						if(stree.glbconstant && pst->rettype.GetConst()[0] != CONST_TOKEN) return false;
						if(stree.frp->rettype.GetConst()[0] == PROPERTY_TOKEN && !IsReadOnlyToken((TokenKind)pst->rettype.GetConst()[0])) return false;
						if((typ.id = TokenTtoTypeId(pst->rettype.GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
						typ.kwsk = pst->rettype;
						tystack += typ.id;
						if(stree.frp != stree.root && typ.id == VOID_TYPEID && IsVoidHikinashiRet(stree, iap, i, *pst)){
							if(TokenTtoTypeId(stree.frp->rettype.GetConst(), il=0) != VOID_TYPEID) return false;
							tystack.Pop();
							retcall = pst;
						}else{
							if(Kansuu(stree, bytecode, iap, i, tystack, ktype, *pst, false) < 0) return false;
							bytecode += CALL_MEIREI;
							if(pst->byteadress == 0){
								pst->backpatch += bytecode.GetLength();
							}
							AtaiCode(bytecode, pst->byteadress);
						}
						ktype = typ;
					}else{
						return false;
					}
				}else{
					if(iap[i] >= 0) return MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall);
					indt = ~iap[i];
					pst = stree.frp->parent;
					cnstf = IsReadOnlyToken((TokenKind)stree.frp->rettype.GetConst()[0]);
					if(pst->tpcount <= indt && indt < pst->tpcount+pst->hikisuucount
						&& pst->funchensuu.GetConst()[indt].type.GetTail() == ~DAINYUU_TOKEN)
					{
						ig = pst->GetMemberOffset(indt);
						if(ktype.id != ADDRESS_TYPEID) bytecode += INDEXTYPE_MEIREI;
						ktype.id = ADDRESS_TYPEID;
						bytecode += PUSHMEMBERHENSUU_MEIREI;
						AtaiCode(bytecode, ig);
						type = pst->funchensuu.GetConst()[indt].type;
						type.Pop();
						if(!Sanshou(stree, bytecode, iap, i, tystack, ktype, type, cnstf)) return false;
					}else{
						cnstf = cnstf || indt < pst->tpcount+pst->hikisuucount;
						ig = 0;
						while(true){
							ig += pst->GetMemberOffset(indt);
							if(iap[i+1] != POINT_TOKEN) break;
							i += 2;
							if(iap[i] == SHOUKAKKOOPEN_TOKEN){
								i++;
								if(iap[i] < 0){
									pst = stree.stat->funclist.Get(~iap[i]);
									i++;
									if(!pst->kakutei && (pst = pst->GetIjousaki(ig, stree, iap+i)) == NULL) return false;
									if(stree.frp->rettype.GetConst()[0] == CONST_TOKEN && pst->rettype.GetConst()[0] != CONST_TOKEN) return false;
									if(cnstf && !IsReadOnlyToken((TokenKind)pst->rettype.GetConst()[0])) return false;
									if(Kansuu(stree, bytecode, iap, i, tystack, ktype, *pst, false) < 0) return false;
									bytecode += MEMBERPUSHMEMBER_MEIREI;
									AtaiCode(bytecode, ig);
									bytecode += CALL_MEIREI;
									if((ktype.id = TokenTtoTypeId(pst->rettype.GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
									ktype.kwsk = pst->rettype;
									tystack += ktype.id;
									if(pst->byteadress == 0){
										pst->backpatch += bytecode.GetLength();
									}
									AtaiCode(bytecode, pst->byteadress);
									bytecode += MEMBERPOP_MEIREI;
								}else{
									return false;
								}
								break;
							}
							if(iap[i] >= 0) return false;
							il = 0;
							if(IsSoushokuToken((TokenKind)pst->funchensuu.GetConst()[indt].type.GetConst()[il])) il++;
							if(pst->funchensuu.GetConst()[indt].type.GetConst()[il] != CLASS_TOKEN
								&& pst->funchensuu.GetConst()[indt].type.GetConst()[il] != ENUM_TOKEN) return false;
							il++;
							tyki = (int32_t)pst->funchensuu.GetConst()[indt].type.GetConst()[il];
							pst = stree.stat->funclist.Get(tyki);
							indt = ~iap[i];
							cnstf = cnstf || IsReadOnlyToken((TokenKind)pst->funchensuu.GetConst()[indt].type.GetConst()[0]);
						}
						if(iap[i] >= 0) break;
						pty = &pst->funchensuu.GetConst()[indt].type;
						il = 0;
						if(IsSoushokuToken((TokenKind)pty->GetConst()[il])) il++;
						if(iap[i+1] == VOID_TOKEN){
							i++;
							if(pty->GetConst()[il] == THREAD_TOKEN){
								if(ktype.id != ADDRESS_TYPEID){
									bytecode += INDEXTYPE_MEIREI;
									ktype.id = ADDRESS_TYPEID;
								}
								bytecode += MEMBERSETNULL_MEIREI;
								AtaiCode(bytecode, ig);
							}else if(pty->GetConst()[il] == REF_TOKEN || pty->GetConst()[il] == LIST_TOKEN){
								if(pty->GetConst()[il+1] == NULL_TOKEN){
									if(ktype.id != DYNREF_TYPEID && ktype.id != DYNLIST_TYPEID){
										bytecode += DYNREFTYPE_MEIREI;
										ktype.id = VOID_TYPEID;
									}
									bytecode += MEMBERSETNULL_MEIREI;
									AtaiCode(bytecode, ig);
								}else if(pty->GetConst()[il+1] != OTHERKIND_TOKEN){
									if(ktype.id != REF_TYPEID && ktype.id != LIST_TYPEID){
										bytecode += REFTYPE_MEIREI;
										ktype.id = VOID_TYPEID;
									}
									bytecode += MEMBERSETNULL_MEIREI;
									AtaiCode(bytecode, ig);
								}
							}else{
								return false;
							}
							i++;//Z~R΂
							if(iap[i] != SEMICOLON_TOKEN) return false;
							break;
						}
						if(iap[i+1] == DELETE_TOKEN || iap[i+1] == NEW_TOKEN){
							if(cnstf) return false;
							i++;
							if(pty->GetConst()[il] == THREAD_TOKEN){
								bytecode += MEMBERKARASANSHOU_MEIREI;
								AtaiCode(bytecode, ig);
								bytecode += SANSHOUTHREADWAIT_MEIREI;
							}else if(pty->GetConst()[il] == REF_TOKEN || pty->GetConst()[il] == LIST_TOKEN){
								if(pty->GetConst()[il+1] != OTHERKIND_TOKEN){
									bytecode += MEMBERKARASANSHOU_MEIREI;
									AtaiCode(bytecode, ig);
									if(iap[i] != NEW_TOKEN || iap[i+1] != DAIKAKKOOPEN_TOKEN){
										if(!RefCountDown(stree, bytecode, ktype, SANSHOUREFCNTDOWN_MEIREI, *pty, iap[i]==DELETE_TOKEN)) return false;
									}
								}
							}else{
								return false;
							}
							if(iap[i] == NEW_TOKEN){
								i++;
								if(pty->GetConst()[il+1] == CONST_TOKEN || pty->GetConst()[il+1] == OTHERKIND_TOKEN) return false;
								if(iap[i] == SHOUKAKKOOPEN_TOKEN){
									i++;
									if(pty->GetConst()[il] == REF_TOKEN){
										if(!RefNew(stree, bytecode, iap, i, tystack, ktype, SANSHOUREFNEW_MEIREI, *pty)) return false;
									}else{
										if(!RefNew(stree, bytecode, iap, i, tystack, ktype, SANSHOULISTNEW_MEIREI, *pty)) return false;
									}
								}else if(iap[i] == DAIKAKKOOPEN_TOKEN){
									i++;
									if(pty->GetConst()[il] != LIST_TOKEN) return false;
									if(!RefIdxNew(stree, bytecode, iap, i, tystack, ktype, *pty)) return false;
									il++;
									type.Clear();
									type.AddTail(pty->GetConst()+il, pty->GetLength()-il);
									if(!Sanshou(stree, bytecode, iap, i, tystack, ktype, type, false)) return false;
								}else{
									return false;
								}
							}else{
								i++;
								if(iap[i] != SEMICOLON_TOKEN) return false;
							}
							break;
						}
						if(iap[i+1] == SHOUKAKKOOPEN_TOKEN && iap[i+2] == TENTEN_TOKEN){
							if(cnstf) return false;
							i += 3;
							bytecode += MEMBERKARASANSHOU_MEIREI;
							AtaiCode(bytecode, ig);
							if(pty->GetConst()[il] != THREAD_TOKEN) return false;
							il++;
							if(Kansuu(stree, bytecode, iap, i, tystack, ktype, *stree.stat->funclist.Get((int32_t)pty->GetConst()[il]), true) < 0) return false;
							bytecode += SANSHOUTHREADRUN_MEIREI;
							ktype.id = VOID_TYPEID;
							break;
						}
						tmpb = pty->GetConst()[il] == CLASS_TOKEN;
						typ.pstre = pst;
						typ.kwsk = *pty;
						if((typ.id = TokenTtoTypeId(pty->GetConst(), il=0)) == UNKNOWN_TYPEID){
							if(pty->GetConst()[0] == PLUGIN_TOKEN && iap[i+1] == KAKKOCOLOOPEN_TOKEN){
								i += 2;
								ktype = typ;
								if((tmpip = PluginKansuu(stree, bytecode, iap, i, tystack, ktype, pst, waketa)) < 0) return false;
								if(stree.frp->rettype.GetConst()[0] == CONST_TOKEN && waketa.GetConst()[0].GetConst()[0] != CONST_TOKEN) return false;
								if(cnstf && !IsReadOnlyToken((TokenKind)waketa.GetConst()[0].GetConst()[0])) return false;
								if((ktype.id = TokenTtoTypeId(waketa.GetConst()[0].GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
								ktype.kwsk = waketa.GetConst()[0];
								bytecode += PLUGINCALL_MEIREI;
								bytecode += (uint8_t)(stree.TypeIdtoSize(ktype.id) + (ktype.id == FLOAT_TYPEID || ktype.id == DOUBLE_TYPEID ? 128 : 0));
								AtaiCode(bytecode, pty->GetTail());
								if(ktype.id != VOID_TYPEID){
									bytecode += TypeIdtoMeireiT(stree, ktype.id);
								}
								tystack += ktype.id;
								break;
							}else{
								return false;
							}
						}
						if(iap[i+1] == NULL_TOKEN && iap[i+2] == TENDAINYUU_TOKEN){
							if(cnstf) return false;
							i += 2;
							bytecode += MEMBERKARASANSHOU_MEIREI;
							AtaiCode(bytecode, ig);
							if(!Append(stree, bytecode, iap, i, tystack, ktype, typ)) return false;
							break;
						}
						if(iap[i+1] == NULL_TOKEN && (DAINYUU_TOKEN <= iap[i+2] && iap[i+2] <= TENDAINYUU_TOKEN)){
							if(cnstf) return false;
							i += 2;
							if(!Dainyuu(stree, bytecode, iap, i, tystack, ktype, typ, ig, PUSHMEMBERHENSUU_MEIREI, tmpb, false)) return false;
							break;
						}
						if(
							(iap[i+1] == NULL_TOKEN && (iap[i+2] == PLSPLS_TOKEN || iap[i+2] == MNSMNS_TOKEN))
							|| iap[i+1] == PLSPLS_TOKEN || iap[i+1] == MNSMNS_TOKEN)
						{
							if(cnstf) return false;
							i++;
							bytecode += MEMBERKARASANSHOU_MEIREI;
							AtaiCode(bytecode, ig);
							if(!IncDec(stree, bytecode, iap, i, tystack, ktype, typ)) return false;
							break;
						}
						ktype.pstre = typ.pstre;
						ktype.kwsk = typ.kwsk;
						if(typ.id != ktype.id){
							if(typ.id == VOID_TYPEID) return false;
							ktype.id = typ.id;
							if(!tmpb && !(ENDTHREAD_TYPEID <= ktype.id && ktype.id < 0)){
								if((deco = TypeIdtoMeireiT(stree, ktype.id)) == NOP_MEIREI) return false;
								bytecode += deco;
							}
						}
						if(cnstf && ENDTHREAD_TYPEID <= ktype.id && ktype.id < 0) return false;
						if(ktype.id != VOID_TYPEID){
							if(pty->GetConst()[0] == CONST_TOKEN){
								bytecode += PUSHTEISUU_MEIREI;
								if(ktype.id == FUNC_TYPEID){
									if(!stree.stat->funclist.Get((int32_t)pty->GetTail())->kakutei) return false;
									if(stree.stat->funclist.Get((int32_t)pty->GetTail())->byteadress == 0){
										stree.stat->funclist.Get((int32_t)pty->GetTail())->backpatch += bytecode.GetLength();
									}
									AtaiCode(bytecode, stree.stat->funclist.Get((int32_t)pty->GetTail())->byteadress);
								}else if(ktype.id == REF_TYPEID){
									bytecode += PUSHGLOBALHENSUU_MEIREI;
									AtaiCode(bytecode, pty->GetTail());
								}else{
									if((deco = TypeIdtoMeireiT(stree, ktype.id)) == NOP_MEIREI) return false;
									TypeAtai(bytecode, deco, pty->GetConst()+il);
								}
							}else{
								if(tmpb || (ENDTHREAD_TYPEID <= ktype.id && ktype.id < 0)){
									bytecode += MEMBERKARASANSHOU_MEIREI;
									AtaiCode(bytecode, ig);
								}else if(ktype.id == DELEGATE_TYPEID){
									bytecode += MEMBERKARASANSHOU_MEIREI;
									AtaiCode(bytecode, ig);
									bytecode += LOADSANSHOUKARASANSHOUSAKI_MEIREI;
									AtaiCode(bytecode, stree.TypeIdtoSize(ktype.id));
								}else{
									bytecode += PUSHMEMBERHENSUU_MEIREI;
									AtaiCode(bytecode, ig);
								}
							}
							if(ktype.id == METHOD_TYPEID && iap[i+1] == KAKKOCOLOOPEN_TOKEN){
								i += 2;
								ig -= pst->GetMemberOffset(indt);
								if((tmpip = KansuuPointer(stree, bytecode, iap, i, tystack, ktype, pst, waketa)) < 0) return false;
								if(stree.frp->rettype.GetConst()[0] == CONST_TOKEN && waketa.GetConst()[0].GetConst()[0] != CONST_TOKEN) return false;
								if(cnstf && !IsReadOnlyToken((TokenKind)waketa.GetConst()[0].GetConst()[0])) return false;
								bytecode += MEMBERPUSHMEMBER_MEIREI;
								AtaiCode(bytecode, ig);
								if(tmpip > 0){
									bytecode += INDEXTYPE_MEIREI;
									bytecode += LOCALPUSH_MEIREI;
									bytecode += PUSHHENSUU_MEIREI;
									AtaiCode(bytecode, tmpip);
								}
								bytecode += STACKCALL_MEIREI;
								if((ktype.id = TokenTtoTypeId(waketa.GetConst()[0].GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
								ktype.kwsk = waketa.GetConst()[0];
								if(tmpip > 0){
									if(ktype.id != VOID_TYPEID){
										bytecode += POPTOHENSUU_MEIREI;
										AtaiCode(bytecode, tmpip+Aliszof<intptr_t>()-Aligner(stree.GetTypeSize(waketa.GetConst()[0])));
									}
									bytecode += HENSUUSTACKKAIHOU_MEIREI;
									AtaiCode(bytecode, Aliszof<intptr_t>()-Aligner(stree.GetTypeSize(waketa.GetConst()[0])));
								}
								bytecode += MEMBERPOP_MEIREI;
							}
						}
						tystack += ktype.id;
					}
				}
				break;
			default:
				if(iap[i] >= 0) return false;
				indt = ~iap[i];
				pst = stree.frp;
				cnstf = false;
				if((indt < pst->tpcount+pst->hikisuucount || (stree.crp->tpcount <= indt && indt < stree.crp->tpcount+stree.crp->hikisuucount))
					&& pst->funchensuu.GetConst()[indt].type.GetTail() == ~DAINYUU_TOKEN)
				{
					ig = stree.crp->GetMemberOffset(indt);
					if(ktype.id != ADDRESS_TYPEID) bytecode += INDEXTYPE_MEIREI;
					ktype.id = ADDRESS_TYPEID;
					bytecode += PUSHHENSUU_MEIREI;
					AtaiCode(bytecode, ig);
					type = pst->funchensuu.GetConst()[indt].type;
					type.Pop();
					if(!Sanshou(stree, bytecode, iap, i, tystack, ktype, type, cnstf)) return false;
				}else{
					cnstf = cnstf || indt < pst->tpcount+pst->hikisuucount || (stree.crp->tpcount <= indt && indt < stree.crp->tpcount+stree.crp->hikisuucount);
					indt = ~iap[i];
					ig = stree.crp->GetMemberOffset(indt);
					while(true){
						if(iap[i+1] != POINT_TOKEN) break;
						i += 2;
						if(iap[i] == SHOUKAKKOOPEN_TOKEN){
							i++;
							if(iap[i] < 0){
								pst = stree.stat->funclist.Get(~iap[i]);
								i++;
								if(!pst->kakutei && (pst = pst->GetIjousaki(ig, stree, iap+i)) == NULL) return false;
								if(stree.frp->rettype.GetConst()[0] == CONST_TOKEN && pst->rettype.GetConst()[0] != CONST_TOKEN) return false;
								if(cnstf && !IsReadOnlyToken((TokenKind)pst->rettype.GetConst()[0])) return false;
								if(Kansuu(stree, bytecode, iap, i, tystack, ktype, *pst, false) < 0) return false;
								bytecode += MEMBERPUSHHENSUU_MEIREI;
								AtaiCode(bytecode, ig);
								bytecode += CALL_MEIREI;
								if(pst->byteadress == 0){
									pst->backpatch += bytecode.GetLength();
								}
								AtaiCode(bytecode, pst->byteadress);
								if((ktype.id = TokenTtoTypeId(pst->rettype.GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
								ktype.kwsk = pst->rettype;
								tystack += ktype.id;
								bytecode += MEMBERPOP_MEIREI;
							}else{
								return false;
							}
							break;
						}
						if(iap[i] >= 0) return false;
						il = 0;
						if(IsSoushokuToken((TokenKind)pst->funchensuu.GetConst()[indt].type.GetConst()[il])) il++;
						if(pst->funchensuu.GetConst()[indt].type.GetConst()[il] != CLASS_TOKEN
							&& pst->funchensuu.GetConst()[indt].type.GetConst()[il] != ENUM_TOKEN) return false;
						il++;
						tyki = (int32_t)pst->funchensuu.GetConst()[indt].type.GetConst()[il];
						pst = stree.stat->funclist.Get(tyki);
						indt = ~iap[i];
						cnstf = cnstf || IsReadOnlyToken((TokenKind)pst->funchensuu.GetConst()[indt].type.GetConst()[0]);
						ig += pst->GetMemberOffset(indt);
					}
					if(iap[i] >= 0) break;
					pty = &pst->funchensuu.GetConst()[indt].type;
					il = 0;
					if(IsSoushokuToken((TokenKind)pty->GetConst()[il])) il++;
					if(iap[i+1] == VOID_TOKEN){
						i++;
						if(pty->GetConst()[il] == THREAD_TOKEN){
							if(ktype.id != ADDRESS_TYPEID){
								bytecode += INDEXTYPE_MEIREI;
								ktype.id = ADDRESS_TYPEID;
							}
							bytecode += HENSUUSETNULL_MEIREI;
							AtaiCode(bytecode, ig);
						}else if(pty->GetConst()[il] == REF_TOKEN || pty->GetConst()[il] == LIST_TOKEN){
							if(pty->GetConst()[il+1] == NULL_TOKEN){
								if(ktype.id != DYNREF_TYPEID && ktype.id != DYNLIST_TYPEID){
									bytecode += DYNREFTYPE_MEIREI;
									ktype.id = VOID_TYPEID;
								}
								bytecode += HENSUUSETNULL_MEIREI;
								AtaiCode(bytecode, ig);
							}else if(pty->GetConst()[il+1] != OTHERKIND_TOKEN){
								if(ktype.id != REF_TYPEID && ktype.id != LIST_TYPEID){
									bytecode += REFTYPE_MEIREI;
									ktype.id = VOID_TYPEID;
								}
								bytecode += HENSUUSETNULL_MEIREI;
								AtaiCode(bytecode, ig);
							}
						}else{
							return false;
						}
						i++;//Z~R΂
						if(iap[i] != SEMICOLON_TOKEN) return false;
						break;
					}
					if(iap[i+1] == DELETE_TOKEN || iap[i+1] == NEW_TOKEN){
						if(cnstf) return false;
						i++;
						if(pty->GetConst()[il] == THREAD_TOKEN){
							bytecode += HENSUUKARASANSHOU_MEIREI;
							AtaiCode(bytecode, ig);
							bytecode += SANSHOUTHREADWAIT_MEIREI;
						}else if(pty->GetConst()[il] == REF_TOKEN || pty->GetConst()[il] == LIST_TOKEN){
							if(pty->GetConst()[il+1] != OTHERKIND_TOKEN){
								bytecode += HENSUUKARASANSHOU_MEIREI;
								AtaiCode(bytecode, ig);
								if(iap[i] != NEW_TOKEN || iap[i+1] != DAIKAKKOOPEN_TOKEN){
									if(!RefCountDown(stree, bytecode, ktype, SANSHOUREFCNTDOWN_MEIREI, *pty, iap[i]==DELETE_TOKEN)) return false;
								}
							}
						}else{
							return false;
						}
						if(iap[i] == NEW_TOKEN){
							i++;
							if(pty->GetConst()[il+1] == CONST_TOKEN || pty->GetConst()[il+1] == OTHERKIND_TOKEN) return false;
							if(iap[i] == SHOUKAKKOOPEN_TOKEN){
								i++;
								if(pty->GetConst()[il] == REF_TOKEN){
									if(!RefNew(stree, bytecode, iap, i, tystack, ktype, SANSHOUREFNEW_MEIREI, *pty)) return false;
								}else{
									if(!RefNew(stree, bytecode, iap, i, tystack, ktype, SANSHOULISTNEW_MEIREI, *pty)) return false;
								}
							}else if(iap[i] == DAIKAKKOOPEN_TOKEN){
								i++;
								if(pty->GetConst()[il] != LIST_TOKEN) return false;
								if(!RefIdxNew(stree, bytecode, iap, i, tystack, ktype, *pty)) return false;
								il++;
								type.Clear();
								type.AddTail(pty->GetConst()+il, pty->GetLength()-il);
								if(!Sanshou(stree, bytecode, iap, i, tystack, ktype, type, false)) return false;
							}else{
								return false;
							}
						}else{
							i++;
							if(iap[i] != SEMICOLON_TOKEN) return false;
						}
						break;
					}
					if(iap[i+1] == SHOUKAKKOOPEN_TOKEN && iap[i+2] == TENTEN_TOKEN){
						if(cnstf) return false;
						i += 3;
						bytecode += HENSUUKARASANSHOU_MEIREI;
						AtaiCode(bytecode, ig);
						if(pty->GetConst()[il] != THREAD_TOKEN) return false;
						il++;
						if(Kansuu(stree, bytecode, iap, i, tystack, ktype, *stree.stat->funclist.Get((int32_t)pty->GetConst()[il]), true) < 0) return false;
						bytecode += SANSHOUTHREADRUN_MEIREI;
						ktype.id = VOID_TYPEID;
						break;
					}
					tmpb = pty->GetConst()[il] == CLASS_TOKEN;
					typ.pstre = pst->funchensuu.GetConst()[indt].tree;
					typ.kwsk = *pty;
					if((typ.id = TokenTtoTypeId(pty->GetConst(), il=0)) == UNKNOWN_TYPEID){
						if(pty->GetConst()[0] == PLUGIN_TOKEN && iap[i+1] == KAKKOCOLOOPEN_TOKEN){
							i += 2;
							ktype = typ;
							if((tmpip = PluginKansuu(stree, bytecode, iap, i, tystack, ktype, pst, waketa)) < 0) return false;
							if(stree.frp->rettype.GetConst()[0] == CONST_TOKEN && waketa.GetConst()[0].GetConst()[0] != CONST_TOKEN) return false;
							if(cnstf && !IsReadOnlyToken((TokenKind)waketa.GetConst()[0].GetConst()[0])) return false;
							if((ktype.id = TokenTtoTypeId(waketa.GetConst()[0].GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
							ktype.kwsk = waketa.GetConst()[0];
							bytecode += PLUGINCALL_MEIREI;
							bytecode += (uint8_t)(stree.TypeIdtoSize(ktype.id) + (ktype.id == FLOAT_TYPEID || ktype.id == DOUBLE_TYPEID ? 128 : 0));
							AtaiCode(bytecode, pty->GetTail());
							if(ktype.id != VOID_TYPEID){
								bytecode += TypeIdtoMeireiT(stree, ktype.id);
							}
							tystack += ktype.id;
							break;
						}else{
							return false;
						}
					}
					if(iap[i+1] == NULL_TOKEN && iap[i+2] == TENDAINYUU_TOKEN){
						if(cnstf) return false;
						i += 2;
						bytecode += HENSUUKARASANSHOU_MEIREI;
						AtaiCode(bytecode, ig);
						if(!Append(stree, bytecode, iap, i, tystack, ktype, typ)) return false;
						break;
					}
					if(iap[i+1] == NULL_TOKEN && (DAINYUU_TOKEN <= iap[i+2] && iap[i+2] <= TENDAINYUU_TOKEN)){
						if(cnstf) return false;
						i += 2;
						if(!Dainyuu(stree, bytecode, iap, i, tystack, ktype, typ, ig, PUSHHENSUU_MEIREI, tmpb , false)) return false;
						break;
					}
					if(
						(iap[i+1] == NULL_TOKEN && (iap[i+2] == PLSPLS_TOKEN || iap[i+2] == MNSMNS_TOKEN))
						|| iap[i+1] == PLSPLS_TOKEN || iap[i+1] == MNSMNS_TOKEN)
					{
						if(cnstf) return false;
						i++;
						bytecode += HENSUUKARASANSHOU_MEIREI;
						AtaiCode(bytecode, ig);
						if(!IncDec(stree, bytecode, iap, i, tystack, ktype, typ)) return false;
						break;
					}
					ktype.pstre = typ.pstre;
					ktype.kwsk = typ.kwsk;
					if(typ.id != ktype.id){
						if(typ.id == VOID_TYPEID) return false;
						ktype.id = typ.id;
						if(!tmpb && !(ENDTHREAD_TYPEID <= ktype.id && ktype.id < 0)){
							if((deco = TypeIdtoMeireiT(stree, ktype.id)) == NOP_MEIREI) return false;
							bytecode += deco;
						}
					}
					if(cnstf && ENDTHREAD_TYPEID <= ktype.id && ktype.id < 0) return false;
					if(ktype.id != VOID_TYPEID){
						if(pty->GetConst()[0] == CONST_TOKEN){
							bytecode += PUSHTEISUU_MEIREI;
							if(ktype.id == FUNC_TYPEID){
								if(!stree.stat->funclist.Get((int32_t)pty->GetTail())->kakutei) return false;
								if(stree.stat->funclist.Get((int32_t)pty->GetTail())->byteadress == 0){
									stree.stat->funclist.Get((int32_t)pty->GetTail())->backpatch += bytecode.GetLength();
								}
								AtaiCode(bytecode, stree.stat->funclist.Get((int32_t)pty->GetTail())->byteadress);
							}else if(ktype.id == REF_TYPEID){
								bytecode += PUSHGLOBALHENSUU_MEIREI;
								AtaiCode(bytecode, pty->GetTail());
							}else{
								if((deco = TypeIdtoMeireiT(stree, ktype.id)) == NOP_MEIREI) return false;
								TypeAtai(bytecode, deco, pty->GetConst()+il);
							}
						}else{
							if(tmpb || (ENDTHREAD_TYPEID <= ktype.id && ktype.id < 0)){
								bytecode += HENSUUKARASANSHOU_MEIREI;
								AtaiCode(bytecode, ig);
							}else if(ktype.id == DELEGATE_TYPEID){
								bytecode += HENSUUKARASANSHOU_MEIREI;
								AtaiCode(bytecode, ig);
								bytecode += LOADSANSHOUKARASANSHOUSAKI_MEIREI;
								AtaiCode(bytecode, stree.TypeIdtoSize(ktype.id));
							}else{
								bytecode += PUSHHENSUU_MEIREI;
								AtaiCode(bytecode, ig);
							}
						}
						if(ktype.id == METHOD_TYPEID && iap[i+1] == KAKKOCOLOOPEN_TOKEN){
							i += 2;
							ig -= (pst == stree.frp ? stree.crp->GetMemberOffset(indt) : pst->GetMemberOffset(indt));
							if((tmpip = KansuuPointer(stree, bytecode, iap, i, tystack, ktype, pst, waketa)) < 0) return false;
							if(stree.frp->rettype.GetConst()[0] == CONST_TOKEN && waketa.GetConst()[0].GetConst()[0] != CONST_TOKEN) return false;
							if(cnstf && !IsReadOnlyToken((TokenKind)waketa.GetConst()[0].GetConst()[0])) return false;
							if(!pst->touka){
								bytecode += MEMBERPUSHHENSUU_MEIREI;
								AtaiCode(bytecode, ig);
							}
							if(tmpip > 0){
								bytecode += INDEXTYPE_MEIREI;
								bytecode += LOCALPUSH_MEIREI;
								bytecode += PUSHHENSUU_MEIREI;
								AtaiCode(bytecode, tmpip);
							}
							bytecode += STACKCALL_MEIREI;
							if((ktype.id = TokenTtoTypeId(waketa.GetConst()[0].GetConst(), il=0)) == UNKNOWN_TYPEID) return false;
							ktype.kwsk = waketa.GetConst()[0];
							if(tmpip > 0){
								if(ktype.id != VOID_TYPEID){
									bytecode += POPTOHENSUU_MEIREI;
									AtaiCode(bytecode, tmpip+Aliszof<intptr_t>()-Aligner(stree.GetTypeSize(waketa.GetConst()[0])));
								}
								bytecode += HENSUUSTACKKAIHOU_MEIREI;
								AtaiCode(bytecode, Aliszof<intptr_t>()-Aligner(stree.GetTypeSize(waketa.GetConst()[0])));
							}
							if(!pst->touka) bytecode += MEMBERPOP_MEIREI;
						}
					}
					tystack += ktype.id;
				}
				break;
			}
			return true;
		}
		bool Sankou(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype)
		{
			intptr_t bp1, bp2;
			KuwashiiType typ;
			SrcTree *retcall = NULL;
			if(tystack.GetTail() != BOOL_TYPEID) return false;
			bytecode += POPNFALSEJUMP_MEIREI;
			bp1 = bytecode.GetLength();
			AtaiCode(bytecode, (intptr_t)0);
			tystack.Pop();
			ktype.id = VOID_TYPEID;
			tystack += WILDCARD_TYPEID;
			for(i++; !(iap[i] == QUESTION_TOKEN || iap[i] == COLON_TOKEN); i++){
				if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
			}
			if(iap[i] == QUESTION_TOKEN) return false;
			typ = ktype;
			bytecode += JUMP_MEIREI;
			bp2 = bytecode.GetLength();
			AtaiCode(bytecode, (intptr_t)0);
			tystack.SetLength(tystack.GetLength()-2);
			*(intptr_t *)(bytecode.GetConst()+bp1) = bytecode.GetLength();
			ktype.id = VOID_TYPEID;
			tystack += typ.id;
			for(i++; !(iap[i] == QUESTION_TOKEN || iap[i] == COLON_TOKEN); i++){
				if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
			}
			if(iap[i] == COLON_TOKEN) return false;
			if(tystack.GetTail() == UNKNOWN_TYPEID) return false;
			if(tystack.GetTail() == BOOL_TYPEID && ktype.id != BOOL_TYPEID){
				bytecode += BOOLTYPE_MEIREI;
				ktype.id = BOOL_TYPEID;
			}
			if(typ.id == REF_TYPEID || typ.id == LIST_TYPEID){
				if(ktype.id != REF_TYPEID && ktype.id != LIST_TYPEID) return false;
				if(!stree.YuruiKataCheck(ktype.kwsk, typ.kwsk, *pomc->filename, pomc->gyousuu)){
					if(!stree.YuruiKataCheck(typ.kwsk, ktype.kwsk, *pomc->filename, pomc->gyousuu)) return false;
					ktype = typ;
				}
			}else if(typ.id == FUNC_TYPEID || typ.id == METHOD_TYPEID){
				if(ktype.id != FUNC_TYPEID && ktype.id != METHOD_TYPEID) return false;
				if(!stree.YuruiKataCheck(ktype.kwsk, typ.kwsk, *pomc->filename, pomc->gyousuu)){
					if(!stree.YuruiKataCheck(typ.kwsk, ktype.kwsk, *pomc->filename, pomc->gyousuu)) return false;
					ktype = typ;
				}
			}else{
				if(!JidouCast(stree, bytecode, tystack, typ.id, ktype.id)) return false;
				ktype = typ;
			}
			*(intptr_t *)(bytecode.GetConst()+bp2) = bytecode.GetLength();
			tystack.Pop();
			tystack.Tail() = ktype.id;
			return true;
		}
		bool AAOO(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype)
		{
			intptr_t bp;
			SrcTree *retcall = NULL;
			if(tystack.GetTail() != BOOL_TYPEID) return false;
			switch(iap[i]){
			case ANDAND_TOKEN:
				bytecode += FALSEJUMP_MEIREI;
				break;
			case OROR_TOKEN:
				bytecode += TRUEJUMP_MEIREI;
				break;
			default:
				return false;
			}
			bp = bytecode.GetLength();
			AtaiCode(bytecode, (intptr_t)0);
			if(ktype.id != BOOL_TYPEID){
				bytecode += BOOLTYPE_MEIREI;
				ktype.id = BOOL_TYPEID;
			}
			bytecode += POP_MEIREI;
			tystack.Tail() = BOOL_TYPEID;
			for(i++; !(iap[i] == ANDAND_TOKEN || iap[i] == OROR_TOKEN); i++){
				if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
			}
			if(tystack.Pop() != BOOL_TYPEID) return false;
			*(intptr_t*)(bytecode.GetConst()+bp) = bytecode.GetLength();
			return true;
		}
		bool IncDecSub(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype, const KuwashiiType &oktype, const MeireiCode deco)
		{
			if(iap[i+1] == SEMICOLON_TOKEN || iap[i+1] == SHARP_TOKEN){
				switch(deco){
				case INCLKARAATAI_MEIREI:
				case INCLKARAMAENOATAI_MEIREI:
					bytecode += INCL_MEIREI;
					break;
				case DECLKARAATAI_MEIREI:
				case DECLKARAMAENOATAI_MEIREI:
					bytecode += DECL_MEIREI;
					break;
				default:
					return false;
				}
				ktype.id = VOID_TYPEID;
				i += 1;
			}else if(iap[i+1] == NULL_TOKEN && (DAINYUU_TOKEN <= iap[i+2] && iap[i+2] <= RSHIFTDAINYUU_TOKEN)){
				switch(deco){
				case INCLKARAATAI_MEIREI:
				case INCLKARAMAENOATAI_MEIREI:
					bytecode += PUSHSTACKADDRESS_MEIREI;
					bytecode += INCL_MEIREI;
					break;
				case DECLKARAATAI_MEIREI:
				case DECLKARAMAENOATAI_MEIREI:
					bytecode += PUSHSTACKADDRESS_MEIREI;
					bytecode += DECL_MEIREI;
					break;
				default:
					return false;
				}
				i += 2;
				if(!Dainyuu(stree, bytecode, iap, i, tystack, ktype, oktype, 0, PUSHSTACKSANSHOUSAKI_MEIREI, false, true)) return false;
			}else if(
				(iap[i+1] == NULL_TOKEN && (iap[i+2] == PLSPLS_TOKEN || iap[i+2] == MNSMNS_TOKEN))
				|| iap[i+1] == PLSPLS_TOKEN || iap[i+1] == MNSMNS_TOKEN)
			{
				switch(deco){
				case INCLKARAATAI_MEIREI:
				case INCLKARAMAENOATAI_MEIREI:
					bytecode += PUSHSTACKADDRESS_MEIREI;
					bytecode += INCL_MEIREI;
					break;
				case DECLKARAATAI_MEIREI:
				case DECLKARAMAENOATAI_MEIREI:
					bytecode += PUSHSTACKADDRESS_MEIREI;
					bytecode += DECL_MEIREI;
					break;
				default:
					return false;
				}
				i++;
				if(!IncDec(stree, bytecode, iap, i, tystack, ktype, oktype)) return false;
			}else{
				bytecode += deco;
				ktype = oktype;
				tystack += ktype.id;
			}
			return true;
		}
		bool IncDec(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype, const KuwashiiType &oktype)
		{
			if(oktype.id < ADDRESS_TYPEID || BYTE_TYPEID < oktype.id) return false;
			if(ktype.id != oktype.id){
				bytecode += TypeIdtoMeireiT(stree, oktype.id);
				ktype = oktype;
			}
			if(iap[i] == NULL_TOKEN){
				i++;
				switch(iap[i]){
				case PLSPLS_TOKEN:
					if(!IncDecSub(stree, bytecode, iap, i, tystack, ktype, oktype, INCLKARAMAENOATAI_MEIREI)) return false;
					break;
				case MNSMNS_TOKEN:
					if(!IncDecSub(stree, bytecode, iap, i, tystack, ktype, oktype, DECLKARAMAENOATAI_MEIREI)) return false;
					break;
				default:
					return false;
				}
			}else{
				switch(iap[i]){
				case PLSPLS_TOKEN:
					if(!IncDecSub(stree, bytecode, iap, i, tystack, ktype, oktype, INCLKARAATAI_MEIREI)) return false;
					break;
				case MNSMNS_TOKEN:
					if(!IncDecSub(stree, bytecode, iap, i, tystack, ktype, oktype, DECLKARAATAI_MEIREI)) return false;
					break;
				default:
					return false;
				}
			}
			return true;
		}
		bool AppendSub(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype, const KuwashiiType &oktype, const MeireiCode deco, const int32_t typ)
		{
			if(iap[i+1] == SEMICOLON_TOKEN || iap[i+1] == SHARP_TOKEN){
				bytecode += deco;
				if(deco == LOADAPPEND_MEIREI) AtaiCode(bytecode, stree.stat->funclist.Get(typ)->GetMemberSize());
				tystack.Pop();
				ktype.id = VOID_TYPEID;
				i += 1;
			}else if(iap[i+1] == NULL_TOKEN && iap[i+2] == TENDAINYUU_TOKEN){
				switch(deco){
				case APPEND_MEIREI:
					bytecode += APPENDKARASANSHOU_MEIREI;
					break;
				case LOADAPPEND_MEIREI:
					bytecode += LOADAPPENDKARASANSHOU_MEIREI;
					AtaiCode(bytecode, stree.stat->funclist.Get(typ)->GetMemberSize());
					break;
				default:
					return false;
				}
				tystack.Pop();
				i += 2;
				if(!Append(stree, bytecode, iap, i, tystack, ktype, oktype)) return false;
			}else if(iap[i+1] == NULL_TOKEN && iap[i+2] == DAINYUU_TOKEN){
				switch(deco){
				case APPEND_MEIREI:
					bytecode += APPENDKARASANSHOU_MEIREI;
					break;
				case LOADAPPEND_MEIREI:
					bytecode += LOADAPPENDKARASANSHOU_MEIREI;
					AtaiCode(bytecode, stree.stat->funclist.Get(typ)->GetMemberSize());
					break;
				default:
					return false;
				}
				tystack.Pop();
				i += 2;
				if(!Dainyuu(stree, bytecode, iap, i, tystack, ktype, oktype, 0, PUSHSTACKSANSHOUSAKI_MEIREI, false, true)) return false;
			}else{
				switch(deco){
				case APPEND_MEIREI:
					bytecode += APPENDKARASANSHOU_MEIREI;
					break;
				case LOADAPPEND_MEIREI:
					bytecode += LOADAPPENDKARASANSHOU_MEIREI;
					AtaiCode(bytecode, stree.stat->funclist.Get(typ)->GetMemberSize());
					break;
				default:
					return false;
				}
				if(ktype.id != REF_TYPEID && ktype.id != LIST_TYPEID) bytecode += REFTYPE_MEIREI;
				bytecode += SANSHOUKARASANSHOUSAKI_MEIREI;
				ktype = oktype;
				tystack.Tail() = ktype.id;
			}
			return true;
		}
		bool Append(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype, const KuwashiiType &oktype)
		{
			int32_t typ;
			intptr_t il;
			SrcTree *retcall = NULL;
			if(oktype.id != LIST_TYPEID) return false;
			for(i++; iap[i] != TENDAINYUU_TOKEN; i++){
				if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
			}
			il = 0;
			if(IsSoushokuToken((TokenKind)oktype.kwsk.GetConst()[il])) il++;
			if(oktype.kwsk.GetConst()[il] != LIST_TOKEN) return false;
			il++;
			if((typ = TokenTtoTypeId(oktype.kwsk.GetConst(), il)) == UNKNOWN_TYPEID) return false;
			if(
				typ < 0
				|| (typ <= ENDCLASS_TYPEID && stree.stat->funclist.Get(typ)->selftype == CLASS_BLOCK && !stree.stat->funclist.Get(typ)->IsOpen()))
			{
				return false;
			}
			if(tystack.GetTail() == UNKNOWN_TYPEID) return false;
			if(tystack.GetTail() == BOOL_TYPEID && ktype.id != BOOL_TYPEID){
				bytecode += BOOLTYPE_MEIREI;
				ktype.id = BOOL_TYPEID;
			}
			if((ktype.id == REF_TYPEID || ktype.id == LIST_TYPEID)){
				if(!stree.YuruiKataCheck(oktype.kwsk, ktype.kwsk, *pomc->filename, pomc->gyousuu)) return false;
				if(!AppendSub(stree, bytecode, iap, i, tystack, ktype, oktype, APPEND_MEIREI, typ)) return false;
			}else{
				if(!JidouCast(stree, bytecode, tystack, typ, ktype.id)) return false;
				ktype.id = typ;
				if(typ <= ENDCLASS_TYPEID && stree.stat->funclist.Get(typ)->selftype == CLASS_BLOCK){
					if(!AppendSub(stree, bytecode, iap, i, tystack, ktype, oktype, LOADAPPEND_MEIREI, typ)) return false;
				}else{
					if(!AppendSub(stree, bytecode, iap, i, tystack, ktype, oktype, APPEND_MEIREI, typ)) return false;
				}
			}
			return true;
		}
		bool Dainyuu(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype, const KuwashiiType &oktype, const intptr_t ig, const MeireiCode deco, const bool classke, const bool mata)
		{
			intptr_t ii;
			int32_t did;
			MeireiCode dekyo;
			SrcTree *pst;
			SrcTree *retcall = NULL;
			if(
				iap[i] == DAINYUU_TOKEN ? oktype.id == VOID_TYPEID || (ENDTHREAD_TYPEID <= oktype.id && oktype.id < 0) : (
					oktype.id <= ENDCLASS_TYPEID ? true : (
						ANDDAINYUU_TOKEN <= iap[i] && iap[i] <= ORDAINYUU_TOKEN ?
						oktype.id == FLOAT_TYPEID || oktype.id == DOUBLE_TYPEID : oktype.id == BOOL_TYPEID || (
							iap[i] == LSHIFTDAINYUU_TOKEN || iap[i] == RSHIFTDAINYUU_TOKEN ?
							oktype.id == FLOAT_TYPEID || oktype.id == DOUBLE_TYPEID : false
						)
					)
				))
			{
				return false;
			}
			if(oktype.kwsk.GetLength() == 2 &&
				oktype.kwsk.GetConst()[0] == REF_TOKEN && oktype.kwsk.GetConst()[1] == OTHERKIND_TOKEN && iap[i+1] == COMMA_TOKEN){
				i++;
				return true;
			}
			if(oktype.id == DELEGATE_TYPEID){
				if(deco != PUSHHENSUU_MEIREI) return false;
			}else if(mata){
				if(iap[i] == DAINYUU_TOKEN && iap[i+1] == COMMA_TOKEN){
					tystack.Tail() = UNKNOWN_TYPEID;
					return true;
				}
				if(iap[i] != DAINYUU_TOKEN) bytecode += PUSHSTACKSANSHOUSAKI_MEIREI;
			}else{
				switch(deco){
				case PUSHHENSUU_MEIREI:
					bytecode += HENSUUKARASANSHOU_MEIREI;
					AtaiCode(bytecode, ig);
					break;
				case PUSHMEMBERHENSUU_MEIREI:
					bytecode += MEMBERKARASANSHOU_MEIREI;
					AtaiCode(bytecode, ig);
					break;
				case PUSHGLOBALHENSUU_MEIREI:
					bytecode += GLOBALKARASANSHOU_MEIREI;
					AtaiCode(bytecode, ig);
					break;
				case PUSHSTACKSANSHOUSAKI_MEIREI:
					break;
				default:
					return false;
				}
				if(iap[i] != DAINYUU_TOKEN){
					if(ktype.id != oktype.id){
						if((dekyo = TypeIdtoMeireiT(stree, oktype.id)) == NOP_MEIREI) return false;
						bytecode += dekyo;
						ktype = oktype;
					}
					bytecode += deco;
					if(deco != PUSHSTACKSANSHOUSAKI_MEIREI) AtaiCode(bytecode, ig);
					tystack += ktype.id;
				}else if(iap[i+1] == COMMA_TOKEN){
					tystack += UNKNOWN_TYPEID;
					ktype = oktype;
					return true;
				}
			}
			tystack += oktype.id;
			for(i++; !(DAINYUU_TOKEN <= iap[i] && iap[i] <= TENDAINYUU_TOKEN); i++){
				if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
			}
			if(tystack.GetTail() == UNKNOWN_TYPEID) return false;
			if(tystack.GetTail() == BOOL_TYPEID && ktype.id != BOOL_TYPEID){
				bytecode += BOOLTYPE_MEIREI;
				ktype.id = BOOL_TYPEID;
			}
			did = REF_TYPEID;//tOOFF
			switch(oktype.id){
			case REF_TYPEID:
			case LIST_TYPEID:
				if((ktype.id != REF_TYPEID && ktype.id != LIST_TYPEID) || !stree.YuruiKataCheck(ktype.kwsk, oktype.kwsk, *pomc->filename, pomc->gyousuu)) return false;
				break;
			case DYNREF_TYPEID:
				if(ktype.id == REF_TYPEID || ktype.id == LIST_TYPEID){
					ii = 0;
					if(IsSoushokuToken((TokenKind)ktype.kwsk.GetConst()[ii])) ii++;
					if(ktype.kwsk.GetConst()[ii] != REF_TOKEN && ktype.kwsk.GetConst()[ii] != LIST_TOKEN) return false;
					ii++;
					did = TokenTtoTypeId(ktype.kwsk.GetConst(), ii);
					if(did == REF_TYPEID || did == LIST_TYPEID || did == DYNREF_TYPEID || did == DYNLIST_TYPEID){
						pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"^G[łB");
						return false;
					}
					bytecode += DYNREFTYPE_MEIREI;
				}else if(ktype.id != DYNREF_TYPEID && ktype.id != DYNLIST_TYPEID){
					pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"^G[łB");
					return false;
				}
				break;
			case DYNLIST_TYPEID:
				if(ktype.id == LIST_TYPEID){
					ii = 0;
					if(IsSoushokuToken((TokenKind)ktype.kwsk.GetConst()[ii])) ii++;
					if(ktype.kwsk.GetConst()[ii] != LIST_TOKEN) return false;
					ii++;
					did = TokenTtoTypeId(ktype.kwsk.GetConst(), ii);
					if(did == REF_TYPEID || did == LIST_TYPEID || did == DYNREF_TYPEID || did == DYNLIST_TYPEID){
						pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"^G[łB");
						return false;
					}
					bytecode += DYNREFTYPE_MEIREI;
				}else if(ktype.id != DYNLIST_TYPEID){
					pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"^G[łB");
					return false;
				}
				break;
			case FUNC_TYPEID:
			case METHOD_TYPEID:
				if((ktype.id != FUNC_TYPEID && ktype.id != METHOD_TYPEID) || !stree.YuruiKataCheck(ktype.kwsk, oktype.kwsk, *pomc->filename, pomc->gyousuu)) return false;
				break;
			case DELEGATE_TYPEID:
				if(ktype.id != DELEGATE_TYPEID || !stree.YuruiKataCheck(ktype.kwsk, oktype.kwsk, *pomc->filename, pomc->gyousuu)) return false;
				pst = oktype.pstre;
				while(pst != ktype.pstre){
					if(pst == NULL){
						pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"qubNŐ錾ꂽdelegate^͐eubN̕ϐɑł܂B");
						return false;
					}
					pst = pst->parent;
				}
				break;
			default:
				if(!JidouCast(stree, bytecode, tystack, oktype.id, ktype.id)) return false;
				break;
			}
			ktype = oktype;
			switch(iap[i]){
			case PULUSDAINYUU_TOKEN:
			case MINUSDAINYUU_TOKEN:
			case KAKERUDAINYUU_TOKEN:
			case WARUDAINYUU_TOKEN:
			case AMARIDAINYUU_TOKEN:
			case ANDDAINYUU_TOKEN:
			case XORDAINYUU_TOKEN:
			case ORDAINYUU_TOKEN:
			case LSHIFTDAINYUU_TOKEN:
			case RSHIFTDAINYUU_TOKEN:
				switch(iap[i]){
				case PULUSDAINYUU_TOKEN:
					bytecode += TASU_MEIREI;
					break;
				case MINUSDAINYUU_TOKEN:
					bytecode += HIKU_MEIREI;
					break;
				case KAKERUDAINYUU_TOKEN:
					bytecode += KAKERU_MEIREI;
					break;
				case WARUDAINYUU_TOKEN:
					bytecode += WARU_MEIREI;
					break;
				case AMARIDAINYUU_TOKEN:
					bytecode += AMARI_MEIREI;
					break;
				case ANDDAINYUU_TOKEN:
					bytecode += AND_MEIREI;
					break;
				case XORDAINYUU_TOKEN:
					bytecode += XOR_MEIREI;
					break;
				case ORDAINYUU_TOKEN:
					bytecode += OR_MEIREI;
					break;
				case LSHIFTDAINYUU_TOKEN:
					bytecode += LSHIFT_MEIREI;
					break;
				case RSHIFTDAINYUU_TOKEN:
					bytecode += RSHIFT_MEIREI;
					break;
				}
				tystack.Pop();
			case DAINYUU_TOKEN:
				if(classke && !stree.stat->funclist.Get(oktype.id)->IsOpen()) return false;
				if(iap[i+1] == SEMICOLON_TOKEN || iap[i+1] == SHARP_TOKEN){
					if(classke){
						bytecode += LOAD_MEIREI;
						AtaiCode(bytecode, stree.stat->funclist.Get(oktype.id)->GetMemberSize());
					}else if(ktype.id == REF_TYPEID || ktype.id == LIST_TYPEID || ktype.id == DYNREF_TYPEID || ktype.id == DYNLIST_TYPEID){
						if(did != REF_TYPEID){
							bytecode += SANSHOUDYNREFSET_MEIREI;
							AtaiCode(bytecode, did);
							bytecode += POPADDRESS_MEIREI;
						}else{
							if(!RefCountDown(stree, bytecode, ktype, SANSHOUREFCOPY_MEIREI, ktype.kwsk, true)) return false;
						}
					}else if(ktype.id == DELEGATE_TYPEID){
						bytecode += POPTOHENSUU_MEIREI;
						AtaiCode(bytecode, ig);
						bytecode += POPTOHENSUU_MEIREI;
						AtaiCode(bytecode, (intptr_t)(ig+Aliszof<intptr_t>()));
						bytecode += POPTOHENSUU_MEIREI;
						AtaiCode(bytecode, (intptr_t)(ig+Aliszof<intptr_t>()*2));
					}else{
						bytecode += DAINYUU_MEIREI;
					}
					tystack.SetLength(tystack.GetLength()-2);
					ktype.id = VOID_TYPEID;
					i++;
				}else if(iap[i+1] == NULL_TOKEN && iap[i+2] == TENDAINYUU_TOKEN){
					if(ktype.id != LIST_TYPEID) return false;
					if(!RefCountDown(stree, bytecode, ktype, SANSHOUREFCOPY_MEIREI, ktype.kwsk, false)) return false;
					tystack.SetLength(tystack.GetLength()-2);
					i += 2;
					if(!Append(stree, bytecode, iap, i, tystack, ktype, oktype)) return false;
				}else if(iap[i+1] == NULL_TOKEN && (DAINYUU_TOKEN <= iap[i+2] && iap[i+2] <= TENDAINYUU_TOKEN)){
					if(classke){
						bytecode += LOADKARASANSHOU_MEIREI;
						AtaiCode(bytecode, stree.stat->funclist.Get(oktype.id)->GetMemberSize());
					}else if(ktype.id == REF_TYPEID || ktype.id == LIST_TYPEID || ktype.id == DYNREF_TYPEID || ktype.id == DYNLIST_TYPEID){
						if(did != REF_TYPEID){
							bytecode += SANSHOUDYNREFSET_MEIREI;
							AtaiCode(bytecode, did);
						}else{
							if(!RefCountDown(stree, bytecode, ktype, SANSHOUREFCOPY_MEIREI, ktype.kwsk, false)) return false;
						}
					}else if(ktype.id == DELEGATE_TYPEID){
						return false;
					}else{
						bytecode += DAINYUUKARASANSHOU_MEIREI;
					}
					tystack.SetLength(tystack.GetLength()-2);
					i += 2;
					if(!Dainyuu(stree, bytecode, iap, i, tystack, ktype, oktype, 0, PUSHSTACKSANSHOUSAKI_MEIREI, false, true)) return false;
				}else if(
					(iap[i+1] == NULL_TOKEN && (iap[i+2] == PLSPLS_TOKEN || iap[i+2] == MNSMNS_TOKEN))
					|| iap[i+1] == PLSPLS_TOKEN || iap[i+1] == MNSMNS_TOKEN)
				{
					bytecode += DAINYUUKARASANSHOU_MEIREI;
					tystack.SetLength(tystack.GetLength()-2);
					i++;
					if(!IncDec(stree, bytecode, iap, i, tystack, ktype, oktype)) return false;
				}else{
					if(classke){
						bytecode += LOADKARASANSHOU_MEIREI;
						AtaiCode(bytecode, stree.stat->funclist.Get(oktype.id)->GetMemberSize());
					}else if(ktype.id == REF_TYPEID || ktype.id == LIST_TYPEID || ktype.id == DYNREF_TYPEID || ktype.id == DYNLIST_TYPEID){
						if(did != REF_TYPEID){
							bytecode += SANSHOUDYNREFSET_MEIREI;
							AtaiCode(bytecode, did);
						}else{
							if(!RefCountDown(stree, bytecode, ktype, SANSHOUREFCOPY_MEIREI, ktype.kwsk, false)) return false;
						}
						ktype = oktype;
						bytecode += (ktype.id == REF_TYPEID || ktype.id == LIST_TYPEID ? REFTYPE_MEIREI : DYNREFTYPE_MEIREI);
						bytecode += SANSHOUKARASANSHOUSAKI_MEIREI;
					}else if(ktype.id == DELEGATE_TYPEID){
						return false;
					}else{
						bytecode += DAINYUUKARASANSHOU_MEIREI;
						bytecode += SANSHOUKARASANSHOUSAKI_MEIREI;
					}
					tystack.Pop();
					tystack.Tail() = ktype.id;
				}
				break;
			default:
				return false;
			}
			return true;
		}
		bool MeijiCast(SrcTree &stree, PWCArray<uint8_t> &bytecode, PWCArray<int32_t> &tystack, const int32_t dtype, int32_t typ)
		{
			if(tystack.GetTail() == UNKNOWN_TYPEID){
				pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"sȃLXgłB");
				return false;
			}
			if(tystack.GetTail() == BOOL_TYPEID && typ != BOOL_TYPEID){
				bytecode += BOOLTYPE_MEIREI;
				typ = BOOL_TYPEID;
			}
			if(typ == dtype) return true;
			if(dtype == BOOL_TYPEID || dtype <= ENDCLASS_TYPEID || typ < 0){
				pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"sȃLXgłB");
				return false;
			}
			if(typ <= ENDCLASS_TYPEID){
				if(stree.stat->funclist.Get(typ)->selftype != ENUM_BLOCK){
					pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"sȃLXgłB");
					return false;
				}
				if(dtype == INT_TYPEID) return true;
			}
			bytecode += TypeIdtoMeireiT(stree, dtype) + CASTBYTETYPE_MEIREI - BYTETYPE_MEIREI;
			tystack.Tail() = dtype;
			return true;
		}
		bool JidouCast(SrcTree &stree, PWCArray<uint8_t> &bytecode, PWCArray<int32_t> &tystack, const int32_t dtype, int32_t typ)
		{
			if(tystack.GetTail() == UNKNOWN_TYPEID){
				pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"^G[łB");
				return false;
			}
			if(tystack.GetTail() == BOOL_TYPEID && typ != BOOL_TYPEID){
				bytecode += BOOLTYPE_MEIREI;
				typ = BOOL_TYPEID;
			}
			if(typ == dtype) return true;
			if( (
				(dtype == ADDRESS_TYPEID || BYTE_TYPEID >= dtype && dtype >= LONG_TYPEID)
				&& (typ == ADDRESS_TYPEID || BYTE_TYPEID >= typ && typ >= LONG_TYPEID)
				)
				|| (UBYTE_TYPEID >= dtype && dtype >= ULONG_TYPEID && UBYTE_TYPEID >= typ && typ >= ULONG_TYPEID)
				|| ((dtype == FLOAT_TYPEID || dtype == DOUBLE_TYPEID) && (typ == FLOAT_TYPEID || typ == DOUBLE_TYPEID)))
			{
				bytecode += TypeIdtoMeireiT(stree, dtype) + CASTBYTETYPE_MEIREI - BYTETYPE_MEIREI;
				tystack.Tail() = dtype;
				return true;
			}
			pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"^G[łB");
			return false;
		}
		bool If(SrcTree &stree, PWCArray<uint8_t> &bytecode, intptr_t &j)
		{
			intptr_t len, bp1, i;
			const intptr_t *iap;
			KuwashiiType ktype;
			PWCArray<int32_t> tystack;
			SrcTree *retcall = NULL;
			tystack.Clear();
			iap = stree.token.GetConst()[j].GetConst();
			pomc->gyousuu = stree.gyo.GetConst()[j];
			ktype.id = VOID_TYPEID;
			i = 1;
			if(iap[i] != SHOUKAKKOOPEN_TOKEN) return false;
			i++;
			tystack += BOOL_TYPEID;
			do{
				if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
				i++;
			}while(iap[i] != SHOUKAKKOCLOSE_TOKEN);
			if(tystack.GetLength() <= 0 || tystack.GetTail() != BOOL_TYPEID) return false;
			bytecode += POPNFALSEJUMP_MEIREI;
			bp1 = bytecode.GetLength();
			AtaiCode(bytecode, (intptr_t)0);
			tystack.SetLength(tystack.GetLength()-2);
			j++;
			iap = stree.token.GetConst()[j].GetConst();
			len = stree.token.GetConst()[j].GetLength();
			pomc->gyousuu = stree.gyo.GetConst()[j];
			ktype.id = VOID_TYPEID;
			if(!KoubunSwitch(stree, bytecode, iap, len, j)) return false;
			*(intptr_t *)(bytecode.GetConst()+bp1) = bytecode.GetLength();
			return true;
		}
		void YobunPoper(PWCArray<uint8_t> &bytecode)
		{
			if(pomc->yobunipop == 0) return;
			bytecode += LOCALPUSH_MEIREI;
			bytecode += HENSUUSTACKKAIHOU_MEIREI;
			AtaiCode(bytecode, pomc->yobunipop);
			pomc->yobunipop = 0;
		}
		bool MainLoop(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, const intptr_t len, intptr_t &i, SrcTree *&retcall)
		{
			KuwashiiType ktype;
			PWCArray<int32_t> tystack;
			tystack.Clear();
			ktype.id = VOID_TYPEID;
			for(; i < len; i++){
				if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
			}
			return true;
		}
		bool MainLoopBase(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, const intptr_t len, intptr_t &i, SrcTree *&retcall)
		{
			switch(stree.selftype){
			case FUNC_BLOCK:
				if(!pomc->fcom.MainLoop(stree, bytecode, iap, len, i, retcall)) return false;
				return true;
			case NORMAL_BLOCK:
				if(!pomc->ncom.MainLoop(stree, bytecode, iap, len, i, retcall)) return false;
				return true;
			case LOCK_BLOCK:
				if(!pomc->lkcom.MainLoop(stree, bytecode, iap, len, i, retcall)) return false;
				return true;
			case SWITCH_BLOCK:
				if(!pomc->swcom.MainLoop(stree, bytecode, iap, len, i, retcall)) return false;
				return true;
			case BRANCH_BLOCK:
				if(!pomc->brcom.MainLoop(stree, bytecode, iap, len, i, retcall)) return false;
				return true;
			case LOOP_BLOCK:
				if(!pomc->lpcom.MainLoop(stree, bytecode, iap, len, i, retcall)) return false;
				return true;
			case DELEGATE_BLOCK:
				if(!pomc->dgcom.MainLoop(stree, bytecode, iap, len, i, retcall)) return false;
				return true;
			}
			return false;
		}
		bool DefDel(SrcTree &stree, PWCArray<uint8_t> &bytecode)
		{
			intptr_t i;
			KuwashiiType ktype;
			PWCArray<int32_t> tystack;
			SrcTree *retcall = NULL;
			ktype.id = VOID_TYPEID;
			for(i = 0; i < stree.defdel.GetLength(); i++){
				if(!MainSwitch(stree, bytecode, stree.defdel.GetConst(), i, tystack, ktype, retcall)) return false;
			}
			return true;
		}
		bool DefDelBase(SrcTree &stree, PWCArray<uint8_t> &bytecode)
		{
			switch(stree.selftype){
			case FUNC_BLOCK:
				if(!pomc->fcom.DefDel(stree, bytecode)) return false;
				return true;
			case NORMAL_BLOCK:
				if(!pomc->ncom.DefDel(stree, bytecode)) return false;
				return true;
			case LOCK_BLOCK:
				if(!pomc->lkcom.DefDel(stree, bytecode)) return false;
				return true;
			case SWITCH_BLOCK:
				if(!pomc->swcom.DefDel(stree, bytecode)) return false;
				return true;
			case BRANCH_BLOCK:
				if(!pomc->brcom.DefDel(stree, bytecode)) return false;
				return true;
			case LOOP_BLOCK:
				if(!pomc->lpcom.DefDel(stree, bytecode)) return false;
				return true;
			case DELEGATE_BLOCK:
				if(!pomc->dgcom.DefDel(stree, bytecode)) return false;
				return true;
			}
			return false;
		}
		bool KoubunSwitch(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, const intptr_t len, intptr_t &j)
		{
			intptr_t i;
			SrcTree *retcall = NULL;
			if(iap[0] == IF_TOKEN){
				if(!If(stree, bytecode, j)) return false;
			}else{
				i = 0;
				if(!MainLoopBase(stree, bytecode, iap, len, i, retcall)) return false;
			}
			return true;
		}
		bool KoubunSwitchBase(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, const intptr_t len, intptr_t &j)
		{
			switch(stree.selftype){
			case FUNC_BLOCK:
				if(!pomc->fcom.KoubunSwitch(stree, bytecode, iap, len, j)) return false;
				return true;
			case NORMAL_BLOCK:
				if(!pomc->ncom.KoubunSwitch(stree, bytecode, iap, len, j)) return false;
				return true;
			case LOCK_BLOCK:
				if(!pomc->lkcom.KoubunSwitch(stree, bytecode, iap, len, j)) return false;
				return true;
			case SWITCH_BLOCK:
				if(!pomc->swcom.KoubunSwitch(stree, bytecode, iap, len, j)) return false;
				return true;
			case BRANCH_BLOCK:
				if(!pomc->brcom.KoubunSwitch(stree, bytecode, iap, len, j)) return false;
				return true;
			case LOOP_BLOCK:
				if(!pomc->lpcom.KoubunSwitch(stree, bytecode, iap, len, j)) return false;
				return true;
			case DELEGATE_BLOCK:
				if(!pomc->dgcom.KoubunSwitch(stree, bytecode, iap, len, j)) return false;
				return true;
			}
			return false;
		}
	public:
		template<typename T> void AtaiCode(PWCArray<uint8_t> &bytecode, const T x)
		{
			bytecode.SetLength(bytecode.GetLength()+sizeof(T));
			*(T*)(bytecode.GetConst()+bytecode.GetLength()-sizeof(T)) = x;
		}
		void SetOmc(OmCompiler *const po)
		{
			pomc = po;
		}
	};
	class FuncCompi : public XCompi
	{
	public:
		bool FuncTreeCompile(SrcTree &stree, PWCArray<uint8_t> &bytecode)
		{
			intptr_t i, j, len, glbbp;
			KuwashiiType ktype;
			PWCArray<int32_t> tystack;
			PWCArray<wchar_t> tmps;
			const intptr_t *iap;
			if(!stree.kakutei){
				if(stree.GetIjousaki(len, stree, NULL) != NULL){
					for(i = 0; i < stree.subfunc.GetLength(); i++){
						if(!FuncTreeCompile(*stree.subfunc.GetConst()[i], bytecode)) return false;
					}
					return true;
				}
				pomc->ss->ErrorMessage.Add(stree.srce->filename, stree.topsrcgyo, L"`łB");
				return false;
			}
			if(stree.selftype == ENUM_BLOCK) return true;
			if(stree.selftype == CLASS_BLOCK || (stree.tpcount > 0 && stree.funchensuu.GetConst()[stree.tpcount-1].type.GetLength() <= 1)){
				for(i = 0; i < stree.subfunc.GetLength(); i++){
					if(!FuncTreeCompile(*stree.subfunc.GetConst()[i], bytecode)) return false;
				}
				return true;
			}
			pomc->filename = &stree.srce->filename;
			stree.byteadress = bytecode.GetLength();
			if(tmps = L"break", stree.myname == tmps){
				stree.parent->deladress = stree.byteadress;
				if(&stree != stree.root && stree.parent->junkandata.GetLength() > 0){
					pomc->ss->CircularGC.SetPLDestructor(stree.parent->sonota.junkanid, stree.byteadress);
				}
			}
			Backpatch(bytecode, stree.backpatch);
			if(stree.root != stree.frp) KansuStackSet(stree, bytecode);
			for(j = 0; j < stree.token.GetLength(); j++){
				iap = stree.token.GetConst()[j].GetConst();
				len = stree.token.GetConst()[j].GetLength();
				if(len == 0) continue;
				pomc->gyousuu = stree.gyo.GetConst()[j];
				if(!KoubunSwitchBase(stree, bytecode, iap, len, j)){
					if(pomc->ss->ErrorMessage.mes.GetLength() == 1) pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"RpCG[łB");
					return false;
				}
			}
			if(stree.selftype == FUNC_BLOCK && stree.rettype.GetTail() != VOID_TOKEN){
				if(stree.token.GetLength() <= 0 || stree.token.GetConst()[j-1].GetConst()[len-1] != RETURN_TOKEN
					|| (stree.token.GetLength() >= 2 && stree.token.GetConst()[j-2].GetConst()[0] == IF_TOKEN))
				{
					pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"return܂B");
					return false;
				}
			}else{
				ktype.id = VOID_TYPEID;
				if(&stree == stree.root){
					bytecode += GBLDELPUSH_MEIREI;
					AtaiCode(bytecode, (intptr_t)(bytecode.GetLength()+sizeof(intptr_t)+1+sizeof(intptr_t)));
					bytecode += JUMP_MEIREI;
					glbbp = bytecode.GetLength();
					AtaiCode(bytecode, (intptr_t)0);
				}
				if(!DefDelBase(stree, bytecode)){
					pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"RpCG[B");
					return false;
				}
				if(stree.root == &stree){
					*(intptr_t*)(bytecode.GetData()+glbbp) = bytecode.GetLength();
				}else{
					KansuStackClear(stree, bytecode);
				}
				bytecode += RETURN_MEIREI;
			}
			for(i = 0; i < stree.subfunc.GetLength(); i++){
				if(!FuncTreeCompile(*stree.subfunc.GetConst()[i], bytecode)) return false;
			}
			return true;
		}
	};
	class NormalCompi : public XCompi
	{
	protected:
		bool Break(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, const intptr_t len, intptr_t &i, PWCArray<int32_t> &tystack, KuwashiiType &ktype, SrcTree *&retcall)
		{
			i++;
			if(stree.breakin){
				bytecode += CALL_MEIREI;
				stree.breakpatch += bytecode.GetLength();
				AtaiCode(bytecode, (intptr_t)0);
			}
			while(iap[i] < 0 || iap[i] == TILDE_TOKEN || iap[i] == MBR_TOKEN){
				for(; ; i++){
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
					if(iap[i] == SEMICOLON_TOKEN) break;
				}
				i++;
			}
			if(iap[i] == COMMA_TOKEN){
				i++;
				if(!MainLoopBase(*stree.parent, bytecode, iap, len, i, retcall)) return false;
			}else{
				if(iap[i] != SEMICOLON_TOKEN) return false;
				i++;
			}
			YobunPoper(bytecode);
			bytecode += JUMP_MEIREI;//COMMAłJUMP(RETURN)Ƃ͌ȂB
			stree.backpatch += bytecode.GetLength();
			AtaiCode(bytecode, (intptr_t)0);
			YobunPoper(bytecode);
			return true;
		}
	public:
		bool MainLoop(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, const intptr_t len, intptr_t &i, SrcTree *&retcall)
		{
			KuwashiiType ktype;
			PWCArray<int32_t> tystack;
			tystack.Clear();
			ktype.id = VOID_TYPEID;
			for(; i < len; i++){
				if(iap[i] == BREAK_TOKEN){
					if(!Break(stree, bytecode, iap, len, i, tystack, ktype, retcall)) return false;
				}else{
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
				}
			}
			return true;
		}
		bool KoubunSwitch(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, const intptr_t len, intptr_t &j)
		{
			SrcTree *retcall = NULL;
			if(iap[0] == BREAK_TOKEN && iap[1] == COLON_TOKEN){
				bytecode += CALL_MEIREI;
				stree.breakpatch += bytecode.GetLength();
				AtaiCode(bytecode, (intptr_t)0);
				if(!DefDelBase(stree, bytecode)) return false;
				bytecode += JUMP_MEIREI;
				stree.backpatch += bytecode.GetLength();
				AtaiCode(bytecode, (intptr_t)0);
				Backpatch(bytecode, stree.breakpatch);
			}else{
				if(!XCompi::KoubunSwitch(stree, bytecode, iap, len, j)) return false;
			}
			return true;
		}
		bool FuncTreeCompile(SrcTree &stree, PWCArray<uint8_t> &bytecode)
		{
			intptr_t j, len;
			PWCArray<int32_t> tystack;
			const intptr_t *iap;
			stree.byteadress = bytecode.GetLength();
			for(j = 0; j < stree.token.GetLength(); j++){
				iap = stree.token.GetConst()[j].GetConst();
				len = stree.token.GetConst()[j].GetLength();
				if(len == 0) continue;
				pomc->gyousuu = stree.gyo.GetConst()[j];
				if(!KoubunSwitchBase(stree, bytecode, iap, len, j)) return false;
			}
			if(stree.breakin){
				bytecode += RETURN_MEIREI;
			}else{
				if(!DefDelBase(stree, bytecode)) return false;
			}
			Backpatch(bytecode, stree.backpatch);
			return true;
		}
	};
	class LockCompi : public NormalCompi
	{
	public:
		bool DefDel(SrcTree &stree, PWCArray<uint8_t> &bytecode)
		{
			intptr_t i;
			if(!NormalCompi::DefDel(stree, bytecode)) return false;
			for(i = 0; i < stree.hikisuucount; i++) bytecode += STACKCALL_MEIREI;
			return true;
		}
		bool MainLoop(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, const intptr_t len, intptr_t &i, SrcTree *&retcall)
		{
			KuwashiiType ktype;
			PWCArray<int32_t> tystack;
			tystack.Clear();
			ktype.id = VOID_TYPEID;
			intptr_t j;
			for(; i < len; i++){
				if(iap[i] == BREAK_TOKEN){
					i++;
					if(stree.breakin){
						bytecode += CALL_MEIREI;
						stree.breakpatch += bytecode.GetLength();
						AtaiCode(bytecode, (intptr_t)0);
					}
					while(iap[i] < 0 || iap[i] == TILDE_TOKEN || iap[i] == MBR_TOKEN){
						for(; ; i++){
							if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
							if(iap[i] == SEMICOLON_TOKEN) break;
						}
						i++;
					}
					for(j = 0; j < stree.hikisuucount; j++) bytecode += STACKCALL_MEIREI;
					if(iap[i] == COMMA_TOKEN){
						i++;
						if(!MainLoopBase(*stree.parent, bytecode, iap, len, i, retcall)) return false;
					}else{
						if(iap[i] != SEMICOLON_TOKEN) return false;
						i++;
					}
					YobunPoper(bytecode);
					bytecode += JUMP_MEIREI;
					stree.backpatch += bytecode.GetLength();
					AtaiCode(bytecode, (intptr_t)0);
				}else{
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
				}
			}
			return true;
		}
	};
	class SwitchCompi : public NormalCompi
	{
		void CaseTable(PWCArray<uint8_t> &bytecode, SwitchExtStruct &ses, const PWCArray<bool> &uetikb, PWCArray<intptr_t> &shnnkn, const uint8_t ichi, const MeireiCode mtype)
		{
			intptr_t i, j, tbli;
			PWCArrayQuad< PWCArray<int> > numbn;
			PWCArray<bool> tikb;
			bytecode += SWITCHJUMP_MEIREI;
			bytecode += ichi;
			tbli = bytecode.GetLength();
			bytecode.SetLength(bytecode.GetLength() + sizeof(intptr_t)*16);
			numbn.SetLength(16);
			for(i = 0; i < uetikb.GetLength(); i++){
				if(uetikb.GetConst()[i]) continue;
				numbn.GetData()[(*((uint8_t *)(ses.atai.GetConst()+i)+(ichi>>1)) >> (ichi&0x01)*4) & 0x0f] += i;
			}
			for(i = 0; i < 16; i++){
				if(numbn.GetConst()[i].GetLength() <= 0){
					shnnkn += tbli + sizeof(intptr_t)*i;
				}else if(numbn.GetConst()[i].GetLength() == 1){
					*(intptr_t *)(bytecode.GetConst()+tbli+sizeof(intptr_t)*i) = bytecode.GetLength();
					bytecode += PUSHTEISUU_MEIREI;
					TypeAtai(bytecode, mtype, &ses.atai.GetConst()[numbn.GetConst()[i].GetConst()[0]]);
					bytecode += EQUAL_MEIREI;
					bytecode += POPNTRUEJUMP_MEIREI;
					ses.ataibp[numbn.GetConst()[i].GetConst()[0]] = bytecode.GetLength();
					AtaiCode(bytecode, (intptr_t)0);
					bytecode += JUMP_MEIREI;
					ses.defbp += bytecode.GetLength();
					AtaiCode(bytecode, (intptr_t)0);
				}
			}
			tikb.SetLength(uetikb.GetLength());
			for(i = 0; i < tikb.GetLength(); i++) tikb.GetData()[i] = true;
			for(i = 0; i < 16; i++){
				if(numbn.GetConst()[i].GetLength() < 2) continue;
				for(j = 0; j < numbn.GetConst()[i].GetLength(); j++) tikb.GetData()[numbn.GetConst()[i].GetConst()[j]] = false;
				*(intptr_t *)(bytecode.GetConst()+tbli+sizeof(intptr_t)*i) = bytecode.GetLength();
				CaseTable(bytecode, ses, tikb, shnnkn, ichi+1, mtype);
				for(j = 0; j < numbn.GetConst()[i].GetLength(); j++) tikb.GetData()[numbn.GetConst()[i].GetConst()[j]] = true;
			}
		}
	public:
		bool MainLoop(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, const intptr_t len, intptr_t &i, SrcTree *&retcall)
		{
			KuwashiiType ktype;
			PWCArray<int32_t> tystack;
			tystack.Clear();
			ktype.id = VOID_TYPEID;
			for(; i < len; i++){
				switch(iap[i]){
				case DEFAULT_TOKEN:
					i++;
					YobunPoper(bytecode);
					bytecode += JUMP_MEIREI;
					if(stree.extswitch->def == 0) stree.extswitch->defbp += bytecode.GetLength();
					AtaiCode(bytecode, stree.extswitch->def);
					break;
				case BREAK_TOKEN:
					if(!Break(stree, bytecode, iap, len, i, tystack, ktype, retcall)) return false;
					break;
				default:
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
					break;
				}
			}
			return true;
		}
		bool KoubunSwitch(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, const intptr_t len, intptr_t &j)
		{
			intptr_t i, ii, iii, sz;
			MeireiCode deco;
			int32_t typ;
			int64_t t64;
			if(iap[0] == CASE_TOKEN || (iap[0] == DEFAULT_TOKEN && (iap[1] == COLON_TOKEN || iap[1] == COLOYAJIRUSHI_TOKEN))){
				if(!stree.extswitch->teeburutukutta){
					PWCArray<bool> tikb;
					PWCArray<intptr_t> popbp;
					for(i = 0; i < stree.token.GetLength(); i++){
						ii = 0;
						if(stree.token.GetConst()[i].GetConst()[ii] != CASE_TOKEN) continue;
						do{
							ii++;
							if((typ = TokenTtoTypeId(stree.token.GetConst()[i].GetConst(), ii)) < 0){
								pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"caseɂl܂B");
								return false;
							}
							stree.extswitch->atai += (int64_t)0;
							sz = stree.CastConst(typ, stree.token.GetConst()[i].GetConst()+ii,
								stree.extswitch->swtype, &stree.extswitch->atai.Tail());
							if(sz < 0){
								pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"caseɂl܂B");
								return false;
							}
							for(iii = 0; iii < stree.extswitch->atai.GetLength()-1; iii++){
								if(stree.extswitch->atai.GetConst()[iii] == stree.extswitch->atai.GetTail()){
									pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"caseɓl܂B");
									return false;
								}
							}
							ii += (sz + sizeof(intptr_t)-1) / sizeof(intptr_t);
						}while(stree.token.GetConst()[i].GetConst()[ii] == COMMA_TOKEN);
					}
					popbp.Clear();
					tikb.SetLength(stree.extswitch->atai.GetLength());
					for(ii = 0; ii < tikb.GetLength(); ii++) tikb.GetData()[ii] = false;
					deco = TypeIdtoMeireiT(stree, stree.extswitch->swtype);
					bytecode += deco;
					CaseTable(bytecode, *stree.extswitch, tikb, popbp, 0, deco);
					Backpatch(bytecode, popbp);
					bytecode += POP_MEIREI;
					bytecode += JUMP_MEIREI;
					stree.extswitch->defbp += bytecode.GetLength();
					AtaiCode(bytecode, (intptr_t)0);
					stree.extswitch->teeburutukutta = true;
				}
				if(iap[len-1] == COLOYAJIRUSHI_TOKEN){
					Backpatch(bytecode, stree.extswitch->tgnsuruusakibp);
				}else{
					bytecode += JUMP_MEIREI;
					stree.extswitch->tgnsuruusakibp += bytecode.GetLength();
					AtaiCode(bytecode, (intptr_t)0);
				}
				if(iap[0] == DEFAULT_TOKEN){
					if(stree.extswitch->def != 0) return false;
					stree.extswitch->def = bytecode.GetLength();
					Backpatch(bytecode, stree.extswitch->defbp);
				}else{
					ii = 0;
					do{
						ii++;
						if((typ = TokenTtoTypeId(iap, ii)) < 0) return false;
						t64 = 0;
						sz = stree.CastConst(typ, iap+ii, stree.extswitch->swtype, &t64);
						if(sz < 0) return false;
						ii += (sz + sizeof(intptr_t)-1) / sizeof(intptr_t);
						for(iii = 0; iii < stree.extswitch->atai.GetLength(); iii++){
							if(stree.extswitch->atai.GetConst()[iii] == t64){
								*(intptr_t*)(bytecode.GetConst()+stree.extswitch->ataibp.GetConst()[iii]) = bytecode.GetLength();
								break;
							}
						}
					}while(iap[ii] == COMMA_TOKEN);
				}
			}else{
				if(iap[0] == BREAK_TOKEN && iap[1] == COLON_TOKEN){
					Backpatch(bytecode, stree.extswitch->tgnsuruusakibp);
					if(stree.extswitch->def == 0) Backpatch(bytecode, stree.extswitch->defbp);
				}
				if(!NormalCompi::KoubunSwitch(stree, bytecode, iap, len, j)) return false;
			}
			return true;
		}
		bool DefDel(SrcTree &stree, PWCArray<uint8_t> &bytecode)
		{
			if(!stree.breakin){
				Backpatch(bytecode, stree.extswitch->tgnsuruusakibp);
				if(stree.extswitch->def == 0) Backpatch(bytecode, stree.extswitch->defbp);
			}
			if(!NormalCompi::DefDel(stree, bytecode)) return false;
			return true;
		}
		bool FuncTreeCompile(SrcTree &stree, PWCArray<uint8_t> &bytecode)
		{
			if(!NormalCompi::FuncTreeCompile(stree, bytecode)) return false;
			if(!stree.extswitch->teeburutukutta){
				bytecode += TypeIdtoMeireiT(stree, stree.extswitch->swtype);
				bytecode += POP_MEIREI;
			}
			return true;
		}
	};
	class BranchCompi : public NormalCompi
	{
		void CondTable(PWCArray<uint8_t> &bytecode, BranchExtStruct &bes, const PWCArray<bool> &uetikb, const uint8_t ichi)
		{
			int32_t i, j;
			intptr_t tbli;
			PWCArrayQuad< PWCArray<int32_t> > numbn;
			PWCArray<bool> tikb;
			bytecode += SWITCHJUMP_MEIREI;
			bytecode += ichi;
			tbli = bytecode.GetLength();
			bytecode.SetLength(bytecode.GetLength() + sizeof(intptr_t)*16);
			numbn.SetLength(16);
			for(i = 0; i < uetikb.GetLength(); i++){
				if(uetikb.GetConst()[i]) continue;
				numbn.GetData()[(*((uint8_t*)(&i)+(ichi>>1)) & (0x0f<<(ichi&0x01)*4)) >> (ichi&0x01)*4] += i;
			}
			for(i = 0; i < 16; i++){
				if(numbn.GetConst()[i].GetLength() <= 0){
					bes.tgnelsesakibp += tbli + sizeof(intptr_t)*i;
				}else if(numbn.GetConst()[i].GetLength() == 1){
					*(intptr_t*)(bytecode.GetConst()+tbli+sizeof(intptr_t)*i) = bytecode.GetLength();
					bytecode += PUSHSTACK_MEIREI;
					bytecode += PUSHTEISUU_MEIREI;
					AtaiCode(bytecode, numbn.GetConst()[i].GetConst()[0]);
					bytecode += EQUAL_MEIREI;
					bytecode += POPNTRUEJUMP_MEIREI;
					bes.labelbp.GetData()[numbn.GetConst()[i].GetConst()[0]] = bytecode.GetLength();
					AtaiCode(bytecode, (intptr_t)0);
					bytecode += JUMP_MEIREI;
					bes.tgnelsesakibp += bytecode.GetLength();
					AtaiCode(bytecode, (intptr_t)0);
				}
			}
			tikb.SetLength(uetikb.GetLength());
			for(i = 0; i < tikb.GetLength(); i++) tikb.GetData()[i] = true;
			for(i = 0; i < 16; i++){
				if(numbn.GetConst()[i].GetLength() < 2) continue;
				for(j = 0; j < numbn.GetConst()[i].GetLength(); j++) tikb.GetData()[numbn.GetConst()[i].GetConst()[j]] = false;
				*(intptr_t*)(bytecode.GetConst()+tbli+sizeof(intptr_t)*i) = bytecode.GetLength();
				CondTable(bytecode, bes, tikb, ichi+1);
				for(j = 0; j < numbn.GetConst()[i].GetLength(); j++) tikb.GetData()[numbn.GetConst()[i].GetConst()[j]] = true;
			}
		}
		void MakeTable(PWCArray<uint8_t> &bytecode, BranchExtStruct &bes)
		{
			intptr_t ii;
			PWCArray<bool> tikb;
			bes.labelbp.SetLength(bes.label.GetLength());
			tikb.SetLength(bes.label.GetLength());
			for(ii = 0; ii < tikb.GetLength(); ii++) tikb.GetData()[ii] = false;
			bytecode += INTTYPE_MEIREI;
			CondTable(bytecode, bes, tikb, 0);
		}
		void LabelBP(PWCArray<uint8_t> &bytecode, BranchExtStruct &bes)
		{
			intptr_t i;
			for(i = 0; i < bes.labelbp.GetLength(); i++){
				if(bes.labelbp.GetConst()[i] != 0){
					*(intptr_t*)(bytecode.GetConst()+bes.labelbp.GetConst()[i]) = bytecode.GetLength();
					bes.labelbp.GetData()[i] = 0;
				}
			}
		}
	public:
		bool MainLoop(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, const intptr_t len, intptr_t &i, SrcTree *&retcall)
		{
			KuwashiiType ktype;
			PWCArray<int32_t> tystack;
			tystack.Clear();
			ktype.id = VOID_TYPEID;
			for(; i < len; i++){
				switch(iap[i]){
				case ELSE_TOKEN:
					i++;
					if(stree.extbranch->mode < 0){
						bytecode += INTTYPE_MEIREI;
						if(stree.extbranch->bunkisita != 0) bytecode += POP_MEIREI;
						bytecode += PUSHTEISUU_MEIREI;
						AtaiCode(bytecode, (int32_t)-1);
					}
					YobunPoper(bytecode);
					bytecode += JUMP_MEIREI;
					stree.extbranch->tgnelsesakibp += bytecode.GetLength();
					AtaiCode(bytecode, (intptr_t)0);
					break;
				case BREAK_TOKEN:
					if(stree.extbranch->mode < 0 && stree.extbranch->bunkisita != 0) pomc->yobunipop += Aliszof<int32_t>();
					if(!Break(stree, bytecode, iap, len, i, tystack, ktype, retcall)) return false;
					break;
				default:
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
					break;
				}
			}
			return true;
		}
		bool KoubunSwitch(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, const intptr_t len, intptr_t &j)
		{
			KuwashiiType ktype;
			PWCArray<int32_t> tystack;
			SrcTree *retcall = NULL;
			intptr_t i;
			if(iap[0] == COND_TOKEN){
				stree.extbranch->bunkisita = 1;
				if(stree.extbranch->nextcond != 0){
					bytecode += JUMP_MEIREI;
					stree.extbranch->tgncommsakibp += bytecode.GetLength();
					AtaiCode(bytecode, (intptr_t)0);
					*(intptr_t*)(bytecode.GetConst()+stree.extbranch->nextcond) = bytecode.GetLength();
					stree.extbranch->nextcond = 0;
				}
				if(stree.extbranch->mode < 0){
					i = 2;
				}else{
					i = 1;
				}
				ktype.id = VOID_TYPEID;
				tystack += BOOL_TYPEID;
				do{
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
					i++;
				}while(iap[i] != COLON_TOKEN);
				if(tystack.GetLength() <= 0 || tystack.GetTail() != BOOL_TYPEID) return false;
				bytecode += POPNFALSEJUMP_MEIREI;
				stree.extbranch->nextcond = bytecode.GetLength();
				AtaiCode(bytecode, (intptr_t)0);
				tystack.SetLength(tystack.GetLength()-2);
				if(stree.extbranch->mode < 0){
					bytecode += INTTYPE_MEIREI;
					bytecode += PUSHTEISUU_MEIREI;
					AtaiCode(bytecode, (int32_t)iap[1]);
				}
			}else if(iap[0] == ELSE_TOKEN && iap[1] == COLON_TOKEN){
				if(stree.extbranch->commnai){
					MakeTable(bytecode, *stree.extbranch);
					stree.extbranch->commnai = false;
				}
				bytecode += JUMP_MEIREI;
				stree.extbranch->tgncommsakibp += bytecode.GetLength();
				AtaiCode(bytecode, (intptr_t)0);
				if(stree.extbranch->nextcond != 0){
					*(intptr_t *)(bytecode.GetConst()+stree.extbranch->nextcond) = bytecode.GetLength();
					stree.extbranch->nextcond = 0;
					if(stree.extbranch->mode < 0 && stree.extbranch->bunkisita > 0){
						bytecode += INTTYPE_MEIREI;
						bytecode += PUSHTEISUU_MEIREI;
						AtaiCode(bytecode, (int32_t)-1);
					}
				}
				Backpatch(bytecode, stree.extbranch->tgnelsesakibp);
			}else if(iap[0] == COMM_TOKEN && iap[1] == COLON_TOKEN){
				stree.extbranch->bunkisita = -1;
				if(stree.extbranch->nextcond != 0){
					*(intptr_t *)(bytecode.GetConst()+stree.extbranch->nextcond) = bytecode.GetLength();
					stree.extbranch->nextcond = 0;
				}
				LabelBP(bytecode, *stree.extbranch);
				Backpatch(bytecode, stree.extbranch->tgnelsesakibp);
				Backpatch(bytecode, stree.extbranch->tgncommsakibp);
				stree.extbranch->commnai = true;
			}else if(iap[0] == DIFF_TOKEN && iap[2] == COLON_TOKEN){
				if(stree.extbranch->commnai){
					MakeTable(bytecode, *stree.extbranch);
					stree.extbranch->commnai = false;
				}
				bytecode += JUMP_MEIREI;
				stree.extbranch->tgncommsakibp += bytecode.GetLength();
				AtaiCode(bytecode, (intptr_t)0);
				*(intptr_t *)(bytecode.GetConst()+stree.extbranch->labelbp.GetConst()[iap[1]]) = bytecode.GetLength();
				stree.extbranch->labelbp.GetData()[iap[1]] = 0;
			}else{
				if(iap[0] == BREAK_TOKEN && iap[1] == COLON_TOKEN){
					LabelBP(bytecode, *stree.extbranch);
					if(stree.extbranch->nextcond != 0){
						*(intptr_t*)(bytecode.GetConst()+stree.extbranch->nextcond) = bytecode.GetLength();
						stree.extbranch->nextcond = 0;
					}
					Backpatch(bytecode, stree.extbranch->tgnelsesakibp);
					Backpatch(bytecode, stree.extbranch->tgncommsakibp);
					if(stree.extbranch->mode < 0 && stree.extbranch->bunkisita != 0){
						bytecode += INTTYPE_MEIREI;
						bytecode += POP_MEIREI;
					}
				}
				if(!NormalCompi::KoubunSwitch(stree, bytecode, iap, len, j)) return false;
			}
			return true;
		}
		bool DefDel(SrcTree &stree, PWCArray<uint8_t> &bytecode)
		{
			if(!stree.breakin){
				LabelBP(bytecode, *stree.extbranch);
				if(stree.extbranch->nextcond != 0){
					*(intptr_t*)(bytecode.GetConst()+stree.extbranch->nextcond) = bytecode.GetLength();
					stree.extbranch->nextcond = 0;
				}
				Backpatch(bytecode, stree.extbranch->tgnelsesakibp);
				Backpatch(bytecode, stree.extbranch->tgncommsakibp);
				if(stree.extbranch->mode < 0 && stree.extbranch->bunkisita != 0){
					bytecode += INTTYPE_MEIREI;
					bytecode += POP_MEIREI;
				}
			}
			if(!NormalCompi::DefDel(stree, bytecode)) return false;
			return true;
		}
	};
	class LoopCompi : public NormalCompi
	{
	public:
		bool MainLoop(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, const intptr_t len, intptr_t &i, SrcTree *&retcall)
		{
			KuwashiiType ktype;
			PWCArray<int32_t> tystack;
			tystack.Clear();
			ktype.id = VOID_TYPEID;
			for(; i < len; i++){
				switch(iap[i]){
				case DO_TOKEN:
					i++;
					YobunPoper(bytecode);
					bytecode += JUMP_MEIREI;
					if(stree.extloop->doad == 0) stree.extloop->doadbp += bytecode.GetLength();
					AtaiCode(bytecode, stree.extloop->doad);
					break;
				case CONTINUE_TOKEN:
					i++;
					YobunPoper(bytecode);
					bytecode += JUMP_MEIREI;
					if(stree.extloop->contad == 0) stree.extloop->contadbp += bytecode.GetLength();
					AtaiCode(bytecode, stree.extloop->contad);
					break;
				case WHILE_TOKEN:
					i++;
					YobunPoper(bytecode);
					bytecode += JUMP_MEIREI;
					if(stree.extloop->nextad == 0) stree.extloop->nextadbp += bytecode.GetLength();
					AtaiCode(bytecode, stree.extloop->nextad);
					break;
				case BREAK_TOKEN:
					if(!Break(stree, bytecode, iap, len, i, tystack, ktype, retcall)) return false;
					break;
				default:
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
					break;
				}
			}
			return true;
		}
		bool KoubunSwitch(SrcTree &stree, PWCArray<uint8_t> &bytecode, const intptr_t *const iap, const intptr_t len, intptr_t &j)
		{
			KuwashiiType ktype;
			PWCArray<int32_t> tystack;
			SrcTree *retcall = NULL;
			intptr_t i;
			if(iap[0] == DO_TOKEN && iap[1] == COLON_TOKEN){
				stree.extloop->doad = bytecode.GetLength();
				Backpatch(bytecode, stree.extloop->doadbp);
			}else if(iap[0] == CONTINUE_TOKEN && iap[1] == COLON_TOKEN){
				stree.extloop->contad = bytecode.GetLength();
				Backpatch(bytecode, stree.extloop->contadbp);
			}else if(iap[0] == WHILE_TOKEN && iap[1] != SEMICOLON_TOKEN){
				stree.extloop->nextad = bytecode.GetLength();
				if(stree.extloop->contadbp.GetLength() > 0){
					pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"continue:܂B");
					return false;
				}
				Backpatch(bytecode, stree.extloop->nextadbp);
				i = 1;
				ktype.id = VOID_TYPEID;
				tystack += BOOL_TYPEID;
				do{
					if(!MainSwitch(stree, bytecode, iap, i, tystack, ktype, retcall)) return false;
					i++;
				}while(iap[i] != COLON_TOKEN);
				if(tystack.GetLength() <= 0 || tystack.GetTail() != BOOL_TYPEID) return false;
				bytecode += POPNTRUEJUMP_MEIREI;
				AtaiCode(bytecode, stree.extloop->doad);
				tystack.SetLength(tystack.GetLength()-2);
			}else{
				if(iap[0] == BREAK_TOKEN && iap[1] == COLON_TOKEN){
					if(stree.extloop->doadbp.GetLength() > 0){
						pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"do:܂B");
						return false;
					}
					if(stree.extloop->contadbp.GetLength() > 0){
						pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"continue:܂B");
						return false;
					}
					if(stree.extloop->nextadbp.GetLength() > 0){
						pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"while:܂B");
						return false;
					}
				}
				if(!NormalCompi::KoubunSwitch(stree, bytecode, iap, len, j)) return false;
			}
			return true;
		}
		bool DefDel(SrcTree &stree, PWCArray<uint8_t> &bytecode)
		{
			if(!stree.breakin){
				if(stree.extloop->doadbp.GetLength() > 0){
					pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"do:܂B");
					return false;
				}
				if(stree.extloop->contadbp.GetLength() > 0){
					pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"continue:܂B");
					return false;
				}
				if(stree.extloop->nextadbp.GetLength() > 0){
					pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"while:܂B");
					return false;
				}
			}
			if(!NormalCompi::DefDel(stree, bytecode)) return false;
			return true;
		}
	};
	class DelegateCompi : public FuncCompi
	{
	public:
		bool FuncTreeCompile(SrcTree &stree, PWCArray<uint8_t> &bytecode)
		{
			intptr_t j;
			KuwashiiType ktype;
			const intptr_t *iap;
			intptr_t siz, zis;
			intptr_t len;
			pomc->filename = &stree.srce->filename;
			stree.byteadress = bytecode.GetLength();
			Backpatch(bytecode, stree.backpatch);
			siz = 0;
			for(j = stree.tpcount; j < stree.tpcount+stree.hikisuucount; j++){
				siz += stree.frp->funchensuu.GetConst()[j].size;
			}
			if(siz > 0){
				bytecode += POPTOLOCALSTACK_MEIREI;
				zis = stree.frp->GetMemberOffset(stree.tpcount+stree.hikisuucount-1);
				bytecode += HENSUUKARASANSHOU_MEIREI;
				AtaiCode(bytecode, zis);
				bytecode += LOCALPUSH_MEIREI;
				bytecode += HENSUUKARASANSHOU_MEIREI;
				AtaiCode(bytecode, Aliszof<intptr_t>());
				bytecode += LOAD_MEIREI;
				AtaiCode(bytecode, siz);
				bytecode += HENSUUSTACKKAIHOU_MEIREI;
				AtaiCode(bytecode, siz+Aliszof<intptr_t>()*2);
				bytecode += POPTOMEMBERSTACK_MEIREI;
			}else{
				bytecode += POPTOLOCALSTACK_MEIREI;
				bytecode += POPTOMEMBERSTACK_MEIREI;
			}
			for(j = 0; j < stree.token.GetLength(); j++){
				iap = stree.token.GetConst()[j].GetConst();
				len = stree.token.GetConst()[j].GetLength();
				if(len == 0) continue;
				pomc->gyousuu = stree.gyo.GetConst()[j];
				if(!KoubunSwitchBase(stree, bytecode, iap, len, j)) return false;
			}
			if(stree.rettype.GetTail() != VOID_TOKEN){
				if(stree.token.GetLength() <= 0 || stree.token.GetConst()[j-1].GetConst()[len-1] != RETURN_TOKEN
					|| (stree.token.GetLength() >= 2 && stree.token.GetConst()[j-2].GetConst()[0] == IF_TOKEN))
				{
					pomc->ss->ErrorMessage.Add(*pomc->filename, pomc->gyousuu, L"return܂B");
					return false;
				}
			}else{
				ktype.id = VOID_TYPEID;
				if(!DefDelBase(stree, bytecode)) return false;
				bytecode += LOCALPOP_MEIREI;
				bytecode += MEMBERPOP_MEIREI;
				bytecode += RETURN_MEIREI;
			}
			return true;
		}
	};
	FuncCompi fcom;
	NormalCompi ncom;
	LockCompi lkcom;
	SwitchCompi swcom;
	BranchCompi brcom;
	LoopCompi lpcom;
	DelegateCompi dgcom;
	OmCompiler(SSZStatic<JunkanGoroshi> *sstat)
	{
		ss = sstat;
		yobunipop = 0;
		fcom.SetOmc(this);
		ncom.SetOmc(this);
		lkcom.SetOmc(this);
		swcom.SetOmc(this);
		brcom.SetOmc(this);
		lpcom.SetOmc(this);
		dgcom.SetOmc(this);
	}
	bool Compile(SrcTree &sourceroot, PWCArray<uint8_t> &bytecode, PWCArray<int8_t> &globalres)
	{
		SrcTree *pst;
		intptr_t i, ttl, bp;
		if(sourceroot.parent != NULL){
			ss->ErrorMessage.Add(sourceroot.srce->filename, sourceroot.topsrcgyo, L"^̍łȂłB");
			return false;
		}
		ttl = sourceroot.stat->literaltotal;
		bytecode.Clear();
		pst = &sourceroot;
		while(pst->subsrc != NULL) pst = pst->subsrc;
		while(pst->parent != NULL){
			if(pst->token.GetLength() > 0 || pst->defdel.GetLength() > 0){
				bytecode += CALL_MEIREI;
				pst->backpatch += bytecode.GetLength();
				fcom.AtaiCode(bytecode, pst->byteadress);
			}
			bytecode += JUMP_MEIREI;
			bp = bytecode.GetLength();
			fcom.AtaiCode(bytecode, (intptr_t)0);
			pst->sonota.golbalgeta = ttl;
			if(!fcom.FuncTreeCompile(*pst, bytecode)) return false;
			*(intptr_t*)(bytecode.GetData()+bp) = bytecode.GetLength();
			ttl += pst->GetMemberSize();
			pst = pst->parent;
		}
		bytecode += CALL_MEIREI;
		sourceroot.backpatch += bytecode.GetLength();
		fcom.AtaiCode(bytecode, sourceroot.byteadress);
		bytecode += RETURN_MEIREI;
		sourceroot.sonota.golbalgeta = ttl;
		if(!fcom.FuncTreeCompile(sourceroot, bytecode)) return false;
		ttl += sourceroot.GetMemberSize();
		globalres.SetLength(ttl);
		ttl = 0;
		for(i = 0; i < sourceroot.stat->literallist.GetLength(); i++){
			memcpy(globalres.GetData()+ttl, sourceroot.stat->literallist.GetConst()[i].GetConst(), sourceroot.stat->literallist.GetConst()[i].GetLength());
			if(sourceroot.stat->literallist.GetConst()[i].GetLength() > sizeof(Reference)+sizeof(HeapObjHead)){
				((Reference*)(globalres.GetData()+ttl))->pointer = (HeapObj*)((intptr_t)(globalres.GetData()+ttl+sizeof(Reference)) | INTPTR_MIN);
			}else{
				((Reference*)(globalres.GetData()+ttl))->pointer = NULL;
			}
			((HeapObjHead*)(globalres.GetData()+ttl+sizeof(Reference)))->data = (int8_t*)(globalres.GetData()+ttl+sizeof(Reference)+sizeof(HeapObjHead));
			ttl += sourceroot.stat->literallist.GetConst()[i].GetLength();
		}
		return true;
	}
};

